#ifndef lint
char yysccsid[] = "@(#)yaccpar	1.4 (Berkeley) 02/25/90";
#endif
#line 2 "parser.y"
/*
 * Public Release 3
 * 
 * $Id: parser.y,v 1.8 1997/11/14 03:24:45 cwr Exp $
 */

/*
 * ------------------------------------------------------------------------
 * 
 * Copyright (c) 1996, 1997 The Regents of the University of Michigan
 * All Rights Reserved
 *  
 * Royalty-free licenses to redistribute GateD Release
 * 3 in whole or in part may be obtained by writing to:
 * 
 * 	Merit GateDaemon Project
 * 	4251 Plymouth Road, Suite C
 * 	Ann Arbor, MI 48105
 *  
 * THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE REGENTS OF THE
 * UNIVERSITY OF MICHIGAN AND MERIT DO NOT WARRANT THAT THE
 * FUNCTIONS CONTAINED IN THE SOFTWARE WILL MEET LICENSEE'S REQUIREMENTS OR
 * THAT OPERATION WILL BE UNINTERRUPTED OR ERROR FREE. The Regents of the
 * University of Michigan and Merit shall not be liable for
 * any special, indirect, incidental or consequential damages with respect
 * to any claim by Licensee or any third party arising from use of the
 * software. GateDaemon was originated and developed through release 3.0
 * by Cornell University and its collaborators.
 * 
 * Please forward bug fixes, enhancements and questions to the
 * gated mailing list: gated-people@gated.merit.edu.
 * 
 * ------------------------------------------------------------------------
 * 
 * Copyright (c) 1990,1991,1992,1993,1994,1995 by Cornell University.
 *     All rights reserved.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT
 * LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * GateD is based on Kirton's EGP, UC Berkeley's routing
 * daemon	 (routed), and DCN's HELLO routing Protocol.
 * Development of GateD has been supported in part by the
 * National Science Foundation.
 * 
 * ------------------------------------------------------------------------
 * 
 * Portions of this software may fall under the following
 * copyrights:
 * 
 * Copyright (c) 1988 Regents of the University of California.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms are
 * permitted provided that the above copyright notice and
 * this paragraph are duplicated in all such forms and that
 * any documentation, advertising materials, and other
 * materials related to such distribution and use
 * acknowledge that the software was developed by the
 * University of California, Berkeley.  The name of the
 * University may not be used to endorse or promote
 * products derived from this software without specific
 * prior written permission.  THIS SOFTWARE IS PROVIDED
 * ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */


#define INCLUDE_TIME
#define	MALLOC_OK
#include "include.h"
#include "inet.h"
#include "parse.h"
#include "krt.h"
#include "rip.h"
#include "hello.h"
#include "icmp.h"
#include "ospf.h"
#include "bgp.h"

#undef	vsprintf

#ifdef	notdef
#define	free(a)		task_mem_free((task *) 0, a)
#define	malloc(a)	task_mem_malloc((task *) 0, a)
#define	calloc(a, b)	task_mem_calloc((task *) 0, a, b)
#define	realloc(a, b)	task_mem_realloc((task *) 0, a, b)
#define	strdup(a)	task_mem_strdup((task *) 0, a)
#endif	/* notdef */

char parse_error[LINE_MAX] = { 0 };
char *parse_filename = 0;

static	proto_t	parse_proto;			/* For correctly tagging ADV lists */
static	proto_t	parse_export_proto;		/* For checking metrics */
static	gw_entry	**parse_gwlist;
static	gw_entry	*parse_gwp;		/* To fake a list for EGP and BGP */
static	int	parse_contig_masks = TRUE;	/* Masks must be contiguous */
static	bgpPeer *bnp;
static	bgpPeerGroup *bgp;
static 	struct bgp_conf *bcp;
static struct AREA *ospf_AREA;
static size_t ospf_auth_length;
static ospf_time_range ospf_md5_generate;
static ospf_time_range ospf_md5_accept;
static ospf_auth *ospf_md5_key_list;
static if_info parse_ifi;	/* Used for configuring interfaces that do not exist */

u_int parse_state = PS_INITIAL;
rtbit_mask protos_seen = 0;

PROTOTYPE(yyerror,
	  static void,
	  (const char *));

#define	free_string(s)	task_mem_free((task *) 0, (caddr_t) s); s = 0
#define	free_charstr(s)	task_mem_free((task *) 0, (caddr_t) (s).ptr); (s).ptr = (char *) 0; (s).len = (s).type = 0
#define	free_bytestr(s)	task_mem_free((task *) 0, (caddr_t) (s).ptr); (s).ptr = (byte *) 0; (s).len = (s).type = 0

#define	PARSE_ERROR	yyerror(parse_error); yynerrs++; YYERROR;

#define	PROTO_SEEN(proto) \
	if (BIT_TEST(protos_seen, RTPROTO_BIT(proto))) { \
	    sprintf(parse_error, "parse_proto_seen: duplicate %s clause", gd_lower(trace_state(rt_proto_bits, proto))); \
	    PARSE_ERROR; \
	} else { \
	    BIT_SET(protos_seen, RTPROTO_BIT(proto)); \
	    parse_proto = proto; \
	} \

#line 140 "parser.y"
typedef union {
    long	num;
    u_long	u_num;
    bytestr	hex;
    charstr	str;
    flag_t	flag;
    time_t	time;
    utime_t	utime;
    as_t	as;
    proto_t	proto;
    pmet_t	metric;
    pref_t	pref;
    if_addr	*ifap;
    if_addr_entry	*ifae;
    adv_entry	*adv;
    gw_entry	*gwp;
    void_t	vt;
    sockaddr_un	*sockaddr;
    trace_file	*trace_file;
#ifdef	PROTO_INET
    struct in_addr in_addr;
#endif	/* PROTO_INET */
    dest_mask	dm;
    struct {
	pmet_t metric;
	adv_entry *adv;
    } metric_adv ;
    struct {
	proto_t proto;
	adv_entry **import_list;
	adv_entry **export_list;
	gw_entry **gw_list;
    } proto_lists;
    config_entry *config;
    config_list *conflist;
#ifndef	GATED_LEXER
#ifdef	PROTO_OSPF
    ospf_config_router *ospf_router;
    ospf_auth *ospf_auth;
#endif	/* PROTO_OSPF */
#ifdef	PROTO_ASPATHS
    as_path *aspath;
    asmatch_t *asp_match;
    asp_range asp_range;
    asp_stack asp_stack;
#endif	/* PROTO_ASPATHS */
#endif	/* GATED_LEXER */
} YYSTYPE;
#line 190 "y.tab.c"
#define L_EOS 257
#define L_ERROR 258
#define L_NUMBER 259
#define L_STRING 260
#define L_HNAME 261
#define L_KEYWORD 262
#define L_HEXSTRING 263
#define L_OCTETSTRING 264
#define T_DIRECT 265
#define T_INTERFACE 266
#define T_INTERFACES 267
#define T_LCLADDR 268
#define T_PROTO 269
#define T_METRIC 270
#define T_METRICIN 271
#define T_METRICOUT 272
#define T_INFINITY 273
#define T_DOWN 274
#define T_PARSE 275
#define T_ADV 276
#define T_DEFAULT 277
#define T_DEFAULTS 278
#define T_YYDEBUG 279
#define T_YYSTATE 280
#define T_YYQUIT 281
#define T_DEBUG 282
#define T_SYSLOG 283
#define T_UPTO 284
#define T_EMERG 285
#define T_ALERT 286
#define T_CRIT 287
#define T_ERR 288
#define T_WARNING 289
#define T_NOTICE 290
#define T_INFO 291
#define T_ON 292
#define T_OFF 293
#define T_GATEWAY 294
#define T_PREFERENCE 295
#define T_PREFERENCE2 296
#define T_DEFAULTMETRIC 297
#define T_NONBROADCAST 298
#define T_BROADCAST 299
#define T_TYPE 300
#define T_POINTOPOINT 301
#define T_MULTICAST 302
#define T_NETMASK 303
#define T_DEFINE 304
#define T_INTERNAL 305
#define T_EXTERNAL 306
#define T_IGNORE 307
#define T_NOP 308
#define T_NOP1 309
#define T_NOP_DM 310
#define T_NOP_ADV 311
#define T_NOP_PROTO 312
#define T_NOP_SOCKADDR 313
#define T_PEERAS 314
#define T_LOCALAS 315
#define T_NEIGHBOR 316
#define T_GROUP 317
#define T_NOGENDEFAULT 318
#define T_VERSION 319
#define T_ACQUIRE 320
#define T_TTL 321
#define T_BGP 322
#define T_HOLDTIME 323
#define T_TEST 324
#define T_KEEP 325
#define T_SENDBUF 326
#define T_RECVBUF 327
#define T_OPEN 328
#define T_KEEPALIVE 329
#define T_INDELAY 330
#define T_OUTDELAY 331
#define T_ROUTING 332
#define T_ANALRETENTIVE 333
#define T_KEEPALIVESALWAYS 334
#define T_NOAGGRID 335
#define T_NOAUTHCHECK 336
#define T_SETPREF 337
#define T_IGNOREFIRSTASHOP 338
#define T_NOV4ASLOOP 339
#define T_V3ASLOOPOKAY 340
#define T_LOGUPDOWN 341
#define T_EGP 342
#define T_ASPATH 343
#define T_ORIGIN 344
#define T_INCOMPLETE 345
#define T_ANY 346
#define T_LOOPS 347
#define T_TAG 348
#define T_KERNEL 349
#define T_SYMBOLS 350
#define T_REMNANTS 351
#define T_IFLIST 352
#define T_ROUTES 353
#define T_NOCHANGE 354
#define T_NOFLUSHATEXIT 355
#define T_REMNANTHOLDTIME 356
#define T_HIGHER 357
#define T_LOWER 358
#define T_FLASH 359
#define T_INTERIOR 360
#define T_BACKGROUND 361
#define T_LIMIT 362
#define T_TRUSTEDGATEWAYS 363
#define T_SOURCEGATEWAYS 364
#define T_RIP 365
#define T_RIPOUT 366
#define T_NORIPOUT 367
#define T_RIPIN 368
#define T_NORIPIN 369
#define T_NOCHECKZERO 370
#define T_HELLOOUT 371
#define T_HELLOIN 372
#define T_NOHELLOOUT 373
#define T_NOHELLOIN 374
#define T_OSPF 375
#define T_OSPF_ASE 376
#define T_BACKBONE 377
#define T_VIRTUALLINK 378
#define T_STUB 379
#define T_STUBHOSTS 380
#define T_RXMITINTERVAL 381
#define T_INFTRANSDELAY 382
#define T_NETWORKS 383
#define T_POLLINTERVAL 384
#define T_ROUTERS 385
#define T_NEIGHBORID 386
#define T_TRANSITAREA 387
#define T_ELIGIBLE 388
#define T_MONITORAUTH 389
#define T_MONITORAUTHKEY 390
#define T_NOMULTICAST 391
#define T_DD 392
#define T_ACK 393
#define T_LSA_BLD 394
#define T_LSA_RX 395
#define T_LSA_TX 396
#define T_EXPORTINTERVAL 397
#define T_EXPORTLIMIT 398
#define T_FIRST 399
#define T_EVERY 400
#define T_START_ACCEPT 401
#define T_STOP_ACCEPT 402
#define T_START_GENERATE 403
#define T_STOP_GENERATE 404
#define T_KEY 405
#define T_ID 406
#define T_AREA 407
#define T_SPF 408
#define T_HELLO 409
#define T_HELLOINTERVAL 410
#define T_ROUTERDEADINTERVAL 411
#define T_PRIORITY 412
#define T_ENABLE 413
#define T_DISABLE 414
#define T_QUERY 415
#define T_QUERYINTERVAL 416
#define T_MODE 417
#define T_ADVERTISE 418
#define T_REGISTER 419
#define T_SNMP 420
#define T_ICMP 421
#define T_ROUTERDISCOVERY 422
#define T_REDIRECT 423
#define T_REDIRECTS 424
#define T_NOREDIRECTS 425
#define T_AUTHTYPE 426
#define T_AUTHKEY 427
#define T_NONE 428
#define T_SIMPLE 429
#define T_AUTH 430
#define T_MD5 431
#define T_SECONDARY 432
#define T_PASSIVE 433
#define T_SIMPLEX 434
#define T_STRICTIFS 435
#define T_SCANINTERVAL 436
#define T_STATIC 437
#define T_MARTIANS 438
#define T_RETAIN 439
#define T_EXPORT 440
#define T_IMPORT 441
#define T_EXACT 442
#define T_REFINE 443
#define T_HOST 444
#define T_MASK 445
#define T_MASKLEN 446
#define T_REJECT 447
#define T_BLACKHOLE 448
#define T_RESTRICT 449
#define T_ALLOW 450
#define T_ANNOUNCE 451
#define T_LISTEN 452
#define T_AS 453
#define T_ROUTERID 454
#define T_IGP 455
#define T_OPTIONS 456
#define T_NOINSTALL 457
#define T_NOSEND 458
#define T_NORESOLV 459
#define T_MARK 460
#define T_TRACEOPTIONS 461
#define T_EXCEPT 462
#define T_REPLACE 463
#define T_SIZE 464
#define T_FILES 465
#define T_K 466
#define T_M 467
#define T_ALL 468
#define T_ROUTE 469
#define T_STATE 470
#define T_POLICY 471
#define T_NORMAL 472
#define T_GENERAL 473
#define T_TASK 474
#define T_TIMER 475
#define T_NOSTAMP 476
#define T_UPDATE 477
#define T_DETAIL 478
#define T_SEND 479
#define T_RECEIVE 480
#define T_ERROR 481
#define T_PACKETS 482
#define T_RESPONSE 483
#define T_REQUEST 484
#define T_OTHER 485
#define T_TRAP 486
#define L_USEC 487
#define T_AGGREGATE 488
#define T_GENERATE 489
#define T_BRIEF 490
#define T_INET 491
#define T_GENDEFAULT 492
#define YYERRCODE 256
short yylhs[] = {                                        -1,
    0,    0,  191,  191,  192,  192,  192,  192,  192,  192,
  192,  192,  192,  192,  193,  193,  193,  194,  124,  124,
  124,  125,  125,  125,  128,  128,  123,  123,  122,  118,
  118,  118,  119,  119,  119,  119,  117,  117,  116,  116,
  116,  116,  116,  116,  116,  116,  120,  120,  121,  121,
  121,  199,  200,  207,  200,  200,  200,  163,  163,  163,
  162,  162,  162,  178,  178,  195,  196,  208,  208,  209,
  209,  209,  209,  209,  164,  210,  164,    3,    3,    2,
    2,    2,    1,    1,    1,    1,    1,    1,    1,    1,
  197,  211,  198,  212,  212,  212,  213,  213,  216,  213,
  214,  214,  218,  218,  215,  215,  215,  217,  217,  219,
  219,  126,  126,  127,  127,  127,  127,  127,  127,  127,
  174,  174,  176,  176,  175,  175,  173,  173,  171,  172,
  172,  145,  146,  146,  201,  202,  202,  202,  202,  202,
  202,  202,  202,  228,  221,  229,  229,  229,  230,  230,
  230,  230,  230,  230,  231,  231,  234,  234,  232,  232,
  235,  235,   36,   36,   36,  233,  233,  236,  236,   37,
   37,   37,   33,   33,   33,   34,   34,   35,   35,   35,
   35,   35,   35,   35,   32,   32,   32,   32,   32,  237,
  226,  238,  238,  239,  239,  239,  240,  240,  240,  240,
  240,  240,  240,  240,  240,  240,   39,   39,   38,   38,
   38,   38,   38,   38,   38,   38,   38,   45,   45,   45,
   47,   47,   41,   41,   41,   42,   42,   42,   43,   43,
   44,   44,   40,   40,   40,   40,   46,  241,  223,  242,
  242,  243,  243,  243,  244,  244,  244,  244,  244,  244,
  244,  244,   49,   49,   48,   48,   48,   48,   48,   48,
   51,   51,   51,   52,   52,   53,   53,   50,   54,  245,
  225,  246,  246,  247,  247,  247,  248,  248,  248,  248,
  248,  250,  248,  251,  248,  248,  252,  248,   77,   77,
   78,   78,  249,  249,  249,  254,  254,  254,  254,  253,
  253,  253,  255,  255,  255,  255,  255,  255,   55,   55,
   56,   56,   58,   58,   58,   57,   57,   60,   60,   60,
   59,   59,   59,   59,   59,   59,   59,   59,   59,  257,
   59,   59,   59,   61,   61,   61,   62,   62,   63,   64,
   64,   65,   76,   76,   75,   68,   68,   69,   69,  258,
   70,   67,   67,  259,  259,  260,  260,  260,  260,   66,
  261,   72,   74,   74,   73,   73,   73,   71,   71,   71,
  256,  256,  256,  262,   82,   82,   82,   81,   80,   80,
   79,   87,   88,   83,   83,   83,   84,   84,   85,   85,
   85,   85,   85,   85,   85,   85,   86,   86,   86,   86,
   86,   86,   92,   92,   92,   91,   91,   90,   90,  263,
  222,  264,  264,  265,  265,  265,  266,  266,  266,  266,
  268,  269,  266,  267,  267,  270,  270,  270,  273,  272,
  274,  272,  271,  271,  271,  271,  271,  271,  271,  271,
  271,  271,  271,  271,  271,  271,  271,  271,  271,  271,
  271,  271,  271,  271,  271,  271,  271,  271,  271,    6,
    5,    5,    5,    5,    5,    7,    7,    7,    8,    8,
   12,   12,   13,   13,   13,   13,   16,   16,    9,    9,
    9,   10,   10,   10,   10,   11,   15,   15,   15,   14,
   14,   14,  275,  224,  276,  276,  276,  277,   94,   94,
   94,   95,   95,   96,   96,   93,   93,   93,   93,   93,
  278,  220,  279,  279,  280,  280,  280,  281,  281,  281,
  281,   98,   98,   97,   97,   99,   99,   99,  100,  100,
  101,  227,  203,  282,  204,  283,  283,  283,  284,  284,
  103,  103,  102,  102,  102,  102,  104,  104,  105,  105,
  105,  105,  205,  206,  206,  206,  206,  206,  206,  206,
  206,  206,  206,  134,  134,  161,  161,  167,  167,  167,
  168,  168,  168,  168,  169,  169,  169,  169,  169,  169,
  169,  169,  169,  169,  130,  106,  106,  131,  181,  181,
  179,  179,  179,  285,  180,  132,   89,  107,  107,  187,
  187,  183,  183,  183,  182,  182,  182,  182,  182,  182,
  182,  182,  182,  182,  182,  182,  136,  137,  139,  138,
  135,  188,  188,  186,  186,  185,  185,  185,  286,  184,
  155,  155,  155,  155,  159,  160,  166,  166,  165,  152,
  152,  156,  153,  153,  157,  157,  157,  158,  154,  151,
  150,  150,    4,  148,  148,  149,  149,  147,  147,  177,
  177,  170,  170,  170,  190,  190,  190,  133,  133,  133,
  129,  129,  115,  115,  111,  111,  112,  112,  113,  113,
  143,  143,  144,  144,  142,  189,  108,  108,  109,  109,
  110,  110,  114,  114,  114,  141,  141,  140,   29,   29,
   28,   28,   28,   27,   27,   30,   31,   31,   31,   31,
   31,   26,   26,   26,   26,   26,   26,  287,   17,   25,
   25,   25,   23,   23,   24,   24,   24,   18,   18,   19,
   19,   21,   21,   21,   21,   21,   20,   20,   20,   22,
   22,   22,
};
short yylen[] = {                                         2,
    0,    1,    1,    2,    1,    1,    2,    2,    2,    2,
    2,    2,    2,    1,    3,    3,    2,    5,    0,    2,
    6,    1,    2,    2,    0,    1,    0,    2,    1,    1,
    1,    3,    1,    1,    2,    2,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    0,    1,    0,    1,
    1,    0,    1,    0,    6,    4,    3,    0,    4,    3,
    1,    1,    1,    0,    1,    0,    3,    1,    2,    1,
    1,    2,    2,    3,    0,    0,    3,    0,    1,    1,
    2,    2,    1,    1,    1,    1,    1,    1,    1,    1,
    0,    0,    6,    0,    3,    3,    2,    3,    0,    5,
    1,    2,    1,    2,    2,    2,    0,    0,    2,    2,
    1,    1,    2,    2,    3,    1,    1,    1,    1,    2,
    0,    2,    1,    2,    1,    1,    1,    2,    1,    1,
    1,    1,    1,    1,    0,    2,    2,    2,    2,    2,
    2,    2,    2,    0,    5,    0,    3,    3,    4,    2,
    2,    2,    2,    2,    1,    2,    1,    1,    1,    2,
    2,    2,    1,    1,    1,    1,    2,    2,    2,    1,
    1,    1,    1,    1,    3,    1,    2,    1,    3,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    0,
    4,    0,    3,    0,    3,    3,    2,    2,    1,    1,
    4,    1,    3,    2,    2,    3,    1,    2,    1,    1,
    1,    1,    2,    2,    2,    3,    3,    0,    1,    1,
    1,    2,    0,    1,    1,    1,    1,    3,    1,    2,
    1,    3,    1,    1,    1,    1,    1,    0,    4,    0,
    3,    0,    3,    3,    2,    2,    1,    1,    4,    3,
    2,    2,    1,    2,    1,    1,    1,    1,    2,    2,
    1,    1,    3,    1,    2,    1,    3,    1,    1,    0,
    4,    0,    3,    0,    3,    3,    4,    4,    2,    2,
    3,    0,    4,    0,    3,    3,    0,    5,    0,    2,
    0,    2,    0,    3,    3,    2,    2,    2,    2,    0,
    3,    3,    2,    2,    4,    4,    2,    6,    3,    4,
    0,    3,    2,    3,    3,    0,    3,    2,    3,    3,
    1,    1,    2,    2,    2,    2,    2,    1,    3,    0,
    4,    3,    2,    1,    2,    4,    0,    2,    1,    0,
    2,    1,    2,    1,    1,    1,    3,    2,    3,    0,
    6,    0,    3,    2,    3,    2,    2,    2,    2,    8,
    0,    3,    1,    1,    1,    1,    1,    1,    1,    1,
    2,    3,    3,    3,    2,    3,    3,    2,    0,    1,
    1,    1,    1,    1,    1,    3,    1,    2,    1,    3,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    0,    4,    3,    1,    1,    0,    1,    0,
    4,    0,    3,    0,    3,    3,    2,    2,    2,    4,
    0,    0,   13,    0,    2,    0,    3,    3,    0,    4,
    0,    5,    2,    2,    1,    2,    2,    2,    2,    2,
    4,    2,    1,    2,    2,    1,    1,    1,    1,    2,
    1,    2,    2,    2,    2,    2,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    3,    1,    2,
    0,    2,    1,    1,    1,    1,    0,    2,    1,    1,
    3,    1,    1,    1,    1,    1,    0,    3,    3,    1,
    1,    1,    0,    5,    0,    3,    3,    4,    1,    1,
    3,    1,    2,    1,    3,    1,    1,    1,    1,    1,
    0,    4,    0,    3,    0,    3,    3,    2,    3,    2,
    4,    1,    2,    1,    1,    1,    1,    3,    1,    2,
    1,    2,    0,    0,    5,    2,    3,    2,    5,    7,
    0,    2,    1,    1,    1,    1,    0,    2,    1,    1,
    1,    1,    0,    1,    7,    6,    6,    6,    7,    6,
    7,    8,    7,    0,    1,    1,    1,    1,    2,    3,
    4,    6,    5,    6,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    0,    1,    1,    1,    4,
    0,    3,    3,    0,    4,    1,    1,    0,    1,    1,
    4,    0,    3,    3,    4,    4,    3,    4,    3,    5,
    4,    5,    4,    4,    3,    3,    1,    1,    1,    1,
    1,    1,    2,    0,    3,    0,    3,    3,    0,    4,
    1,    1,    1,    1,    1,    1,    1,    2,    1,    1,
    1,    2,    1,    1,    1,    3,    3,    2,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    0,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    0,    2,    1,    1,
    0,    2,    1,    1,    1,    1,    1,    2,    1,    1,
    1,    1,    1,    3,    5,    0,    2,    1,    0,    2,
    1,    1,    2,    0,    2,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    0,    5,    0,
    1,    1,    1,    3,    1,    1,    1,    1,    3,    1,
    2,    1,    2,    2,    2,    4,    1,    1,    3,    1,
    2,    3,
};
short yydefred[] = {                                      0,
    0,   14,    0,    0,    0,    0,    0,    0,    3,    5,
    6,    0,    0,    0,    0,    0,    0,   13,  673,  674,
    0,    0,   17,  686,   27,    0,    4,    0,    7,   92,
    8,   53,   54,    0,    0,    9,  410,  144,  190,  270,
  238,    0,  493,  511,   10,    0,    0,    0,    0,    0,
    0,    0,    0,  534,   11,  554,    0,    0,    0,    0,
   12,   15,   16,    0,   26,    0,    0,   70,   71,    0,
    0,    0,   68,    0,    0,  698,    0,  651,  665,  666,
  691,  692,  652,  650,  641,    0,  667,  640,    0,    0,
    0,    0,    0,  532,    0,    0,  136,  137,  138,  139,
  140,  141,  142,  143,    0,    0,    0,  636,  644,    0,
  566,  567,    0,  643,    0,   37,   38,   30,   46,   39,
   42,   43,   40,   41,   44,   45,   29,   33,   34,    0,
    0,   28,    0,   90,    0,   83,   84,   85,   86,   87,
   88,   89,   80,    0,    0,   73,   79,    0,    0,   67,
   69,   94,   58,    0,    0,   57,    0,  146,    0,    0,
    0,  495,    0,    0,  671,  486,  670,  597,  669,  668,
  672,    0,  585,    0,    0,  596,    0,    0,    0,    0,
  588,    0,    0,    0,    0,   18,    0,   35,   36,  689,
    0,  690,    0,   81,   82,    0,  685,  682,   76,   74,
    0,    0,  697,   56,  414,  411,    0,  194,  191,  274,
  271,  242,  239,    0,  515,  512,    0,  546,  543,    0,
    0,  544,  545,    0,    0,    0,    0,    0,    0,    0,
  123,    0,  718,    0,    0,    0,    0,    0,  649,  647,
  646,  565,    0,    0,    0,   23,   24,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,   61,    0,
   62,   63,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  145,    0,    0,    0,    0,    0,    0,  494,    0,
    0,  538,  642,    0,    0,  535,    0,  536,  342,  341,
    0,    0,    0,  658,  131,  659,  127,  129,    0,  130,
  660,  639,  637,    0,  124,    0,  600,    0,    0,    0,
  589,    0,    0,  705,  706,    0,    0,    0,    0,    0,
    0,  568,   21,    0,   77,   96,  125,    0,    0,   99,
  103,    0,    0,  101,   93,   95,   60,   55,  663,  664,
  648,   65,    0,    0,    0,    0,    0,    0,    0,  413,
    0,  148,  154,  151,    0,    0,    0,  159,    0,    0,
    0,  166,  157,  158,    0,  155,   27,  147,    0,    0,
    0,    0,  200,  199,    0,    0,  202,    0,    0,  193,
    0,    0,    0,    0,  344,    0,  284,    0,    0,    0,
    0,  273,  287,    0,    0,    0,    0,    0,  248,  247,
    0,    0,    0,  241,    0,  497,   27,  496,    0,    0,
    0,    0,    0,  514,    0,    0,    0,  537,    0,  700,
  701,    0,    0,  128,  638,  676,  675,  678,  602,  560,
  738,    0,    0,    0,    0,  730,  737,  591,  556,  558,
    0,  557,    0,  576,  584,  578,  581,  582,  583,  580,
  577,  575,  579,    0,    0,    0,  569,  695,    0,    0,
  116,  117,  118,  119,    0,    0,  112,    0,  104,  102,
   59,  416,  417,  418,  419,    0,   27,  415,  163,  164,
  165,  161,  687,    0,  162,  160,  169,  171,  172,  170,
  168,  167,  156,    0,  196,    0,  197,  198,    0,    0,
    0,   27,  195,  276,  293,    0,    0,  363,  364,    0,
  282,    0,  279,  280,  343,  345,   27,    0,  275,  244,
    0,  245,  246,    0,    0,   27,  243,    0,  517,    0,
  518,  520,   27,  516,  547,    0,  703,  561,  559,    0,
    0,    0,    0,  731,    0,  733,  734,  735,    0,  555,
    0,    0,    0,    0,    0,  570,  563,    0,  114,  120,
  113,    0,    0,  108,  461,  462,  465,  464,  463,    0,
    0,  688,  184,  180,  181,  182,  173,   48,  183,  149,
    0,  176,  178,    0,    0,    0,    0,  211,  212,  209,
  210,    0,    0,  207,    0,  221,  224,  225,    0,  206,
    0,    0,  290,    0,  286,  350,    0,  281,  346,    0,
  369,  285,  370,  368,    0,  300,    0,    0,  257,  255,
  258,  256,  253,    0,    0,  499,  498,    0,  502,  504,
    0,  524,  525,  522,    0,    0,    0,    0,    0,    0,
    0,  601,    0,  739,  726,  727,  721,  725,    0,  723,
  719,    0,    0,    0,    0,  587,  590,  594,    0,  562,
    0,    0,    0,  571,  115,  654,  105,  655,  106,    0,
    0,  480,  466,  420,    0,  469,  479,    0,    0,  177,
   50,   51,    0,  213,  214,    0,  215,    0,  208,  222,
  226,  201,    0,  229,  231,    0,    0,    0,    0,    0,
    0,  277,    0,  292,    0,    0,    0,  283,  396,  391,
  393,  392,  394,  384,  395,  278,    0,  387,  389,    0,
    0,  259,  260,  254,  261,  249,    0,  264,  266,    0,
    0,  503,    0,  523,  526,  521,    0,  529,  531,  549,
  550,  551,  552,  548,  542,  547,  604,  617,  621,  620,
  237,  382,  383,  269,  618,  712,  619,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  603,
    0,  736,    0,  593,    0,  592,    0,    0,  573,  111,
    0,  109,    0,    0,  470,    0,    0,  188,  186,  187,
  185,  189,  179,  220,  219,  217,  216,    0,  230,    0,
  295,  297,  339,  296,  299,  298,  294,    0,  347,    0,
  348,    0,  388,    0,    0,    0,    0,    0,    0,    0,
    0,  288,    0,    0,  265,    0,    0,  509,  508,  507,
  510,  506,  505,    0,  530,    0,    0,    0,    0,    0,
  622,    0,  615,  616,    0,  609,    0,    0,    0,  607,
  724,  742,  635,    0,  633,  634,  631,  632,  574,  572,
  656,  110,  657,    0,    0,    0,  483,  485,  484,  482,
  481,    0,  233,  235,  234,  236,  232,    0,  349,    0,
  399,  402,  398,  401,  397,  400,  390,  302,  307,    0,
    0,    0,  304,    0,  403,  365,  366,  367,  303,  301,
    0,  268,  267,    0,  611,    0,  613,  614,  626,  623,
    0,  605,  606,  608,  684,  683,  595,  473,  474,  476,
  475,  472,    0,  421,    0,    0,    0,    0,  338,    0,
    0,    0,    0,  612,    0,  610,  478,  424,    0,  351,
    0,    0,  309,    0,    0,    0,  305,    0,  371,    0,
  306,    0,  406,  407,    0,  599,  625,  629,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  310,    0,    0,
  328,    0,    0,    0,  321,  322,    0,    0,    0,    0,
    0,    0,  374,  373,  372,  405,  409,    0,  628,    0,
  627,    0,    0,    0,    0,    0,    0,  435,    0,    0,
    0,    0,    0,    0,    0,    0,  446,  447,  448,  449,
    0,  451,  457,  458,  459,  443,    0,    0,  425,    0,
  356,  357,  358,  359,  353,    0,  354,    0,    0,  334,
    0,    0,  323,  324,  326,  327,  381,  325,  333,    0,
  330,    0,    0,  312,    0,  313,  308,  404,    0,  439,
  132,  133,  134,  433,  436,  437,  438,  434,  442,  653,
  456,  440,  445,  444,  452,  460,  453,  454,  455,  450,
   27,  426,    0,  355,  335,    0,    0,  317,    0,  318,
  329,    0,  332,  361,  315,  314,  680,  679,  630,    0,
    0,    0,    0,    0,    0,  320,  319,  331,    0,  441,
    0,  429,  431,  423,    0,    0,  375,    0,  336,    0,
  380,  378,  362,  428,    0,    0,  427,    0,  377,  376,
  424,  487,    0,    0,    0,    0,    0,  432,    0,  491,
  492,  490,  360,  489,  488,
};
short yydgoto[] = {                                       7,
  143,  144,  146, 1061,  570, 1065,  674,  675,  676,  871,
  171,  865,  922, 1129, 1125,  924,  234,  433,  434,  435,
  436,  653,  649,  650,  651,  758,  236,  420,  292,  421,
  759,  793,  580,  581,  582,  482,  491,  594,  595,  877,
  599,  692,  693,  694,  796,  760,  600,  623,  624,  903,
  726,  727,  728,  761,  889,  943,  968,  980, 1030, 1031,
 1032,  893,  802,  227,  290, 1021,  940,  608,  706,  609,
  612, 1083,  899,  511,  515,  393,  507,  605, 1038, 1112,
 1093, 1094,  716,  717,  718,  887,  762,  763,  172,  988,
  952,  933,  833,  627,  628,  629,  634,  635,  736,  737,
  738,  221,  639,  637,  744,  658,  958,  485,  427,   83,
  803,  842, 1089,  147,   21,  677,  129,  130,  131,  678,
  683,  132,   64,   25,  193,  466,  467,   66,  173,  174,
  180,  175,  176,  243,  765,  766,  767,  768,  769,  437,
  155,  198,  312,  917, 1050, 1051,  295,  930,  862,   84,
   85,  302,  110,  240,  854,  222,  260,  856,  857,  112,
  113,  263,  202,  200,  303,  304,  321,  322,  454,  341,
  297,  298,  328,  231,  329,  232,  305,  343,  549,  659,
  313,  643,  540,  959,  935,  910,  309,  843,   87,   88,
    8,    9,   10,   11,   12,   29,   13,   31,   14,   36,
   15,   45,   16,   55,   17,   61,   75,   72,   73,  250,
   74,  201,  256,  333,  564,  468,  670,  334,  782,   46,
   47,   48,   49,   50,   51,   52,   53,   90,  207,  273,
  365,  357,  361,  366,  358,  362,   91,  209,  274,  381,
   93,  213,  276,  405,   92,  211,  275,  394,  602,  610,
  512,  518,  721,  703,  823,  931, 1082,  705,  965,  966,
 1099,  932,   89,  206,  264,  351,  960,  938, 1018, 1091,
 1019, 1105, 1115, 1116,   95,  214,  280,   96,  216,  281,
  415,  105,  224,  225,  775,  990,  310,
};
short yysindex[] = {                                   -162,
 -137,    0, -130, -131,  -28,  -66,    0, -162,    0,    0,
    0, -169,   22,  216, 1936, -110,  196,    0,    0,    0,
   72,   88,    0,    0,    0, -105,    0,  279,    0,    0,
    0,    0,    0,  120, 1488,    0,    0,    0,    0,    0,
    0,  167,    0,    0,    0,  208,  232,  239,  274,  296,
  304,  308,  327,    0,    0,    0,  294,  320, 1436, 1436,
    0,    0,    0, 1260,    0,   85, 1636,    0,    0,  335,
  312,  349,    0,  476,  492,    0,  278,    0,    0,    0,
    0,    0,    0,    0,    0,  376,    0,    0, -130,  512,
 -130, -130, -130,    0,  541, -130,    0,    0,    0,    0,
    0,    0,    0,    0,  564,  640, 1908,    0,    0,  -32,
    0,    0,  312,    0,  312,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  455,
  794,    0,  416,    0, 1589,    0,    0,    0,    0,    0,
    0,    0,    0, 1589,  656,    0,    0,  468,  467,    0,
    0,    0,    0,  471,  506,    0,  643,    0,  645,  665,
  667,    0,  670, 1783,    0,    0,    0,    0,    0,    0,
    0,  495,    0,  361, -117,    0,  487,  484,  381, -117,
    0,  625,  416,  359,  735,    0, 1196,    0,    0,    0,
  106,    0,  396,    0,    0,  604,    0,    0,    0,    0,
  -48,  434,    0,    0,    0,    0,  -21,    0,    0,    0,
    0,    0,    0,  -95,    0,    0,  616,    0,    0, 1488,
  140,    0,    0,  514,  626,  641,  557,  120, 1078,  838,
    0, -227,    0, -242,  416, -242,  120, -242,    0,    0,
    0,    0,  784,  642, 1196,    0,    0,  416,  856, 1488,
  675,  925, 1488,  183,  677,  682,  684,  697,    0,  266,
    0,    0,  466,  -63,  710,  416,  335, -141, -208,  302,
  -66,    0,  712,  177,  153,  -16,  716,  -66,    0,  717,
  -75,    0,    0, 1078,  838,    0,  730,    0,    0,    0,
 -163, -227, -227,    0,    0,    0,    0,    0, 1078,    0,
    0,    0,    0, 1488,    0,  249,    0,  801,  733,   93,
    0,  863,  740,    0,    0,  743, -242,  749,  642, 1184,
  245,    0,    0,  753,    0,    0,    0, 1078, 1193,    0,
    0,  335,  183,    0,    0,    0,    0,    0,    0,    0,
    0,    0,  760,  765,  468,  468,  249,  728,  -66,    0,
  772,    0,    0,    0,   14,   -7, -141,    0,  416,   68,
 -208,    0,    0,    0,  302,    0,    0,    0,  775,  925,
  468,  249,    0,    0, 1488, 1488,    0,  612,  -66,    0,
  790,  795,  931,  657,    0,  260,    0,  335,  416,  416,
  -66,    0,    0,  802,  803,  925,  468,  249,    0,    0,
 1488, 1488,  -66,    0,  805,    0,    0,    0,  816,  925,
  468,  838,  -66,    0,  818,  312,  792,    0,  416,    0,
    0,  821,  822,    0,    0,    0,    0,    0,    0,    0,
    0,   93,  -85,   93,   73,    0,    0,    0,    0,    0,
  823,    0,  390,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  175,  825,  829,    0,    0,  796,  468,
    0,    0,    0,    0,  120, 1193,    0, -161,    0,    0,
    0,    0,    0,    0,    0,  161,    0,    0,    0,    0,
    0,    0,    0,  831,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1681,    0, 1562,    0,    0, 1488, 1488,
  352,    0,    0,    0,    0,  416,  695,    0,    0,  -64,
    0,  729,    0,    0,    0,    0,    0,  986,    0,    0,
  846,    0,    0, 1488, 1488,    0,    0, 1983,    0,  401,
    0,    0,    0,    0,    0,  658,    0,    0,    0,  454,
   52,  458,   93,    0,  416,    0,    0,    0,  -91,    0,
  864,  416,  120, -242,  874,    0,    0,  468,    0,    0,
    0, 1099, 1099,    0,    0,    0,    0,    0,    0,  799,
 1796,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1698,    0,    0,  337,  249,  249,  878,    0,    0,    0,
    0,  352,  704,    0, 1562,    0,    0,    0,  -66,    0,
 1999,  592,    0,  416,    0,    0,  734,    0,    0,  729,
    0,    0,    0,    0, 1583,    0,  249,  249,    0,    0,
    0,    0,    0,  846, 2010,    0,    0, 1512,    0,    0,
  337,    0,    0,    0,  401, 1807,  363,  120,  312,  881,
 1226,    0,  884,    0,    0,    0,    0,    0, 1022,    0,
    0,   93, 1019, 1104,  895,    0,    0,    0,  897,    0,
 -242, -242,  898,    0,    0,    0,    0,    0,    0,  550,
  120,    0,    0,    0, 1546,    0,    0,  337, 1731,    0,
    0,    0, -182,    0,    0,  306,    0,  352,    0,    0,
    0,    0, 1769,    0,    0,  337,  900,  249,  468,  641,
 -163,    0,  906,    0,  729,  -77,  909,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1602,    0,    0,  337,
  460,    0,    0,    0,    0,    0, 2025,    0,    0,  337,
 2036,    0,  503,    0,    0,    0, 1944,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  487,  820, -117,
 -117, -196, -196,  721, -196,  792,  792,  792, -196,    0,
  258,    0,  416,    0,  980,    0,  920,  922,    0,    0,
  493,    0,  912, 1294,    0,   62, 1731,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 2036,    0,  411,
    0,    0,    0,    0,    0,    0,    0,  776,    0,  934,
    0, 1633,    0,  259,  938,  925,  811,  929, 1084, 1110,
  368,    0,  977, 2036,    0,  755, 2036,    0,    0,    0,
    0,    0,    0, 2047,    0,  363, -196,  416, -196, -196,
    0, 1112,    0,    0,  120,    0, -196, -196, -196,    0,
    0,    0,    0, -154,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  222,  979, 1294,    0,    0,    0,    0,
    0, 2036,    0,    0,    0,    0,    0,  987,    0, 1633,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  194,
 1099,  249,    0, 1099,    0,    0,    0,    0,    0,    0,
 2036,    0,    0, 2047,    0, -196,    0,    0,    0,    0,
 -196,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,  925,    0, 1128,  929, 1129,  867,    0,  985,
  967, 1004,  659,    0,  -71,    0,    0,    0,  663,    0,
 1147, 2018,    0,  416,  249, 1018,    0, 1027,    0, 1029,
    0,  828,    0,    0, 1031,    0,    0,    0, 1039, 1795,
 1048, 1048, 1048, 1048,   49, 1054, 2012,    0,  335,  335,
    0,  335,  335, 1055,    0,    0,  729,  436,  888,  292,
 1065, 1129,    0,    0,    0,    0,    0, 1077,    0,  980,
    0, 1078,  249, 1488,  468,  468,  120,    0, 1076, 1087,
  335, -345,  416,  416,  335,  335,    0,    0,    0,    0,
  468,    0,    0,    0,    0,    0,  -66, 1225,    0, 1302,
    0,    0,    0,    0,    0, 1093,    0,  335, 1228,    0,
 1020, 1096,    0,    0,    0,    0,    0,    0,    0,  -64,
    0,  532, 1108,    0, 1109,    0,    0,    0, -192,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1097,    0,    0, 1488, 1114,    0, 1116,    0,
    0,  729,    0,    0,    0,    0,    0,    0,    0, 1796,
  -69, 1330, 1121, 1394,  993,    0,    0,    0,  729,    0,
 1135,    0,    0,    0, 1137, 1136,    0, 1139,    0, 1141,
    0,    0,    0,    0, 1099, 1276,    0, 1144,    0,    0,
    0,    0, 1342, 1795,   59, 1148, 1149,    0, 1151,    0,
    0,    0,    0,    0,    0,
};
short yyrindex[] = {                                     79,
    0,    0,    0,    0,    0, 1356,    0,  189,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1331,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1427,    0,    0,    0,  450,
  163,    0,    0,    0,    0,    0, 1156,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  -78,
    0,    0,  -98,    0, 1291,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1162,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  489,  700,    0,    0,    0,  585,    0,
    0,    0,    0,    0,    0,    0, 1170,    0, 1172, 1185,
 1188,    0, 1194,    0,    0,    0,    0,    0,    0,    0,
    0,  -47,    0,    0,  -65,    0, 1001,  -86,    0,  -81,
    0,    0,    0, 1332,    0,    0,    0,    0,    0,    0,
  992,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,  -33,    0,    0,    0,
    0, 1336,    0, 1291,    0, 1291,    0, 1291,    0,    0,
    0,    0,    0,    0, 1204,    0,    0,    0,  789,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  191,
    0,    0, 1206,    0,    0,    0,    0,    0,    0,    0,
 1491,    0,    0,    0,    0,    0,    0, 1715,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1336, 1336,    0,    0,    0,    0,    0,  301,    0,
    0,    0,    0,  214,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1291,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1049,    0,    0,
    0,    0,  208,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1818,    0,
    0,    0,    0,    0,    0,    0, 1207,    0,    0,    0,
 1208,    0,    0,    0, 1213,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1892,    0,
    0,    0,    0, -188,    0,    0,    0,    0,    0,    0,
 1464,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1910,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1878,    0,    0,   98, 1171,    0,  -55,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    5,  251,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1291,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  232,    0,  139,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  830,    0,    0,    0,    0,  506, 1218,
 1216,    0,    0,    0,    0,    0, 1220,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1223, 1230,    0,    0,  522,    0,    0,
    0,    0,    0,    0,    0, 1021,    0,    0,    0,    0,
    0,  -26,    0,    0,    0,    0,    0,    0, 1176,    0,
    0,    0,    0, 1291,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
  685,    0,    0,    0,    0,    0,    0,    0,    0,    0,
  392,    0,    0,   84,    0,    0,    0,    0,    0,    0,
    0, 1216,    0,    0, 1232,    0,    0,    0,    0,    0,
 1132,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1825,    0,    0,    0,    0,    0,
    0,    0,    0, 1233,  365,    0,    0,  528,    0,    0,
  851,    0,    0,    0,  802,    0,  822,    0,   98,    0,
    0,    0,    0,    0,    0,    0,    0,    0,   41,    0,
    0,   10,    0, 1367,    0,    0,    0,    0,    0,    0,
 1291, 1291,    0,    0,    0,    0,    0,    0,    0,  274,
    0,    0,    0,    0,  875,    0,    0,   75,  830,    0,
    0,    0,    0,    0,    0, 1519,    0, 1216,    0,    0,
    0,    0, -175,    0,    0,  842,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,  571,    0,    0, 1017,
    0,    0,    0,    0,    0,    0, -149,    0,    0, 1011,
  522,    0,    0,    0,    0,    0,  805,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  -31,
  125,   57,   67, -197,    1,   -2,   -2,   -2,    1,    0,
    0,    0, 1370,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1190,  685,    0,    0,  671,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1132,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1825,    0,    0,    0,    0,    0,  684,    0,    0,
    0,    0,    0,  365,    0,    0,  849,    0,    0,    0,
    0,    0,    0,    0,    0, 1241,    1,    0,    1,    1,
    0, 1243,    0,    0,    0,    0,    1,    1,    1,    0,
    0,    0,    0, 1247,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1246,  964,    0,    0,    0,    0,
    0,  419,    0,    0,    0,    0,    0,    0,    0,  696,
    0,    0,    0,    0,    0,    0,    0,    0,    0,   27,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 -120,    0,    0, 1250,    0,    1,    0,    0,    0,    0,
    1,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,  710,   27, 1255,    0,    0,    0,
    0,    0,    0,    0, 1293,    0,    0,    0,    0,    0,
 1265,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  717,    0,    0,    0,    0,    0,    0,    0, 1409,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1255,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1818,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1277,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  685,
    0,    0,    0,    0, 1280,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  733,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,
};
short yygindex[] = {                                      0,
  286,    0,    0,    0,    0,  535,  453,  761, -604,    0,
    0,    0,    0,    0,    0,    0, -414, 1115, 1003,    0,
 -382,    0,    0,  773,    0,    0,    0,  850,    0, -218,
    0,    0,    0,  871, -524,    0,    0,  963,    0,    0,
    0,    0,  764, -608,    0,    0, -489,  939,    0,    0,
    0,  742, -628,    0,    0,  586,    0,    0, -811,    0,
  543, -471, -787,    0,  889, -485,    0,  538,    0, -546,
 -578,    0,    0, -890,  644,    0,    0,    0,    0,    0,
  496,    0,    0,  780, -642,    0,    0,    0,    0,    0,
    0,    0,    0,    0,  865, -584,  959,    0,    0,  766,
 -701,    0,    0,  852,    0,    0,    0,    0, -113, -123,
 -292, -220,    0, -241,  886,  -45,  280,    0, 1408,  119,
  155,    0, -320, -247,    0,    0, 1133,    0,  962, 1497,
    0,    0, 1498,    0,    0,    0,    0,    0,    0,  -27,
    0, -330,  -49,    0,    0,    0,  617, -534,    0, -164,
  -57,  -34,    0,  827,  635, -198,  -51, 1428,  504, -189,
 1573,    0,    0,    0, -239,  554, 1315, -232,    0,    0,
 -173,    0, 1407, -390, -347, -159, -187,    0,    0,    0,
 -203,    0,    0,    0,    0,    0,  673, 1412,   -6,  -54,
    0, 1630,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1567,    0,
    0,    0,    0,    0,    0,    0,    0, 1309,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1278, 1288, 1286,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  683,
    0,  718,    0,    0,    0,    0,  540,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1438,    0,    0,    0,
};
#define YYTABLESIZE 2522
short yytable[] = {                                      26,
   86,  109,  109,  259,  114,  114,   77,  111,  111,  192,
  325,  308,  262,  428,  473,  474,  314,  239,  128,  191,
  238,  149,  496,  367,  681,  354,  536,  667,  669,  279,
  407,  708,  316,  657,  318,  835,  704,  484,  543,  554,
  497,  121,  306,  732,  645,  728,  494,  809,  521,  414,
  729,  544,  148,  957,  475, 1104,  680,  121,  607,  192,
  707,  350,  530,  184,  425,  185,  522,  702,  289,  241,
  785,  308,  308,  306,  813,  340,  255,  306,    1,  498,
  531,  227, 1063,  788,  799,  188,  528, 1041,  457,  699,
  469,  121,  644,    1,    2,  190,  720,  417,  825,   81,
   82,  477,  687,  272,  929,  523,  109,  262,  404,  114,
  416,  192,  223,  441,  546,  547,    3,    4,    5,   18,
  121,  315, 1064,  677,  192,  424,  808,   22,  728,  559,
  981,  502,  432,  729,  323,  548,  263,  562,  431,  563,
  148,  128,  192,  517,  109,  713,  513,  114,  229,  337,
  707, 1084,  353,  359,  424,  526,  571,  983,  355,  810,
  277,   19,   20,  722,  655,  533,  109,  192, 1045,  114,
  789,  296,  223, 1025,  300,  543,  230,  315,  645,  677,
  409,  601,  192, 1128,  955,  283, 1101,  645,    2,  677,
  410,  645,  344,   24,  296,  545,  615,  300,  797,  188,
  293,  702,  835,  360,  121,  625,  311,  251,  704,  317,
  457,  289,  636,  121,  702,  645,  645,  252,  330,  411,
  356,  307,  340,  192,  532,  121,  296,  665,   23,  300,
  720,  345,  346,  347,  265,  192,  699,  813,  121,  395,
  790,  296,  732,  720,  300,  487, 1102,  121,  192,  396,
  555,  483,  841,  348,  121,  253, 1087,  677,  542,  425,
  425,  785,  680,  799,   26,  192,  192,  121,  720,  544,
  296,   26,  825,  300,  192,  514,  516,  392,  397,  479,
  398,  399,  400,  337,  425,  425,   28,  412,   30,  419,
  732,  732,  684,  685,  915,  192,  732,  722,    6,  791,
  340,  380,  792,   47,   47,  537,   47,   47,   47,   47,
  722,  714,  296,  677, 1127,  300,  708,   78,   24,   79,
   80,   81,   82,  677,  722,  723,   54,  606,   62,   47,
   47,  266,   47,  661,  267,  722,  661,  268,  296,  269,
  606,  300,   26,  837,   63,   91,  401,  402,  728,   49,
  663,   76,  296,  729,  681,  300,  928,   65,   47,   47,
  656,   47,  704,  645,  645,  278,  535,  121,  804,  456,
  645,  645,   26,  480,  732,  847,  848,  849,   76,  956,
 1103,  121,  192,  121,   26,  413,   52,  553,  613,  867,
  868,  681,  603,  702,  121,  107,   26,  349, 1039,  716,
  135,  340,   49,   49,  710,  284,   26,  254,  382,  717,
  189,  645,  182,  183,  711,  699, 1044,  121,  927,  681,
  194,  192,  720,  122,  488,  489,  490,  135,  192,  195,
  383,  654,  369,  285,  270,  384,   49,  560,  315,  271,
  107,  107,  370,  135,  403,  681,  121,  662,  583,  961,
  962,  963,  964,  135,  941,   91,  681,  777,  778,   94,
  662,  192,  192,  892,   97,  565,  566,  715,  890,  148,
  661,  371,  709,  372,  373,  374, 1022, 1023, 1024,  661,
  192,  481,  630,  661,  567,  662,  613,  135,   98,  722,
  704,  926,  568,  192,  192,   99,   52,  553,  135,  135,
  455,  135,  220, 1098,   56,  614,   49,  190,  661,  732,
  135,   81,   82,  320,  551,  533,   52,  233,  553,  553,
 1113,  426,  552,   32,  189,  662,  853,  668,  668,  385,
  100,   52,   52,  918,   66,  583,  681,  135,  869,  375,
  376,  386,  387,  870,  681,  681,  377, 1043,  133,  388,
  389,   49,  101,  135,  681,  695,   49,  122,  258,  390,
  102,   67,  106,  135,  103,   49,  553,  553,   49,  719,
  122,  246,  247,  121,  192,  937,  855,  192,  642,  729,
 1121,  613,  630,  104,  822,  858,  919,  315,  107,  746,
  739,  378,  690,  145,  732,  122,  920,  135,  152,  645,
  839,  840,  646,  614,  794,  150,  148,  795,  135,  135,
  745,  135,  584,  391,  153,  569,  239,  331,  332,  906,
  681,  681,  681,  311,  154,  533,   52,  553,  553,  553,
  921,   67,  156,  583,  158,   57,   58,  379,  286,  662,
  662,   52,   52,  783,   66,  455,  631,  695,  174,  192,
  881,  882,  661,   33,  681,  363,  364,   47,  320,  852,
  661,  661,  661,  162, 1056, 1057,  661,  883,   34,   35,
  661,  719,  969,  970,  190,  228,  553,  553,   81,   82,
 1070,  729,  971,   59,   60,  630,  164,  508,  509,  257,
  510,  739,   78,   24,   79,   80,   81,   82,  614,  584,
 1054,  972,  973,  974,  975,  976,   78,  339,  340,  640,
  108,  186,  648,  196,  192,  815,  702,  109,  977,  696,
  114,  978,  641,  979,  315,  816,  197, 1033, 1034,  203,
 1035, 1036,   78,  720,  953,  884,   68,   69,   70,  122,
  885,  583,  886,  730,   47,   72,  631,  122,  122,  122,
 1090,   78,  695,  122, 1055,   81,   82,  122,  296, 1062,
  199,  300,  204, 1068, 1069,  205,  719,  208,  192, 1071,
   71,   72,   78,   24,   79,   80,   81,   82,  729,  596,
  597,  630,  598,  951,  500,  733, 1075,  210,  739,  212,
  218,  855,  215,  828,  226,  896,  897,  584,  898,  645,
  858,  740,  646,  647,  916,  861,   68,   69,   70,  741,
  742,  696,   47,  228,   47,  681,  682,  911,   47,  743,
  192,  192,  693,  219,  632,  633,  695,  385, 1088,  233,
  516,  235,  786,  237,  719,  720, 1095,  817,  818,  819,
   71,   75,  820,   47,   47,  730,   47,  697,  242,  631,
  800,  780,  781,  613, 1095,  729,  668,  244,  739,  668,
  248,  698,  249,  508,  509,  296, 1040,   75,  300,  192,
   47,   47,  282,   47,  814,  109,   47,  220,  114,  192,
  192,  954,  288,   78,  826,  821,  699,   81,   82, 1066,
 1066,  700,  873,  874,  875,  876,  668,   47,   47,  289,
   47,   47,   47,   47,  291,  584,  319,   78,   78,   78,
  320,  694,  648,  324,  950,  342,  696,   78,   24,   79,
   80,   81,   82,  429,  829,  830, 1130,  175,  499,  500,
  720,  326,  109,  335,  296,  114,   47, 1053,  336,  701,
  337,   78,  730,   47,   47,  631,   72,   72,   72,   47,
   47,  165,  386,  338,  524,  525,  693,  220,  613,  508,
  509,  166,   47,   47,  422,  423,  352,  693,  368, 1058,
  614,  693,  406,  408,  157,  613,  159,  160,  161,   47,
   72,  163,  693,  831,  832,  438,  418,  611,   24,  430,
  696,   81,   82,  693,  693,  693,  439,  467,  720,  440,
   47,   47,   47,   47,  167,  442,   47,   47,   47,   47,
   26,  458,   47,   47,  693,  168,  471,  693,  693,  730,
  693,  472,  693,   47,  693,  693,  693,  476,  478,  693,
  693,  495,  693,  693,  693,  693,  693,  693,  693,  693,
  693,  501,   75,   75,   75,  694,  503,   47,  169,   47,
   47,  504,   47,  505,   47,  506,  694,  229,  519,  520,
  694,  527,  170,  961,  962,  963,  964,  109,  116,  117,
  114,  694,  529, 1131,  534,  614,   75,  538,  539,  550,
  668,  556,  694,  694,  694,  557,  468,   47,   47,  572,
  558,  947,  614,   47,  604,   47,   78,   24,   79,   80,
   81,   82,  220,  694,   47,  501,  694,  694,  616,  694,
  638,  694,  671,  694,  694,  694,  617,  618,  694,  694,
  660,  694,  694,  694,  694,  694,  694,  694,  694,  694,
  664,  467,  693,  688,  693,  693,  686,  747,  606,   47,
  770,   49,  467,  772, 1078,  771,  467,  773,  301,   47,
   47,  774,   47,  776,  779,   47,  801,  693,  693,  693,
  693,   47,  807,   47,   47,  811,   47,  838,  467,  467,
  467,  126,   47,  845,   47,   47,  859,   47,  860,   47,
  864,  878,   47,   78,   24,   79,   80,   81,   82,  467,
  879,  693,  467,  467,  888,  467,  891,  467,  892,  467,
  467,  467,   47,   47,  467,  467,  894,  467,  467,  467,
  467,  467,  467,  467,  467,  467,  619,  620,  621,  622,
  468,  694,  946,  694,  694,   78,   24,   79,   80,   81,
   82,  468,  895,  900,  909,  468,  902,  294,   78,   24,
   79,   80,   81,   82,  923,  925,  694,  694,  694,  694,
  939,  942,   47,  944,  945,  187,  108,  468,  468,  468,
  949,  119,  120,  121,  122,  123,  124,  125,  126,  967,
   47,   47,   49,   49,  984, 1077,  987,  541,  468,  666,
  694,  468,  468,  985,  468,  986,  468,  989,  468,  468,
  468,   47,   47,  468,  468,  991,  468,  468,  468,  468,
  468,  468,  468,  468,  468,  126, 1020,  467,   47,   47,
 1027,   47,  471, 1037,   47,  541,  126, 1042,  126,  126,
  126, 1046,  126,   49,   49,   49,   49,   47,   47,   47,
   47,   49,   49, 1048, 1059,  467,   78,   24,   79,   80,
   81,   82,  126,  126,  126, 1060,  126, 1072, 1073, 1074,
 1076,   47, 1080,   47,   47, 1092,   47,   78,   24,   79,
   80,   81,   82,  126, 1085, 1086,  126,  126,  477,  126,
 1096,  126, 1097,  126,  126,  126, 1106, 1107,  126,  126,
 1111,  126,  126,  126,  126,  126,  126,  126,  126,  126,
  294, 1114,  327, 1117, 1118, 1119,  468, 1120, 1122, 1126,
  969,  970, 1123, 1028, 1029, 1134, 1133, 1135,   49,   49,
  971,  666,  696,  681,  126,  126,  126,  126,   31,  126,
  126,  126,  126,  220,  468,   49,  412,  121,  192,  972,
  973,  974,  975,  976,  586,  586,  586,  586,  586,  586,
   47,  272,   47,   47,  240,   47,  977,  853,  444,  978,
  513,  979,  586,  486,  564,  471,   22,  471,  677,  541,
   32,  471,   64,  152,  153,  121,  459,  541,  541,  150,
  116,  117,  126,  126,  205,  223,  291,  541,  126,  251,
  126,  126,  126,  471,  471,  471,  252,  460,  203,  250,
  748,  740,   49,   49,  741,  126,  126,  540,   49,  624,
   49,  126,  749,  681,  471,  445,  528,  471,  471,  126,
  471,  311,  471,  477,  471,  471,  471,  477, 1109,  471,
  471,  316,  471,  471,  471,  471,  471,  471,  471,  471,
  471,  422,  446,  677,  116,  117,  379,  165, 1067,  477,
  477,  477, 1100,  851,  866,  652,  541,  166,  447,  787,
  806,  598,  598,  598,  598,  598,  598,  689,  448,  449,
  477,  872,  724,  477,  477,  901,  477, 1047,  477,  598,
  477,  477,  477, 1079,  750,  477,  477, 1081,  477,  477,
  477,  477,  477,  477,  477,  477,  477,  982,  805, 1110,
  751,  880,  450,  734,  245,  827,   25,  836,  561,  904,
  752,  753,  764,  179,  181,   25,   25,  863, 1052,  121,
   47,   47,   25,   47,   47,   47,   47,  121,  121,  586,
  451,   25,  471,  121, 1049,  461,  462,  121, 1132,  261,
   19,   19,  115,  443,  754,  299,  672,   27,  151,  463,
  464,  470,  493,  586,  486,  465,  492, 1026,  948, 1108,
  471,  452,   78,   24,   79,   80,   81,   82,   25,   25,
 1124,  287,  755,  119,  120,  121,  122,  123,  124,  125,
  126,  453,    0,   25,    0,    0,    0,    0,  477,    0,
   25,   25,   25,   25,    0,    0,    0,  118,    0,    0,
    0,    0,   20,  756,   78,   24,   79,   80,   81,   82,
    0,   20,   20,    0,    0,    0,  477,    0,   20,    0,
    0,    0,  108,  757,    0,    0,    0,   20,    0,    0,
    0,    0,   25,   25,   25,   25,   25,  119,  120,  121,
  122,  123,  124,  125,  126,  127,  598,    0,   25,   25,
    0,    0,    0,    0,    0,   19,   78,   24,   79,   80,
   81,   82,   25,   25,   20,   20,   19,    0,   25,    0,
  598,  119,  120,  121,  122,  123,  124,  125,  126,   20,
    0,  578,    0,    0,    0,  218,   20,   20,   20,   20,
    0,   19,    0,   19,    0,    0,    0,    0,    0,  218,
  218,    0,    0,    0,   25,    0,    0,    0,   25,   25,
   25,   25,   25,   25,   25,   25,   25,   25,   25,   25,
   25,   25,   25,   25,   25,   25,   25,    0,   20,   20,
   20,   20,   20,   19,   19,   19,   19,   19,   19,   19,
   19,   19,  585,  586,   20,   20,    0,  218,    0,    0,
   19,   19,   19,   19,    0,    0,    0,    0,   20,   20,
    0,    0,    0,    0,   20,   19,   19,   19,   19,   19,
    0,    0,    0,    0,  709,    0,    0,    0,    0,    0,
  134,   19,   19,  136,  137,  138,  139,  140,  141,  142,
  587,    0,    0,  709,  218,  218,  218,  218,  672,    0,
    0,   19,    0,    0,   20,   20,   20,   20,   20,   20,
   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,
   20,   20,   20,   19,  709,    0,    0,  134,   19,  135,
  136,  137,  138,  139,  140,  141,  142,  588,  589,  590,
  591,   19,   19,   19,   19,   19,   19,   19,   19,   19,
   19,   19,   19,   19,    0,   19,    0,   19,  218,   19,
  218,    0,    0,    0,    0,    0,    0,    0,   19,   19,
   19,   19,   19,   19,   19,   19,   19,    0,   19,   19,
   19,  573,   19,  731,   19,   19,  710,  711,  712,  119,
  120,  121,  122,  123,  124,  125,  126,    0,  573,  578,
  713,  592,    0,  593,    0,  710,  711,  712,    0,    0,
    0,    0,    0,    0,    0,   19,    0,  784,    0,  713,
  714,    0,    0,  119,  120,  121,  122,  123,  124,  125,
  126,  573,    0,  578,    0,    0,  710,  711,  712,    0,
  574,  575,  576,    0,    0,    0,    0,    0,  217,    0,
  713,   78,   24,   79,   80,   81,   82,  574,  575,  576,
  119,  120,  121,  122,  123,  124,  125,  126,  127,  218,
  578,    0,  992,  812,    0,    0,  993,    0,  715,  119,
  120,  121,  122,  123,  124,  125,  126,    0,    0,  578,
  574,  575,  576,    0,    0,    0,    0,  715,  994,  995,
  996,    0,  219,    0,    0,    0,    0,    0,    0,    0,
  119,  120,  121,  122,  123,  124,  125,  126,  577,  997,
  578,    0,  998,  999,    0, 1000,    0, 1001,  715, 1002,
 1003, 1004,    0,    0, 1005, 1006,    0, 1007, 1008, 1009,
 1010, 1011, 1012, 1013, 1014, 1015,   19,   19,  672,    0,
    0,    0,   19,    0,    0,   19,   19,    0,  119,  120,
  121,  122,  123,  124,  125,  126,  127,    0,  578,  679,
   19,    0,    0,    0,  579,  119,  120,  121,  122,  123,
  124,  125,  126,    0,  844,  578,  846,    0,    0,    0,
  850,  579,   19,   19,   19,   19,   19,   19,   19,   19,
   19,    0,   19,   19,   19,   19,   19,    0,  119,  120,
  121,  122,  123,  124,  125,  126,    0,    0,  578,    0,
    0,    0,    0,    0,  579,    0,   47,   47,    0,  165,
    0,    0,    0,  673,    0,    0,  220, 1016,    0,  177,
  798,    0,    0,   47,  735,    0,  119,  120,  121,  122,
  123,  124,  125,  126,    0,   19,  578,    0,  905,    0,
  907,  908,    0,    0,    0, 1017,    0,   37,  912,  913,
  914,    0,    0,  119,  120,  121,  122,  123,  124,  125,
  126,  127,  167,  578,  119,  120,  121,  122,  123,  124,
  125,  126,  127,  178,   38,   19,   19,   19,   19,   19,
   19,   19,   19,   19,   19,   19,   19,   19,    0,   19,
   39,   47,    0,   47,   47,   19,   47,    0,   47,    0,
   40,    0,    0,    0,    0,    0,  169,  934,    0,   19,
    0,    0,  936,    0,    0,    0,    0,    0,    0,    0,
  170,    0,    0,    0,    0,    0,    0,   19,    0,    0,
    0,    0,    0,    0,   41,   19,   19,   19,   19,   19,
   19,   19,   19,   19,    0,   42,   43,    0,   44,   19,
   19,   19,   19,   19,   19,   19,   19,   19,    0,   19,
   19,   19,    0,   19,   19,   19,   19,   19,   19,   19,
   19,   19,   19,   19,   19,   19,    0,   19,   19,   19,
    0,   19,  969,  970,    0, 1028, 1029,    0,  969,  970,
    0,    0,  971,    0,    0,  834,    0,    0,  971,    0,
  626,  119,  120,  121,  122,  123,  124,  125,  126,    0,
    0,  972,  973,  974,  975,  976,  691,  972,  973,  974,
  975,  976,    0,    0,    0,    0,    0,  725,  977,    0,
    0,  978,    0,  979,  977,    0,    0,  978,    0,  979,
  119,  120,  121,  122,  123,  124,  125,  126,  127,    0,
  578,    0,    0,    0,    0,    0,  119,  120,  121,  122,
  123,  124,  125,  126,  127,    0,  578,  119,  120,  121,
  122,  123,  124,  125,  126,  127,  824,  578,    0,    0,
    0,    0,  119,  120,  121,  122,  123,  124,  125,  126,
    0,    0,  578,  119,  120,  121,  122,  123,  124,  125,
  126,    0,    0,  578,  119,  120,  121,  122,  123,  124,
  125,  126,
};
short yycheck[] = {                                       6,
   35,   59,   60,  202,   59,   60,   34,   59,   60,  133,
  250,  232,  202,  306,  345,  346,  235,  182,   64,  133,
  180,   71,  370,  271,  123,  267,  417,  562,  563,  125,
  278,  610,  236,  125,  238,  737,  123,   45,  124,  454,
  371,  123,  270,  628,  123,   41,  367,  125,  396,  125,
   41,  434,  295,  125,  347,  125,  581,  123,  123,  183,
  607,  125,  410,  113,  304,  115,  397,  123,  257,  183,
  675,  292,  293,  270,  717,  123,  125,  270,    0,  372,
  411,  257,  428,  266,  693,  131,  407,  978,  321,  123,
  332,  123,   41,  256,  257,  259,  123,  285,  727,  263,
  264,  349,  592,  125,  892,  398,  164,  257,  125,  164,
  284,  235,  164,  317,   42,   43,  279,  280,  281,  257,
  123,  235,  468,  123,  248,  299,  705,  259,  124,  460,
  942,  379,   40,  124,  248,   63,  257,  299,   46,  301,
  295,  187,  266,  391,  202,  343,  388,  202,  266,  123,
  348, 1042,  266,  362,  328,  403,  477,  945,  300,  706,
  256,  292,  293,  123,  256,  413,  224,  291,  980,  224,
  353,  229,  224,  125,  229,  124,  294,  291,  257,  123,
  256,  502,  306,  125,  256,  220,  256,  266,    0,  123,
  266,  270,  256,  260,  252,  123,  517,  252,  688,  245,
  228,  257,  904,  412,  270,  526,  449,  256,  295,  237,
  443,  400,  533,  295,  270,  294,  295,  266,  253,  295,
  362,  449,  270,  347,  412,  257,  284,  558,  257,  284,
  257,  295,  296,  297,  256,  359,  270,  880,  270,  256,
  423,  299,  827,  270,  299,  359,  316,  123,  372,  266,
  454,  259,  449,  317,  257,  304,  449,  257,  344,  499,
  500,  866,  787,  872,  271,  389,  390,  270,  295,  652,
  328,  278,  901,  328,  398,  389,  390,  125,  295,  266,
  297,  298,  299,  257,  524,  525,  456,  363,  267,  453,
   40,   41,  585,  586,  449,  419,   46,  257,  461,  482,
  348,  125,  485,  479,  480,  419,  482,  483,  484,  485,
  270,  343,  370,  257,  256,  370,  348,  259,  260,  261,
  262,  263,  264,  257,  617,  618,  437,  405,  257,  479,
  480,  353,  482,  552,  356,  295,  123,  359,  396,  361,
  405,  396,  349,  758,  257,  267,  363,  364,  344,  266,
  554,  259,  410,  344,  257,  410,  891,  463,  479,  480,
  452,  482,  449,  442,  443,  461,  416,  449,  699,  125,
  449,  450,  379,  360,  124,  766,  767,  768,  259,  451,
  450,  257,  506,  449,  391,  461,  308,  309,  512,  328,
  329,  490,  506,  449,  270,  257,  403,  461,  977,  343,
  322,  449,  328,  329,  348,  266,  413,  456,  256,  343,
  131,  490,  445,  446,  348,  449,  125,  449,  890,  257,
  135,  545,  449,  123,  357,  358,  359,  349,  552,  144,
  278,  545,  256,  294,  456,  283,  353,  465,  552,  461,
  302,  303,  266,  365,  461,  283,  449,  257,  494,  401,
  402,  403,  404,  375,  926,  267,  294,  661,  662,  293,
  270,  585,  586,  270,  257,  305,  306,  343,  816,  295,
  257,  295,  348,  297,  298,  299,  962,  963,  964,  266,
  604,  468,  528,  270,  324,  295,  610,  409,  257,  449,
  604,  298,  332,  617,  618,  257,  308,  309,  420,  421,
  256,  423,  444, 1082,  309,  512,  423,  259,  295,  259,
  322,  263,  264,  269,  125,  437,  438,  343,  440,  441,
 1099,  273,  348,  308,  245,  553,  468,  562,  563,  377,
  257,  453,  454,  312,  456,  581,  439,  349,  477,  363,
  364,  389,  390,  482,  447,  448,  370,  256,  464,  397,
  398,  477,  257,  365,  457,  601,  482,  257,  125,  407,
  257,  283,  269,  375,  257,  482,  488,  489,  485,  615,
  270,  466,  467,  449,  698,  923,  775,  701,  125,  625,
 1115,  705,  628,  257,  125,  775,  365,  701,  269,  639,
  636,  415,  599,  259,  344,  295,  375,  409,  123,  342,
  760,  761,  345,  610,  299,  257,  295,  302,  420,  421,
  638,  423,  494,  461,  123,  455,  781,  435,  436,  838,
  458,  459,  460,  449,  347,  437,  438,  453,  440,  441,
  409,  283,  257,  679,  123,  440,  441,  461,  125,  449,
  450,  453,  454,  671,  456,  256,  528,  693,  257,  773,
  392,  393,  439,  438,  492,  354,  355,  266,  269,  773,
  447,  448,  449,  123,  995,  996,  453,  409,  453,  454,
  457,  717,  381,  382,  259,  257,  488,  489,  263,  264,
 1011,  727,  391,  488,  489,  731,  123,  428,  429,  256,
  431,  737,  259,  260,  261,  262,  263,  264,  705,  581,
  993,  410,  411,  412,  413,  414,  257,  442,  443,  256,
  277,  257,  455,   58,  838,  256,  125,  775,  427,  601,
  775,  430,  269,  432,  838,  266,  259,  969,  970,  259,
  972,  973,  283,  615,  933,  477,  458,  459,  460,  439,
  482,  787,  484,  625,  353,  257,  628,  447,  448,  449,
 1071,  259,  798,  453,  994,  263,  264,  457,  816, 1001,
  294,  816,  257, 1005, 1006,  123,  812,  123,  892, 1017,
  492,  283,  259,  260,  261,  262,  263,  264,  824,  428,
  429,  827,  431,  125,  257,  631, 1028,  123,  834,  123,
  277,  990,  123,  291,  300,  428,  429,  679,  431,  342,
  990,  439,  345,  346,  854,  313,  458,  459,  460,  447,
  448,  693,  291,  453,  423,  479,  480,  845,  291,  457,
  944,  945,  123,  310,  424,  425,  872,  257, 1049,  343,
  944,  348,  678,  453,  880,  717, 1076,  378,  379,  380,
  492,  257,  383,  479,  480,  727,  482,  256,  490,  731,
  696,  302,  303,  977, 1094,  901,  891,  123,  904,  894,
  465,  270,  259,  428,  429,  923,  431,  283,  923,  993,
  479,  480,  257,  482,  720,  933,  485,  444,  933, 1003,
 1004,  933,  257,  259,  730,  426,  295,  263,  264, 1003,
 1004,  300,  482,  483,  484,  485,  931,  479,  480,  259,
  482,  483,  484,  485,  348,  787,  123,  458,  459,  460,
  269,  123,  455,   58,  256,  450,  798,  259,  260,  261,
  262,  263,  264,  123,  422,  423, 1125,  257,  375,  376,
  812,  257,  990,  257,  992,  990,  266,  992,  257,  348,
  257,  492,  824,  422,  423,  827,  458,  459,  460,  422,
  423,  312,  257,  257,  401,  402,  257,  444, 1082,  428,
  429,  322,  392,  393,  292,  293,  257,  268,  257,  997,
  977,  272,  257,  257,   89, 1099,   91,   92,   93,  409,
  492,   96,  283,  481,  482,  123,  257,  259,  260,  257,
  872,  263,  264,  294,  295,  296,  257,  123,  880,  257,
  479,  480,  481,  482,  365,  257,  479,  480,  481,  482,
 1017,  259,  328,  329,  315,  376,  257,  318,  319,  901,
  321,  257,  323,  353,  325,  326,  327,  300,  257,  330,
  331,  257,  333,  334,  335,  336,  337,  338,  339,  340,
  341,  430,  458,  459,  460,  257,  257,  477,  409,  479,
  480,  257,  482,  123,  484,  399,  268,  266,  257,  257,
  272,  257,  423,  401,  402,  403,  404, 1125,  275,  276,
 1125,  283,  257, 1125,  257, 1082,  492,  257,  257,  257,
 1115,  257,  294,  295,  296,  257,  123,  392,  393,  259,
  295,  125, 1099,  423,  400,  266,  259,  260,  261,  262,
  263,  264,  444,  315,  409,  257,  318,  319,  123,  321,
  453,  323,  314,  325,  326,  327,  271,  272,  330,  331,
  257,  333,  334,  335,  336,  337,  338,  339,  340,  341,
  257,  257,  433,  430,  435,  436,  259,  257,  405,  291,
  257,  291,  268,  125,  125,  124,  272,   44,  311,  479,
  480,  257,  482,  257,  257,  485,  257,  458,  459,  460,
  461,  477,  257,  479,  480,  257,  482,  348,  294,  295,
  296,  123,  477,  453,  479,  480,  257,  482,  257,  484,
  269,  406,  353,  259,  260,  261,  262,  263,  264,  315,
  257,  492,  318,  319,  257,  321,  386,  323,  270,  325,
  326,  327,  328,  329,  330,  331,  123,  333,  334,  335,
  336,  337,  338,  339,  340,  341,  371,  372,  373,  374,
  257,  433,  256,  435,  436,  259,  260,  261,  262,  263,
  264,  268,  123,  257,  123,  272,  482,  313,  259,  260,
  261,  262,  263,  264,  266,  259,  458,  459,  460,  461,
  123,  123,  423,  387,  270,  462,  277,  294,  295,  296,
  257,  468,  469,  470,  471,  472,  473,  474,  475,  123,
  422,  423,  422,  423,  257,  256,  449,  257,  315,  313,
  492,  318,  319,  257,  321,  257,  323,  257,  325,  326,
  327,  328,  329,  330,  331,  257,  333,  334,  335,  336,
  337,  338,  339,  340,  341,  257,  259,  433,  479,  480,
  257,  482,  123,  259,  485,  295,  268,  430,  270,  271,
  272,  257,  274,  482,  483,  484,  485,  479,  480,  481,
  482,  481,  482,  257,  259,  461,  259,  260,  261,  262,
  263,  264,  294,  295,  296,  259,  298,  123,   47,  257,
  123,  477,  257,  479,  480,  259,  482,  259,  260,  261,
  262,  263,  264,  315,  257,  257,  318,  319,  123,  321,
  257,  323,  257,  325,  326,  327,   47,  257,  330,  331,
  388,  333,  334,  335,  336,  337,  338,  339,  340,  341,
  313,  257,  468,  257,  259,  257,  433,  257,  123,   58,
  381,  382,  259,  384,  385,  257,  259,  257,  392,  393,
  391,  313,  257,  123,  366,  367,  368,  369,  257,  371,
  372,  373,  374,  444,  461,  409,  257,  257,  257,  410,
  411,  412,  413,  414,  259,  260,  261,  262,  263,  264,
  477,  257,  479,  480,  257,  482,  427,  468,  265,  430,
  257,  432,  277,  453,  123,  266,  465,  268,  123,  439,
  257,  272,  257,  257,  257,  295,  274,  447,  448,  257,
  275,  276,  424,  425,  257,  260,  257,  457,  430,  257,
  432,  433,  434,  294,  295,  296,  257,  295,  257,  257,
  265,  125,  482,  477,  125,  447,  448,  257,  482,  257,
  484,  453,  277,  257,  315,  322,  257,  318,  319,  461,
  321,  257,  323,  268,  325,  326,  327,  272,  125,  330,
  331,  257,  333,  334,  335,  336,  337,  338,  339,  340,
  341,  123,  349,  257,  275,  276,  257,  312, 1004,  294,
  295,  296, 1090,  771,  784,  543,  432,  322,  365,  679,
  701,  259,  260,  261,  262,  263,  264,  595,  375,  376,
  315,  798,  624,  318,  319,  824,  321,  982,  323,  277,
  325,  326,  327, 1031,  349,  330,  331, 1040,  333,  334,
  335,  336,  337,  338,  339,  340,  341,  944,  700, 1094,
  365,  812,  409,  635,  187,  731,  266,  746,  466,  834,
  375,  376,  641,  107,  107,  275,  276,  781,  992,  439,
  479,  480,  282,  482,  483,  484,  485,  447,  448,  444,
  437,  291,  433,  453,  990,  433,  434,  457, 1125,  202,
  275,  276,   60,  319,  409,  229,  343,    8,   72,  447,
  448,  333,  365,  468,  357,  453,  361,  965,  931,  256,
  461,  468,  259,  260,  261,  262,  263,  264,  328,  329,
 1121,  224,  437,  468,  469,  470,  471,  472,  473,  474,
  475,  488,   -1,  343,   -1,   -1,   -1,   -1,  433,   -1,
  350,  351,  352,  353,   -1,   -1,   -1,  428,   -1,   -1,
   -1,   -1,  266,  468,  259,  260,  261,  262,  263,  264,
   -1,  275,  276,   -1,   -1,   -1,  461,   -1,  282,   -1,
   -1,   -1,  277,  488,   -1,   -1,   -1,  291,   -1,   -1,
   -1,   -1,  392,  393,  394,  395,  396,  468,  469,  470,
  471,  472,  473,  474,  475,  476,  444,   -1,  408,  409,
   -1,   -1,   -1,   -1,   -1,  282,  259,  260,  261,  262,
  263,  264,  422,  423,  328,  329,  266,   -1,  428,   -1,
  468,  468,  469,  470,  471,  472,  473,  474,  475,  343,
   -1,  478,   -1,   -1,   -1,  257,  350,  351,  352,  353,
   -1,  291,   -1,  428,   -1,   -1,   -1,   -1,   -1,  271,
  272,   -1,   -1,   -1,  464,   -1,   -1,   -1,  468,  469,
  470,  471,  472,  473,  474,  475,  476,  477,  478,  479,
  480,  481,  482,  483,  484,  485,  486,   -1,  392,  393,
  394,  395,  396,  468,  469,  470,  471,  472,  473,  474,
  475,  476,  271,  272,  408,  409,   -1,  319,   -1,   -1,
  350,  351,  352,  353,   -1,   -1,   -1,   -1,  422,  423,
   -1,   -1,   -1,   -1,  428,  392,  393,  394,  395,  396,
   -1,   -1,   -1,   -1,  282,   -1,   -1,   -1,   -1,   -1,
  282,  408,  409,  285,  286,  287,  288,  289,  290,  291,
  319,   -1,   -1,  282,  366,  367,  368,  369,  343,   -1,
   -1,  428,   -1,   -1,  468,  469,  470,  471,  472,  473,
  474,  475,  476,  477,  478,  479,  480,  481,  482,  483,
  484,  485,  486,  423,  282,   -1,   -1,  282,  428,  284,
  285,  286,  287,  288,  289,  290,  291,  366,  367,  368,
  369,  468,  469,  470,  471,  472,  473,  474,  475,  476,
  477,  478,  479,  480,   -1,  482,   -1,  484,  430,  486,
  432,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  468,  469,
  470,  471,  472,  473,  474,  475,  476,   -1,  478,  479,
  480,  291,  482,  462,  484,  485,  394,  395,  396,  468,
  469,  470,  471,  472,  473,  474,  475,   -1,  291,  478,
  408,  430,   -1,  432,   -1,  394,  395,  396,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  291,   -1,  462,   -1,  408,
  428,   -1,   -1,  468,  469,  470,  471,  472,  473,  474,
  475,  291,   -1,  478,   -1,   -1,  394,  395,  396,   -1,
  350,  351,  352,   -1,   -1,   -1,   -1,   -1,  256,   -1,
  408,  259,  260,  261,  262,  263,  264,  350,  351,  352,
  468,  469,  470,  471,  472,  473,  474,  475,  476,  277,
  478,   -1,  268,  462,   -1,   -1,  272,   -1,  486,  468,
  469,  470,  471,  472,  473,  474,  475,   -1,   -1,  478,
  350,  351,  352,   -1,   -1,   -1,   -1,  486,  294,  295,
  296,   -1,  310,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  468,  469,  470,  471,  472,  473,  474,  475,  428,  315,
  478,   -1,  318,  319,   -1,  321,   -1,  323,  486,  325,
  326,  327,   -1,   -1,  330,  331,   -1,  333,  334,  335,
  336,  337,  338,  339,  340,  341,  422,  423,  343,   -1,
   -1,   -1,  428,   -1,   -1,  328,  329,   -1,  468,  469,
  470,  471,  472,  473,  474,  475,  476,   -1,  478,  462,
  343,   -1,   -1,   -1,  484,  468,  469,  470,  471,  472,
  473,  474,  475,   -1,  763,  478,  765,   -1,   -1,   -1,
  769,  484,  468,  469,  470,  471,  472,  473,  474,  475,
  476,   -1,  478,  479,  480,  481,  482,   -1,  468,  469,
  470,  471,  472,  473,  474,  475,   -1,   -1,  478,   -1,
   -1,   -1,   -1,   -1,  484,   -1,  392,  393,   -1,  312,
   -1,   -1,   -1,  428,   -1,   -1,  444,  433,   -1,  322,
  462,   -1,   -1,  409,  428,   -1,  468,  469,  470,  471,
  472,  473,  474,  475,   -1,  428,  478,   -1,  837,   -1,
  839,  840,   -1,   -1,   -1,  461,   -1,  322,  847,  848,
  849,   -1,   -1,  468,  469,  470,  471,  472,  473,  474,
  475,  476,  365,  478,  468,  469,  470,  471,  472,  473,
  474,  475,  476,  376,  349,  468,  469,  470,  471,  472,
  473,  474,  475,  476,  477,  478,  479,  480,   -1,  482,
  365,  477,   -1,  479,  480,  428,  482,   -1,  484,   -1,
  375,   -1,   -1,   -1,   -1,   -1,  409,  906,   -1,  428,
   -1,   -1,  911,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  423,   -1,   -1,   -1,   -1,   -1,   -1,  428,   -1,   -1,
   -1,   -1,   -1,   -1,  409,  468,  469,  470,  471,  472,
  473,  474,  475,  476,   -1,  420,  421,   -1,  423,  468,
  469,  470,  471,  472,  473,  474,  475,  476,   -1,  478,
  479,  480,   -1,  482,  483,  484,  485,  468,  469,  470,
  471,  472,  473,  474,  475,  476,   -1,  478,  479,  480,
   -1,  482,  381,  382,   -1,  384,  385,   -1,  381,  382,
   -1,   -1,  391,   -1,   -1,  462,   -1,   -1,  391,   -1,
  428,  468,  469,  470,  471,  472,  473,  474,  475,   -1,
   -1,  410,  411,  412,  413,  414,  428,  410,  411,  412,
  413,  414,   -1,   -1,   -1,   -1,   -1,  428,  427,   -1,
   -1,  430,   -1,  432,  427,   -1,   -1,  430,   -1,  432,
  468,  469,  470,  471,  472,  473,  474,  475,  476,   -1,
  478,   -1,   -1,   -1,   -1,   -1,  468,  469,  470,  471,
  472,  473,  474,  475,  476,   -1,  478,  468,  469,  470,
  471,  472,  473,  474,  475,  476,  462,  478,   -1,   -1,
   -1,   -1,  468,  469,  470,  471,  472,  473,  474,  475,
   -1,   -1,  478,  468,  469,  470,  471,  472,  473,  474,
  475,   -1,   -1,  478,  468,  469,  470,  471,  472,  473,
  474,  475,
};
#define YYFINAL 7
#ifndef YYDEBUG
#define YYDEBUG 0
#endif
#define YYMAXTOKEN 492
#if YYDEBUG
char *yyname[] = {
"end-of-file",0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,"'('","')'","'*'","'+'","','","'-'","'.'","'/'",0,0,0,0,0,0,0,0,0,0,
"':'",0,0,0,0,"'?'",0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,"'{'","'|'","'}'",0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,"L_EOS","L_ERROR","L_NUMBER","L_STRING","L_HNAME","L_KEYWORD",
"L_HEXSTRING","L_OCTETSTRING","T_DIRECT","T_INTERFACE","T_INTERFACES",
"T_LCLADDR","T_PROTO","T_METRIC","T_METRICIN","T_METRICOUT","T_INFINITY",
"T_DOWN","T_PARSE","T_ADV","T_DEFAULT","T_DEFAULTS","T_YYDEBUG","T_YYSTATE",
"T_YYQUIT","T_DEBUG","T_SYSLOG","T_UPTO","T_EMERG","T_ALERT","T_CRIT","T_ERR",
"T_WARNING","T_NOTICE","T_INFO","T_ON","T_OFF","T_GATEWAY","T_PREFERENCE",
"T_PREFERENCE2","T_DEFAULTMETRIC","T_NONBROADCAST","T_BROADCAST","T_TYPE",
"T_POINTOPOINT","T_MULTICAST","T_NETMASK","T_DEFINE","T_INTERNAL","T_EXTERNAL",
"T_IGNORE","T_NOP","T_NOP1","T_NOP_DM","T_NOP_ADV","T_NOP_PROTO",
"T_NOP_SOCKADDR","T_PEERAS","T_LOCALAS","T_NEIGHBOR","T_GROUP","T_NOGENDEFAULT",
"T_VERSION","T_ACQUIRE","T_TTL","T_BGP","T_HOLDTIME","T_TEST","T_KEEP",
"T_SENDBUF","T_RECVBUF","T_OPEN","T_KEEPALIVE","T_INDELAY","T_OUTDELAY",
"T_ROUTING","T_ANALRETENTIVE","T_KEEPALIVESALWAYS","T_NOAGGRID","T_NOAUTHCHECK",
"T_SETPREF","T_IGNOREFIRSTASHOP","T_NOV4ASLOOP","T_V3ASLOOPOKAY","T_LOGUPDOWN",
"T_EGP","T_ASPATH","T_ORIGIN","T_INCOMPLETE","T_ANY","T_LOOPS","T_TAG",
"T_KERNEL","T_SYMBOLS","T_REMNANTS","T_IFLIST","T_ROUTES","T_NOCHANGE",
"T_NOFLUSHATEXIT","T_REMNANTHOLDTIME","T_HIGHER","T_LOWER","T_FLASH",
"T_INTERIOR","T_BACKGROUND","T_LIMIT","T_TRUSTEDGATEWAYS","T_SOURCEGATEWAYS",
"T_RIP","T_RIPOUT","T_NORIPOUT","T_RIPIN","T_NORIPIN","T_NOCHECKZERO",
"T_HELLOOUT","T_HELLOIN","T_NOHELLOOUT","T_NOHELLOIN","T_OSPF","T_OSPF_ASE",
"T_BACKBONE","T_VIRTUALLINK","T_STUB","T_STUBHOSTS","T_RXMITINTERVAL",
"T_INFTRANSDELAY","T_NETWORKS","T_POLLINTERVAL","T_ROUTERS","T_NEIGHBORID",
"T_TRANSITAREA","T_ELIGIBLE","T_MONITORAUTH","T_MONITORAUTHKEY","T_NOMULTICAST",
"T_DD","T_ACK","T_LSA_BLD","T_LSA_RX","T_LSA_TX","T_EXPORTINTERVAL",
"T_EXPORTLIMIT","T_FIRST","T_EVERY","T_START_ACCEPT","T_STOP_ACCEPT",
"T_START_GENERATE","T_STOP_GENERATE","T_KEY","T_ID","T_AREA","T_SPF","T_HELLO",
"T_HELLOINTERVAL","T_ROUTERDEADINTERVAL","T_PRIORITY","T_ENABLE","T_DISABLE",
"T_QUERY","T_QUERYINTERVAL","T_MODE","T_ADVERTISE","T_REGISTER","T_SNMP",
"T_ICMP","T_ROUTERDISCOVERY","T_REDIRECT","T_REDIRECTS","T_NOREDIRECTS",
"T_AUTHTYPE","T_AUTHKEY","T_NONE","T_SIMPLE","T_AUTH","T_MD5","T_SECONDARY",
"T_PASSIVE","T_SIMPLEX","T_STRICTIFS","T_SCANINTERVAL","T_STATIC","T_MARTIANS",
"T_RETAIN","T_EXPORT","T_IMPORT","T_EXACT","T_REFINE","T_HOST","T_MASK",
"T_MASKLEN","T_REJECT","T_BLACKHOLE","T_RESTRICT","T_ALLOW","T_ANNOUNCE",
"T_LISTEN","T_AS","T_ROUTERID","T_IGP","T_OPTIONS","T_NOINSTALL","T_NOSEND",
"T_NORESOLV","T_MARK","T_TRACEOPTIONS","T_EXCEPT","T_REPLACE","T_SIZE",
"T_FILES","T_K","T_M","T_ALL","T_ROUTE","T_STATE","T_POLICY","T_NORMAL",
"T_GENERAL","T_TASK","T_TIMER","T_NOSTAMP","T_UPDATE","T_DETAIL","T_SEND",
"T_RECEIVE","T_ERROR","T_PACKETS","T_RESPONSE","T_REQUEST","T_OTHER","T_TRAP",
"L_USEC","T_AGGREGATE","T_GENERATE","T_BRIEF","T_INET","T_GENDEFAULT",
};
char *yyrule[] = {
"$accept : config",
"config :",
"config : statements",
"statements : statement",
"statements : statements statement",
"statement : parse_statement",
"statement : trace_statement",
"statement : option_order option_statement",
"statement : interface_order interface_statement",
"statement : define_order define_statement",
"statement : proto_order proto_statement",
"statement : route_order route_statement",
"statement : control_order control_statement",
"statement : error L_EOS",
"statement : L_EOS",
"parse_statement : T_YYDEBUG onoff_option L_EOS",
"parse_statement : T_YYSTATE L_NUMBER L_EOS",
"parse_statement : T_YYQUIT L_EOS",
"trace_statement : T_TRACEOPTIONS trace_file_option trace_control_option trace_trace_options_none L_EOS",
"trace_file_option :",
"trace_file_option : string trace_replace",
"trace_file_option : string trace_replace T_SIZE trace_size T_FILES number",
"trace_size : number",
"trace_size : number T_K",
"trace_size : number T_M",
"trace_replace :",
"trace_replace : T_REPLACE",
"trace_control_option :",
"trace_control_option : trace_control_option trace_control_options",
"trace_control_options : T_NOSTAMP",
"trace_trace_options_none : T_NONE",
"trace_trace_options_none : trace_trace_options",
"trace_trace_options_none : trace_trace_options T_EXCEPT trace_trace_options",
"trace_trace_options : trace_option",
"trace_trace_options : trace_option_global",
"trace_trace_options : trace_trace_options trace_option",
"trace_trace_options : trace_trace_options trace_option_global",
"trace_option_global : T_PARSE",
"trace_option_global : T_ADV",
"trace_option : T_ROUTE",
"trace_option : T_NORMAL",
"trace_option : T_GENERAL",
"trace_option : T_STATE",
"trace_option : T_POLICY",
"trace_option : T_TASK",
"trace_option : T_TIMER",
"trace_option : T_ALL",
"trace_option_detail :",
"trace_option_detail : T_DETAIL",
"trace_option_sendrecv :",
"trace_option_sendrecv : T_SEND",
"trace_option_sendrecv : T_RECEIVE",
"define_order :",
"define_statement : T_NOP",
"$$1 :",
"define_statement : T_MARTIANS $$1 '{' martian_list_inet '}' L_EOS",
"define_statement : T_AS as as_loop_option L_EOS",
"define_statement : T_ROUTERID host_inet L_EOS",
"martian_list_inet :",
"martian_list_inet : martian_list_inet martian_inet martian_option L_EOS",
"martian_list_inet : martian_list_inet error L_EOS",
"martian_inet : host_mask_inet",
"martian_inet : network_mask_flags_inet",
"martian_inet : default_mask_inet",
"martian_option :",
"martian_option : T_ALLOW",
"option_order :",
"option_statement : T_OPTIONS option_list L_EOS",
"option_list : option",
"option_list : option_list option",
"option : T_NOSEND",
"option : T_NORESOLV",
"option : T_SYSLOG syslog_mask",
"option : T_MARK mark_time",
"option : T_GENDEFAULT preference_option default_gateway_option",
"default_gateway_option :",
"$$2 :",
"default_gateway_option : T_GATEWAY $$2 gateway_inet",
"mark_time :",
"mark_time : time",
"syslog_mask : syslog_level",
"syslog_mask : T_UPTO syslog_level",
"syslog_mask : syslog_mask syslog_level",
"syslog_level : T_EMERG",
"syslog_level : T_ALERT",
"syslog_level : T_CRIT",
"syslog_level : T_ERR",
"syslog_level : T_WARNING",
"syslog_level : T_NOTICE",
"syslog_level : T_INFO",
"syslog_level : T_DEBUG",
"interface_order :",
"$$3 :",
"interface_statement : T_INTERFACES $$3 '{' interface_stmts '}' L_EOS",
"interface_stmts :",
"interface_stmts : interface_stmts interface_stmt L_EOS",
"interface_stmts : interface_stmts error L_EOS",
"interface_stmt : T_OPTIONS interface_options",
"interface_stmt : T_INTERFACE interface_all_list interface_define_options",
"$$4 :",
"interface_stmt : T_DEFINE host_inet $$4 interface_type_options interface_common_options",
"interface_options : interface_option",
"interface_options : interface_options interface_option",
"interface_option : T_STRICTIFS",
"interface_option : T_SCANINTERVAL time",
"interface_type_options : T_BROADCAST host",
"interface_type_options : T_POINTOPOINT host",
"interface_type_options :",
"interface_common_options :",
"interface_common_options : interface_common_options interface_common_option",
"interface_common_option : T_NETMASK mask",
"interface_common_option : T_MULTICAST",
"interface_define_options : interface_define_option",
"interface_define_options : interface_define_options interface_define_option",
"interface_define_option : T_PREFERENCE preference",
"interface_define_option : T_DOWN T_PREFERENCE preference",
"interface_define_option : T_PASSIVE",
"interface_define_option : T_SIMPLEX",
"interface_define_option : T_REJECT",
"interface_define_option : T_BLACKHOLE",
"interface_define_option : T_AS as",
"interface_list_option :",
"interface_list_option : T_INTERFACE interface_list",
"interface_gateway_list : interface_list_option",
"interface_gateway_list : T_GATEWAY gateway_list",
"interface_all_list : T_ALL",
"interface_all_list : interface_list",
"interface_list : interface_all",
"interface_list : interface_list interface_all",
"interface_all : interface_all_adv",
"interface_all_adv : host_name",
"interface_all_adv : addr",
"interface_local : interface_addr",
"interface_addr : addr",
"interface_addr : host_name",
"proto_order :",
"proto_statement : redirect_statement L_EOS",
"proto_statement : kernel_statement L_EOS",
"proto_statement : bgp_statement L_EOS",
"proto_statement : hello_statement L_EOS",
"proto_statement : icmp_statement L_EOS",
"proto_statement : ospf_statement L_EOS",
"proto_statement : rip_statement L_EOS",
"proto_statement : snmp_statement L_EOS",
"$$5 :",
"kernel_statement : T_KERNEL $$5 '{' kernel_group '}'",
"kernel_group :",
"kernel_group : kernel_group kernel_stmt L_EOS",
"kernel_group : kernel_group error L_EOS",
"kernel_stmt : T_TRACEOPTIONS trace_file_option trace_control_option kernel_trace_options_none",
"kernel_stmt : T_OPTIONS kernel_options",
"kernel_stmt : T_REMNANTHOLDTIME time",
"kernel_stmt : T_FLASH kernel_flash_options",
"kernel_stmt : T_BACKGROUND kernel_bg_options",
"kernel_stmt : T_ROUTES number",
"kernel_options : kernel_option",
"kernel_options : kernel_options kernel_option",
"kernel_option : T_NOCHANGE",
"kernel_option : T_NOFLUSHATEXIT",
"kernel_flash_options : kernel_flash_option",
"kernel_flash_options : kernel_flash_options kernel_flash_option",
"kernel_flash_option : T_TYPE kernel_flash_type",
"kernel_flash_option : T_LIMIT neg_number",
"kernel_flash_type : T_INTERFACE",
"kernel_flash_type : T_INTERIOR",
"kernel_flash_type : T_ALL",
"kernel_bg_options : kernel_bg_option",
"kernel_bg_options : kernel_bg_options kernel_bg_option",
"kernel_bg_option : T_PRIORITY kernel_bg_type",
"kernel_bg_option : T_LIMIT number",
"kernel_bg_type : T_FLASH",
"kernel_bg_type : T_HIGHER",
"kernel_bg_type : T_LOWER",
"kernel_trace_options_none : T_NONE",
"kernel_trace_options_none : kernel_trace_options",
"kernel_trace_options_none : kernel_trace_options T_EXCEPT kernel_trace_options",
"kernel_trace_options : kernel_trace_option",
"kernel_trace_options : kernel_trace_options kernel_trace_option",
"kernel_trace_option : trace_option",
"kernel_trace_option : trace_option_detail trace_option_sendrecv kernel_trace_option_index",
"kernel_trace_option : T_SYMBOLS",
"kernel_trace_option : T_REMNANTS",
"kernel_trace_option : T_IFLIST",
"kernel_trace_option : T_REQUEST",
"kernel_trace_option : T_INFO",
"kernel_trace_option_index : T_PACKETS",
"kernel_trace_option_index : T_ROUTES",
"kernel_trace_option_index : T_REDIRECT",
"kernel_trace_option_index : T_INTERFACE",
"kernel_trace_option_index : T_OTHER",
"$$6 :",
"rip_statement : T_RIP $$6 onoff_option rip_group",
"rip_group :",
"rip_group : '{' rip_group_stmts '}'",
"rip_group_stmts :",
"rip_group_stmts : rip_group_stmts rip_group_stmt L_EOS",
"rip_group_stmts : rip_group_stmts error L_EOS",
"rip_group_stmt : T_PREFERENCE preference",
"rip_group_stmt : T_DEFAULTMETRIC metric",
"rip_group_stmt : T_BROADCAST",
"rip_group_stmt : T_NONBROADCAST",
"rip_group_stmt : T_TRACEOPTIONS trace_file_option trace_control_option rip_trace_options_none",
"rip_group_stmt : T_NOCHECKZERO",
"rip_group_stmt : T_INTERFACE interface_all_list rip_interface_options",
"rip_group_stmt : T_TRUSTEDGATEWAYS gateway_list_inet",
"rip_group_stmt : T_SOURCEGATEWAYS gateway_list_inet",
"rip_group_stmt : T_QUERY T_AUTH rip_auth",
"rip_interface_options : rip_interface_option",
"rip_interface_options : rip_interface_options rip_interface_option",
"rip_interface_option : T_RIPIN",
"rip_interface_option : T_NORIPIN",
"rip_interface_option : T_RIPOUT",
"rip_interface_option : T_NORIPOUT",
"rip_interface_option : T_METRICIN metric",
"rip_interface_option : T_METRICOUT metric",
"rip_interface_option : T_AUTH rip_auth",
"rip_interface_option : T_SECONDARY T_AUTH rip_auth",
"rip_interface_option : T_VERSION L_NUMBER rip_v2_options",
"rip_v2_options :",
"rip_v2_options : T_MULTICAST",
"rip_v2_options : T_BROADCAST",
"rip_auth : T_NONE",
"rip_auth : rip_auth_type string",
"rip_auth_type :",
"rip_auth_type : T_SIMPLE",
"rip_auth_type : T_MD5",
"rip_trace_options_none : T_NONE",
"rip_trace_options_none : rip_trace_options",
"rip_trace_options_none : rip_trace_options T_EXCEPT rip_trace_options",
"rip_trace_options : rip_trace_option",
"rip_trace_options : rip_trace_options rip_trace_option",
"rip_trace_option : trace_option",
"rip_trace_option : trace_option_detail trace_option_sendrecv rip_trace_option_index",
"rip_trace_option_index : T_PACKETS",
"rip_trace_option_index : T_REQUEST",
"rip_trace_option_index : T_RESPONSE",
"rip_trace_option_index : T_OTHER",
"rip_prop_init : T_RIP",
"$$7 :",
"hello_statement : T_HELLO $$7 onoff_option hello_group",
"hello_group :",
"hello_group : '{' hello_group_stmts '}'",
"hello_group_stmts :",
"hello_group_stmts : hello_group_stmts hello_group_stmt L_EOS",
"hello_group_stmts : hello_group_stmts error L_EOS",
"hello_group_stmt : T_PREFERENCE preference",
"hello_group_stmt : T_DEFAULTMETRIC metric",
"hello_group_stmt : T_BROADCAST",
"hello_group_stmt : T_NONBROADCAST",
"hello_group_stmt : T_TRACEOPTIONS trace_file_option trace_control_option hello_trace_options_none",
"hello_group_stmt : T_INTERFACE interface_all_list hello_interface_options",
"hello_group_stmt : T_TRUSTEDGATEWAYS gateway_list_inet",
"hello_group_stmt : T_SOURCEGATEWAYS gateway_list_inet",
"hello_interface_options : hello_interface_option",
"hello_interface_options : hello_interface_options hello_interface_option",
"hello_interface_option : T_HELLOIN",
"hello_interface_option : T_NOHELLOIN",
"hello_interface_option : T_HELLOOUT",
"hello_interface_option : T_NOHELLOOUT",
"hello_interface_option : T_METRICIN metric",
"hello_interface_option : T_METRICOUT metric",
"hello_trace_options_none : T_NONE",
"hello_trace_options_none : hello_trace_options",
"hello_trace_options_none : hello_trace_options T_EXCEPT hello_trace_options",
"hello_trace_options : hello_trace_option",
"hello_trace_options : hello_trace_options hello_trace_option",
"hello_trace_option : trace_option",
"hello_trace_option : trace_option_detail trace_option_sendrecv hello_trace_option_index",
"hello_trace_option_index : T_PACKETS",
"hello_prop_init : T_HELLO",
"$$8 :",
"ospf_statement : T_OSPF $$8 onoff_option ospf_group",
"ospf_group :",
"ospf_group : '{' ospf_group_stmts '}'",
"ospf_group_stmts :",
"ospf_group_stmts : ospf_group_stmts ospf_group_stmt L_EOS",
"ospf_group_stmts : ospf_group_stmts error L_EOS",
"ospf_group_stmt : T_DEFAULTS '{' ospf_defaults '}'",
"ospf_group_stmt : T_TRACEOPTIONS trace_file_option trace_control_option ospf_trace_options_none",
"ospf_group_stmt : T_EXPORTINTERVAL time",
"ospf_group_stmt : T_EXPORTLIMIT number",
"ospf_group_stmt : T_MONITORAUTH T_MD5 ospf_md5_auth",
"$$9 :",
"ospf_group_stmt : T_MONITORAUTH ospf_simple_auth_type $$9 ospf_auth_key",
"$$10 :",
"ospf_group_stmt : T_MONITORAUTHKEY $$10 ospf_auth_key",
"ospf_group_stmt : T_SYSLOG ospf_syslog_first ospf_syslog_every",
"$$11 :",
"ospf_group_stmt : ospf_area_option $$11 '{' ospf_area_stmts '}'",
"ospf_syslog_first :",
"ospf_syslog_first : T_FIRST number",
"ospf_syslog_every :",
"ospf_syslog_every : T_EVERY number",
"ospf_defaults :",
"ospf_defaults : ospf_defaults ospf_default L_EOS",
"ospf_defaults : ospf_defaults error L_EOS",
"ospf_default : T_PREFERENCE preference",
"ospf_default : T_METRIC ospf_cost",
"ospf_default : T_TAG tag_as",
"ospf_default : T_TYPE ospf_type",
"ospf_area_stmts :",
"ospf_area_stmts : ospf_area_stmts ospf_area_stmt L_EOS",
"ospf_area_stmts : ospf_area_stmts error L_EOS",
"ospf_area_stmt : T_AUTHTYPE ospf_auth_type",
"ospf_area_stmt : T_STUB ospf_cost_option",
"ospf_area_stmt : T_STUBHOSTS '{' ospf_stub_list '}'",
"ospf_area_stmt : T_NETWORKS '{' ospf_area_range_list '}'",
"ospf_area_stmt : T_INTERFACE ospf_interface",
"ospf_area_stmt : T_VIRTUALLINK T_NEIGHBORID host T_TRANSITAREA ospf_area ospf_common",
"ospf_interface : interface_all_list ospf_cost_option ospf_common",
"ospf_interface : interface_all_list T_NONBROADCAST ospf_cost_option ospf_nonbroadcast",
"ospf_common :",
"ospf_common : '{' ospf_common_options '}'",
"ospf_common_options : ospf_common_option L_EOS",
"ospf_common_options : ospf_common_options ospf_common_option L_EOS",
"ospf_common_options : ospf_common_options error L_EOS",
"ospf_nonbroadcast :",
"ospf_nonbroadcast : '{' ospf_nonbroadcast_options '}'",
"ospf_nonbroadcast_options : ospf_nonbroadcast_option L_EOS",
"ospf_nonbroadcast_options : ospf_nonbroadcast_options ospf_nonbroadcast_option L_EOS",
"ospf_nonbroadcast_options : ospf_nonbroadcast_options error L_EOS",
"ospf_common_option : T_ENABLE",
"ospf_common_option : T_DISABLE",
"ospf_common_option : T_RXMITINTERVAL time",
"ospf_common_option : T_INFTRANSDELAY time",
"ospf_common_option : T_PRIORITY ospf_priority",
"ospf_common_option : T_HELLOINTERVAL time",
"ospf_common_option : T_ROUTERDEADINTERVAL time",
"ospf_common_option : T_NOMULTICAST",
"ospf_common_option : T_AUTH T_MD5 ospf_md5_auth",
"$$12 :",
"ospf_common_option : T_AUTH ospf_simple_auth_type $$12 ospf_auth_key",
"ospf_common_option : T_SECONDARY T_AUTH ospf_auth",
"ospf_common_option : T_AUTHKEY ospf_auth_key",
"ospf_nonbroadcast_option : ospf_common_option",
"ospf_nonbroadcast_option : T_POLLINTERVAL time",
"ospf_nonbroadcast_option : T_ROUTERS '{' ospf_router_eligible_list '}'",
"ospf_cost_option :",
"ospf_cost_option : T_METRIC ospf_cost",
"ospf_cost : metric",
"ospf_type_option :",
"ospf_type_option : T_TYPE ospf_type",
"ospf_type : L_NUMBER",
"ospf_area_option : T_AREA ospf_area",
"ospf_area_option : T_BACKBONE",
"ospf_area : number",
"ospf_md5_auth : ospf_md5_key",
"ospf_md5_auth : '{' ospf_md5_keys '}'",
"ospf_md5_keys : ospf_md5_key L_EOS",
"ospf_md5_keys : ospf_md5_keys ospf_md5_key L_EOS",
"$$13 :",
"ospf_md5_key : T_KEY $$13 ospf_auth_key T_ID L_NUMBER ospf_md5_time_ranges",
"ospf_md5_time_ranges :",
"ospf_md5_time_ranges : '{' ospf_md5_time_range_values '}'",
"ospf_md5_time_range_values : ospf_md5_time_range_value L_EOS",
"ospf_md5_time_range_values : ospf_md5_time_range_values ospf_md5_time_range_value L_EOS",
"ospf_md5_time_range_value : T_START_ACCEPT ospf_md5_datetime",
"ospf_md5_time_range_value : T_STOP_ACCEPT ospf_md5_datetime",
"ospf_md5_time_range_value : T_START_GENERATE ospf_md5_datetime",
"ospf_md5_time_range_value : T_STOP_GENERATE ospf_md5_datetime",
"ospf_md5_datetime : L_NUMBER '/' L_NUMBER '/' L_NUMBER L_NUMBER ':' L_NUMBER",
"$$14 :",
"ospf_auth : ospf_simple_auth_type $$14 ospf_auth_key",
"ospf_simple_auth_type : T_NONE",
"ospf_simple_auth_type : T_SIMPLE",
"ospf_auth_type : T_NONE",
"ospf_auth_type : T_SIMPLE",
"ospf_auth_type : T_MD5",
"ospf_auth_key : string",
"ospf_auth_key : L_NUMBER",
"ospf_auth_key : byte_string",
"ospf_stub_list : ospf_stub L_EOS",
"ospf_stub_list : ospf_stub_list ospf_stub L_EOS",
"ospf_stub_list : ospf_stub_list error L_EOS",
"ospf_stub : host T_METRIC ospf_cost",
"ospf_router_eligible_list : ospf_router_eligible L_EOS",
"ospf_router_eligible_list : ospf_router_eligible_list ospf_router_eligible L_EOS",
"ospf_router_eligible_list : ospf_router_eligible_list error L_EOS",
"ospf_router_eligible : gateway_inet ospf_eligible_option",
"ospf_eligible_option :",
"ospf_eligible_option : T_ELIGIBLE",
"ospf_priority : L_NUMBER",
"ospf_prop_init : T_OSPF",
"ospf_ase_prop_init : T_OSPF_ASE",
"ospf_trace_options_none : T_NONE",
"ospf_trace_options_none : ospf_trace_options",
"ospf_trace_options_none : ospf_trace_options T_EXCEPT ospf_trace_options",
"ospf_trace_options : ospf_trace_option",
"ospf_trace_options : ospf_trace_options ospf_trace_option",
"ospf_trace_option : trace_option",
"ospf_trace_option : trace_option_detail trace_option_sendrecv ospf_trace_option_index",
"ospf_trace_option : T_LSA_BLD",
"ospf_trace_option : T_LSA_TX",
"ospf_trace_option : T_LSA_RX",
"ospf_trace_option : T_SPF",
"ospf_trace_option : T_TRAP",
"ospf_trace_option : T_DEBUG",
"ospf_trace_option_index : T_PACKETS",
"ospf_trace_option_index : T_HELLO",
"ospf_trace_option_index : T_DD",
"ospf_trace_option_index : T_REQUEST",
"ospf_trace_option_index : T_UPDATE",
"ospf_trace_option_index : T_ACK",
"ospf_area_range_list :",
"ospf_area_range_list : ospf_area_range_list ospf_area_range ospf_area_range_option L_EOS",
"ospf_area_range_list : ospf_area_range_list error L_EOS",
"ospf_area_range : host_mask_inet",
"ospf_area_range : network_mask_inet",
"ospf_area_range_option :",
"ospf_area_range_option : T_RESTRICT",
"$$15 :",
"bgp_statement : T_BGP $$15 onoff_option bgp_group",
"bgp_group :",
"bgp_group : '{' bgp_group_stmts '}'",
"bgp_group_stmts :",
"bgp_group_stmts : bgp_group_stmts bgp_group_stmt L_EOS",
"bgp_group_stmts : bgp_group_stmts error L_EOS",
"bgp_group_stmt : T_PREFERENCE preference",
"bgp_group_stmt : T_PREFERENCE2 preference",
"bgp_group_stmt : T_DEFAULTMETRIC metric",
"bgp_group_stmt : T_TRACEOPTIONS trace_file_option trace_control_option bgp_trace_options_none",
"$$16 :",
"$$17 :",
"bgp_group_stmt : T_GROUP T_TYPE bgp_linktype T_PEERAS as bgp_proto_option bgp_interface $$16 bgp_peer_options $$17 '{' bgp_peer_stmts '}'",
"bgp_peer_options :",
"bgp_peer_options : bgp_peer_options bgp_peer_option",
"bgp_peer_stmts :",
"bgp_peer_stmts : bgp_peer_stmts bgp_peer_stmt L_EOS",
"bgp_peer_stmts : bgp_peer_stmts error L_EOS",
"$$18 :",
"bgp_peer_stmt : T_NEIGHBOR $$18 host bgp_peer_options",
"$$19 :",
"bgp_peer_stmt : T_ALLOW $$19 '{' bgp_allow_list '}'",
"bgp_peer_option : T_METRICOUT metric",
"bgp_peer_option : T_LOCALAS as",
"bgp_peer_option : T_NOGENDEFAULT",
"bgp_peer_option : T_GATEWAY gateway_inet",
"bgp_peer_option : T_PREFERENCE preference",
"bgp_peer_option : T_PREFERENCE2 preference",
"bgp_peer_option : T_LCLADDR interface_local",
"bgp_peer_option : T_HOLDTIME time",
"bgp_peer_option : T_TRACEOPTIONS trace_file_option trace_control_option bgp_trace_options_none",
"bgp_peer_option : T_VERSION L_NUMBER",
"bgp_peer_option : T_PASSIVE",
"bgp_peer_option : T_KEEP T_ALL",
"bgp_peer_option : T_KEEP T_NONE",
"bgp_peer_option : T_ANALRETENTIVE",
"bgp_peer_option : T_KEEPALIVESALWAYS",
"bgp_peer_option : T_NOAGGRID",
"bgp_peer_option : T_NOAUTHCHECK",
"bgp_peer_option : T_SETPREF preference",
"bgp_peer_option : T_IGNOREFIRSTASHOP",
"bgp_peer_option : T_SENDBUF bgp_kbufsize",
"bgp_peer_option : T_RECVBUF bgp_kbufsize",
"bgp_peer_option : T_INDELAY time",
"bgp_peer_option : T_OUTDELAY time",
"bgp_peer_option : T_TTL inet_ttl",
"bgp_peer_option : T_NOV4ASLOOP",
"bgp_peer_option : T_V3ASLOOPOKAY",
"bgp_peer_option : T_LOGUPDOWN",
"bgp_kbufsize : number",
"bgp_linktype : T_INTERNAL",
"bgp_linktype : T_EXTERNAL",
"bgp_linktype : T_IGP",
"bgp_linktype : T_ROUTING",
"bgp_linktype : T_TEST",
"bgp_trace_options_none : T_NONE",
"bgp_trace_options_none : bgp_trace_options",
"bgp_trace_options_none : bgp_trace_options T_EXCEPT bgp_trace_options",
"bgp_trace_options : bgp_trace_option",
"bgp_trace_options : bgp_trace_options bgp_trace_option",
"bgp_proto_option :",
"bgp_proto_option : T_PROTO bgp_proto",
"bgp_proto : T_NOP_PROTO",
"bgp_proto : T_RIP",
"bgp_proto : T_HELLO",
"bgp_proto : T_OSPF",
"bgp_interface :",
"bgp_interface : T_INTERFACE interface_all_list",
"bgp_trace_option : trace_option",
"bgp_trace_option : T_ASPATH",
"bgp_trace_option : trace_option_detail trace_option_sendrecv bgp_trace_option_index",
"bgp_trace_option_index : T_PACKETS",
"bgp_trace_option_index : T_OPEN",
"bgp_trace_option_index : T_UPDATE",
"bgp_trace_option_index : T_KEEPALIVE",
"bgp_prop_init : T_BGP",
"bgp_allow_list :",
"bgp_allow_list : bgp_allow_list bgp_allow L_EOS",
"bgp_allow_list : bgp_allow_list error L_EOS",
"bgp_allow : all_mask_inet",
"bgp_allow : host_mask_inet",
"bgp_allow : network_mask_inet",
"$$20 :",
"icmp_statement : T_ICMP $$20 '{' icmp_group_stmts '}'",
"icmp_group_stmts :",
"icmp_group_stmts : icmp_group_stmts icmp_group_stmt L_EOS",
"icmp_group_stmts : icmp_group_stmts error L_EOS",
"icmp_group_stmt : T_TRACEOPTIONS trace_file_option trace_control_option icmp_trace_options_none",
"icmp_trace_options_none : T_NONE",
"icmp_trace_options_none : icmp_trace_options",
"icmp_trace_options_none : icmp_trace_options T_EXCEPT icmp_trace_options",
"icmp_trace_options : icmp_trace_option",
"icmp_trace_options : icmp_trace_options icmp_trace_option",
"icmp_trace_option : trace_option",
"icmp_trace_option : trace_option_detail trace_option_sendrecv icmp_trace_option_index",
"icmp_trace_option_index : T_PACKETS",
"icmp_trace_option_index : T_REDIRECT",
"icmp_trace_option_index : T_ROUTERDISCOVERY",
"icmp_trace_option_index : T_INFO",
"icmp_trace_option_index : T_ERROR",
"$$21 :",
"redirect_statement : T_REDIRECT $$21 onoff_option redirect_group",
"redirect_group :",
"redirect_group : '{' redirect_group_stmts '}'",
"redirect_group_stmts :",
"redirect_group_stmts : redirect_group_stmts redirect_group_stmt L_EOS",
"redirect_group_stmts : redirect_group_stmts error L_EOS",
"redirect_group_stmt : T_PREFERENCE preference",
"redirect_group_stmt : T_INTERFACE interface_all_list redirect_interface_options",
"redirect_group_stmt : T_TRUSTEDGATEWAYS gateway_list",
"redirect_group_stmt : T_TRACEOPTIONS trace_file_option trace_control_option redirect_trace_options_none",
"redirect_interface_options : redirect_interface_option",
"redirect_interface_options : redirect_interface_options redirect_interface_option",
"redirect_interface_option : T_REDIRECTS",
"redirect_interface_option : T_NOREDIRECTS",
"redirect_trace_options_none : T_NONE",
"redirect_trace_options_none : redirect_trace_options",
"redirect_trace_options_none : redirect_trace_options T_EXCEPT redirect_trace_options",
"redirect_trace_options : redirect_trace_option",
"redirect_trace_options : redirect_trace_options redirect_trace_option",
"redirect_trace_option : trace_option",
"snmp_statement : T_SNMP T_OFF",
"route_order :",
"$$22 :",
"route_statement : T_STATIC $$22 '{' route_stmts '}'",
"route_stmts : route_stmt L_EOS",
"route_stmts : route_stmts route_stmt L_EOS",
"route_stmts : error L_EOS",
"route_stmt : route_dest T_INTERFACE interface_all preference_option rts_options",
"route_stmt : route_dest T_GATEWAY gateway_list interface_list_option route_aspath preference_option rts_options",
"route_aspath :",
"route_aspath : T_AS as",
"route_dest : T_NOP_DM",
"route_dest : host_mask_inet",
"route_dest : network_mask_inet",
"route_dest : T_DEFAULT",
"rts_options :",
"rts_options : rts_options rts_option",
"rts_option : T_RETAIN",
"rts_option : T_REJECT",
"rts_option : T_BLACKHOLE",
"rts_option : T_NOINSTALL",
"control_order :",
"control_statement : T_NOP1",
"control_statement : T_IMPORT T_PROTO control_exterior T_AS as import_option_inet L_EOS",
"control_statement : T_IMPORT T_PROTO T_BGP aspath_match import_option_inet L_EOS",
"control_statement : T_IMPORT T_PROTO import_interior interface_gateway_list import_option_inet L_EOS",
"control_statement : T_IMPORT T_PROTO T_OSPF_ASE tag_option import_option_inet L_EOS",
"control_statement : T_EXPORT T_PROTO control_exterior T_AS as prop_source_option_inet L_EOS",
"control_statement : T_EXPORT T_PROTO export_interior interface_gateway_list prop_source_option_inet L_EOS",
"control_statement : T_EXPORT T_PROTO export_ospf_ase ospf_type_option tag_as_option prop_source_option_inet L_EOS",
"control_statement : T_AGGREGATE aggregate_dest_inet preference_option aggregate_options '{' aggregate_list_inet '}' L_EOS",
"control_statement : T_GENERATE aggregate_dest_inet preference_option '{' aggregate_list_inet '}' L_EOS",
"aggregate_options :",
"aggregate_options : T_BRIEF",
"aggregate_dest_inet : network_mask_inet",
"aggregate_dest_inet : default_mask_inet",
"aggregate_list_inet : aggregate_proto_inet",
"aggregate_list_inet : aggregate_list_inet aggregate_proto_inet",
"aggregate_list_inet : aggregate_list_inet error L_EOS",
"aggregate_proto_inet : T_PROTO aggregate_protos_inet import_option_inet L_EOS",
"aggregate_proto_inet : T_PROTO aggregate_protos_inet T_AS as import_option_inet L_EOS",
"aggregate_proto_inet : T_PROTO aggregate_protos_inet aspath_match import_option_inet L_EOS",
"aggregate_proto_inet : T_PROTO aggregate_protos_inet T_TAG tag import_option_inet L_EOS",
"aggregate_protos_inet : T_ALL",
"aggregate_protos_inet : T_DIRECT",
"aggregate_protos_inet : T_STATIC",
"aggregate_protos_inet : T_KERNEL",
"aggregate_protos_inet : T_AGGREGATE",
"aggregate_protos_inet : T_HELLO",
"aggregate_protos_inet : T_RIP",
"aggregate_protos_inet : T_OSPF",
"aggregate_protos_inet : T_OSPF_ASE",
"aggregate_protos_inet : T_BGP",
"control_exterior : proto_exterior",
"import_listen_option :",
"import_listen_option : T_LISTEN",
"import_interior : proto_interior",
"import_option_inet : T_RESTRICT",
"import_option_inet : preference_option '{' import_list_inet '}'",
"import_list_inet :",
"import_list_inet : import_list_inet import_listen_inet L_EOS",
"import_list_inet : import_list_inet error L_EOS",
"$$23 :",
"import_listen_inet : import_listen_option $$23 dest_mask_inet preference_restrict_option",
"export_interior : proto_interior",
"export_ospf_ase : T_OSPF_ASE",
"prop_announce_option :",
"prop_announce_option : T_ANNOUNCE",
"prop_source_option_inet : T_RESTRICT",
"prop_source_option_inet : metric_option '{' prop_source_list_inet '}'",
"prop_source_list_inet :",
"prop_source_list_inet : prop_source_list_inet prop_source_inet L_EOS",
"prop_source_list_inet : prop_source_list_inet error L_EOS",
"prop_source_inet : T_PROTO prop_direct interface_list_option prop_metric_option_inet",
"prop_source_inet : T_PROTO prop_static interface_list_option prop_metric_option_inet",
"prop_source_inet : T_PROTO prop_aggregate prop_metric_option_inet",
"prop_source_inet : T_PROTO prop_kernel interface_list_option prop_metric_option_inet",
"prop_source_inet : T_PROTO prop_default prop_metric_option_inet",
"prop_source_inet : T_PROTO proto_exterior T_AS as prop_metric_option_inet",
"prop_source_inet : T_PROTO aspath_prop_init aspath_match prop_metric_option_inet",
"prop_source_inet : T_PROTO tag_prop_init T_TAG tag prop_metric_option_inet",
"prop_source_inet : T_PROTO rip_prop_init interface_gateway_list prop_metric_option_inet",
"prop_source_inet : T_PROTO hello_prop_init interface_gateway_list prop_metric_option_inet",
"prop_source_inet : T_PROTO ospf_prop_init prop_metric_option_inet",
"prop_source_inet : T_PROTO ospf_ase_prop_init prop_metric_option_inet",
"prop_direct : T_DIRECT",
"prop_static : T_STATIC",
"prop_aggregate : T_AGGREGATE",
"prop_kernel : T_KERNEL",
"prop_default : T_DEFAULT",
"prop_metric_option_inet : T_RESTRICT",
"prop_metric_option_inet : metric_option prop_restrict_list_option_inet",
"prop_restrict_list_option_inet :",
"prop_restrict_list_option_inet : '{' prop_restrict_list_inet '}'",
"prop_restrict_list_inet :",
"prop_restrict_list_inet : prop_restrict_list_inet prop_restrict_inet L_EOS",
"prop_restrict_list_inet : prop_restrict_list_inet error L_EOS",
"$$24 :",
"prop_restrict_inet : prop_announce_option $$24 dest_mask_inet metric_restrict_option",
"dest_mask_inet : all_mask_inet",
"dest_mask_inet : default_mask_inet",
"dest_mask_inet : host_mask_inet",
"dest_mask_inet : network_mask_flags_inet",
"all_mask_inet : T_ALL",
"default_mask_inet : T_DEFAULT",
"gateway_list_inet : gateway_inet",
"gateway_list_inet : gateway_list_inet gateway_inet",
"gateway_inet : host_inet",
"host_inet : host_name",
"host_inet : addr_inet",
"host_mask_inet : T_HOST host_inet",
"network_inet : host_name",
"network_inet : addr_inet",
"network_mask_inet : network_inet",
"network_mask_inet : network_inet T_MASKLEN number",
"network_mask_inet : network_inet T_MASK mask_inet",
"network_mask_flags_inet : network_mask_inet dm_flags",
"mask_inet : in_addr",
"addr_inet : in_addr",
"in_addr : L_NUMBER",
"in_addr : byte_string",
"inet_ttl : L_NUMBER",
"host : T_NOP_SOCKADDR",
"host : host_inet",
"mask : T_NOP_SOCKADDR",
"mask : mask_inet",
"addr : T_NOP_SOCKADDR",
"addr : addr_inet",
"gateway_list : T_NOP_ADV",
"gateway_list : gateway_list_inet",
"dm_flags :",
"dm_flags : T_EXACT",
"dm_flags : T_REFINE",
"host_name : L_HNAME",
"host_name : L_KEYWORD",
"host_name : string",
"proto_interior : T_REDIRECT",
"proto_interior : T_HELLO",
"proto_interior : T_RIP",
"proto_exterior : T_NOP_PROTO",
"proto_exterior : bgp_prop_init",
"onoff_option : T_ON",
"onoff_option : T_OFF",
"metric : number",
"metric : T_INFINITY",
"metric_option :",
"metric_option : T_METRIC metric",
"metric_restrict_option : metric_option",
"metric_restrict_option : T_RESTRICT",
"preference_option :",
"preference_option : T_PREFERENCE preference",
"preference_restrict_option : preference_option",
"preference_restrict_option : T_RESTRICT",
"preference : L_NUMBER",
"string : L_STRING",
"neg_number : L_NUMBER",
"neg_number : '-' L_NUMBER",
"number : L_NUMBER",
"number : byte_string",
"byte_string : L_HEXSTRING",
"byte_string : L_OCTETSTRING",
"time : L_NUMBER",
"time : L_NUMBER ':' L_NUMBER",
"time : L_NUMBER ':' L_NUMBER ':' L_NUMBER",
"as_loop_option :",
"as_loop_option : T_LOOPS L_NUMBER",
"as : L_NUMBER",
"tag_as_option :",
"tag_as_option : T_TAG tag_as",
"tag_as : tag",
"tag_as : T_AS",
"tag_as : T_AS number",
"tag_option :",
"tag_option : T_TAG tag",
"tag : number",
"tag_prop_init : proto_exterior",
"tag_prop_init : rip_prop_init",
"tag_prop_init : hello_prop_init",
"tag_prop_init : ospf_prop_init",
"tag_prop_init : ospf_ase_prop_init",
"aspath_prop_init : T_ALL",
"aspath_prop_init : proto_exterior",
"aspath_prop_init : rip_prop_init",
"aspath_prop_init : hello_prop_init",
"aspath_prop_init : ospf_prop_init",
"aspath_prop_init : ospf_ase_prop_init",
"$$25 :",
"aspath_match : T_ASPATH $$25 aspath_regex T_ORIGIN aspath_origin",
"aspath_origin :",
"aspath_origin : T_ANY",
"aspath_origin : aspath_origin_list",
"aspath_origin_list : aspath_origins",
"aspath_origin_list : aspath_origin_list '|' aspath_origins",
"aspath_origins : T_IGP",
"aspath_origins : T_EGP",
"aspath_origins : T_INCOMPLETE",
"aspath_regex : aspath_sum",
"aspath_regex : aspath_regex '|' aspath_sum",
"aspath_sum : aspath_term",
"aspath_sum : aspath_sum aspath_term",
"aspath_term : aspath_symbol",
"aspath_term : aspath_symbol '*'",
"aspath_term : aspath_symbol '+'",
"aspath_term : aspath_symbol '?'",
"aspath_term : aspath_symbol '{' aspath_range '}'",
"aspath_symbol : as",
"aspath_symbol : '.'",
"aspath_symbol : '(' aspath_regex ')'",
"aspath_range : number",
"aspath_range : number ','",
"aspath_range : number ',' number",
};
#endif
#define yyclearin (yychar=(-1))
#define yyerrok (yyerrflag=0)
#ifndef YYSTACKSIZE
#ifdef YYMAXDEPTH
#define YYSTACKSIZE YYMAXDEPTH
#else
#define YYSTACKSIZE 300
#endif
#endif
int yydebug;
int yynerrs;
int yyerrflag;
int yychar;
short *yyssp;
YYSTYPE *yyvsp;
YYSTYPE yyval;
YYSTYPE yylval;
#define yystacksize YYSTACKSIZE
short yyss[YYSTACKSIZE];
YYSTYPE yyvs[YYSTACKSIZE];
#line 5746 "parser.y"

/*
 *	Log any parsing errors
 */
static void
yyerror(s)
const char *s;
{
    byte *cp;

    tracef("parse: %s ",
	   parse_where());

    switch (yychar) {
    case L_HEXSTRING:
	tracef("%s at '",
	       s);
	for (cp = yylval.hex.ptr; (cp - yylval.hex.ptr) < yylval.hex.len; cp++) {
	    tracef("%02x",
		   *cp);
	}
	tracef("'");
	break;	

    case L_OCTETSTRING:
	tracef("%s at '",
	       s);
	for (cp = yylval.hex.ptr; (cp - yylval.hex.ptr) < yylval.hex.len; cp++) {
	    tracef("%s%u",
		   (cp == yylval.hex.ptr) ? "" : ".",
		   *cp);
	}
	tracef("'");
	break;	

    case L_NUMBER:
	tracef("%s at '%d'",
	       s,
	       yylval.num);
	break;

    case L_EOS:
	tracef("%s at 'end-of-statement'",
	       s);
	break;

    case L_ERROR:
	tracef("%s",
	       parse_error);
	break;

#ifdef	YYEOF
    case YYEOF:
#else	/* YYEOF */
    case 0:
#endif	/* YYEOF */
	tracef("%s at 'end-of-file'",
	       s);
	break;

    default:
	if (!parse_keyword_lookup(yychar)) {
#if	YYDEBUG != 0
#ifdef	YYTRANSLATE
	    tracef("%s at '%s'",
		   s,
		   yytname[YYTRANSLATE(yychar)]);
#else	/* YYTRANSLATE */
	    tracef("%s at '%s'",
		   s,
		   yyname[yychar]);
#endif	/* YYTRANSLATE */
#else	/* YYDEBUG */
	    tracef("%s",
		   s);
#endif	/* YYDEBUG */
	    break;
	}
	tracef("%s at '%s' (%u) length %d",
	       s,
	       yylval.str.ptr,
	       yylval.str.type,
	       yylval.str.len);
	break;

    case L_STRING:
    case L_HNAME:
    case L_KEYWORD:
	tracef("%s at '%s' length %d",
	       s,
	       yylval.str.ptr,
	       yylval.str.len);
	break;

#ifdef	YYEMPTY
    case YYEMPTY:
	tracef("%s",
	       s);
	break;
#endif	/* YYEMPTY */
    }
    trace_log_tf(trace_global,
		 TRC_NL_BEFORE|TRC_NL_AFTER,
		 LOG_ERR,
		 (NULL));
}


#line 2432 "y.tab.c"
#define YYABORT goto yyabort
#define YYACCEPT goto yyaccept
#define YYERROR goto yyerrlab
int
yyparse()
{
    register int yym, yyn, yystate;
#if YYDEBUG
    register char *yys;
    extern char *getenv();

    if (yys = getenv("YYDEBUG"))
    {
        yyn = *yys;
        if (yyn >= '0' && yyn <= '9')
            yydebug = yyn - '0';
    }
#endif

    yynerrs = 0;
    yyerrflag = 0;
    yychar = (-1);

    yyssp = yyss;
    yyvsp = yyvs;
    *yyssp = yystate = 0;

yyloop:
    if (yyn = yydefred[yystate]) goto yyreduce;
    if (yychar < 0)
    {
        if ((yychar = yylex()) < 0) yychar = 0;
#if YYDEBUG
        if (yydebug)
        {
            yys = 0;
            if (yychar <= YYMAXTOKEN) yys = yyname[yychar];
            if (!yys) yys = "illegal-symbol";
            printf("yydebug: state %d, reading %d (%s)\n", yystate,
                    yychar, yys);
        }
#endif
    }
    if ((yyn = yysindex[yystate]) && (yyn += yychar) >= 0 &&
            yyn <= YYTABLESIZE && yycheck[yyn] == yychar)
    {
#if YYDEBUG
        if (yydebug)
            printf("yydebug: state %d, shifting to state %d\n",
                    yystate, yytable[yyn]);
#endif
        if (yyssp >= yyss + yystacksize - 1)
        {
            goto yyoverflow;
        }
        *++yyssp = yystate = yytable[yyn];
        *++yyvsp = yylval;
        yychar = (-1);
        if (yyerrflag > 0)  --yyerrflag;
        goto yyloop;
    }
    if ((yyn = yyrindex[yystate]) && (yyn += yychar) >= 0 &&
            yyn <= YYTABLESIZE && yycheck[yyn] == yychar)
    {
        yyn = yytable[yyn];
        goto yyreduce;
    }
    if (yyerrflag) goto yyinrecovery;
#ifdef lint
    goto yynewerror;
#endif
yynewerror:
    yyerror("syntax error");
#ifdef lint
    goto yyerrlab;
#endif
yyerrlab:
    ++yynerrs;
yyinrecovery:
    if (yyerrflag < 3)
    {
        yyerrflag = 3;
        for (;;)
        {
            if ((yyn = yysindex[*yyssp]) && (yyn += YYERRCODE) >= 0 &&
                    yyn <= YYTABLESIZE && yycheck[yyn] == YYERRCODE)
            {
#if YYDEBUG
                if (yydebug)
                    printf("yydebug: state %d, error recovery shifting\
 to state %d\n", *yyssp, yytable[yyn]);
#endif
                if (yyssp >= yyss + yystacksize - 1)
                {
                    goto yyoverflow;
                }
                *++yyssp = yystate = yytable[yyn];
                *++yyvsp = yylval;
                goto yyloop;
            }
            else
            {
#if YYDEBUG
                if (yydebug)
                    printf("yydebug: error recovery discarding state %d\n",
                            *yyssp);
#endif
                if (yyssp <= yyss) goto yyabort;
                --yyssp;
                --yyvsp;
            }
        }
    }
    else
    {
        if (yychar == 0) goto yyabort;
#if YYDEBUG
        if (yydebug)
        {
            yys = 0;
            if (yychar <= YYMAXTOKEN) yys = yyname[yychar];
            if (!yys) yys = "illegal-symbol";
            printf("yydebug: state %d, error recovery discards token %d (%s)\n",
                    yystate, yychar, yys);
        }
#endif
        yychar = (-1);
        goto yyloop;
    }
yyreduce:
#if YYDEBUG
    if (yydebug)
        printf("yydebug: state %d, reducing by rule %d (%s)\n",
                yystate, yyn, yyrule[yyn]);
#endif
    yym = yylen[yyn];
    yyval = yyvsp[1-yym];
    switch (yyn)
    {
case 13:
#line 441 "parser.y"
{
				yyerrok;
			}
break;
case 15:
#line 450 "parser.y"
{
#if	YYDEBUG != 0
			    if (yyvsp[-1].num  == T_OFF) {
				yydebug = 0;
			    } else {
				yydebug = 1;
			    }
#endif	/* YYDEBUG */
			}
break;
case 16:
#line 460 "parser.y"
{
#if	YYDEBUG != 0
			    if (yyvsp[-1].num  < 0 || yyvsp[-1].num  > PS_MAX) {
				(void) sprintf(parse_error, "invalid yystate value: %d",
					       yyvsp[-1].num );
				PARSE_ERROR;
			    }
			    parse_state = yyvsp[-1].num ;
#endif	/* YYDEBUG */
			}
break;
case 17:
#line 471 "parser.y"
{
#if	YYDEBUG != 0
			    task_quit(0);
#endif	/* YYDEBUG */
			}
break;
case 18:
#line 481 "parser.y"
{
			    if (yyvsp[-3].trace_file  && !BIT_TEST(task_state, TASKS_TEST)) {

				trace_store(trace_global, yyvsp[-1].flag , yyvsp[-2].flag , yyvsp[-3].trace_file , (bits *) 0);

				if (yyvsp[-1].flag ) {
				    /* Turn on or change tracing */

				    trace_on(yyvsp[-3].trace_file );
				} else {
				    /* Turn off tracing */

				    trace_off(yyvsp[-3].trace_file );
				}
				trace_display(trace_global, TR_ALL);
			    }
			}
break;
case 19:
#line 502 "parser.y"
{
			    yyval.trace_file  = trace_file_locate((char *) 0,
						   (off_t) 0,
						   (u_int) 0,
						   (flag_t) 0);
			}
break;
case 20:
#line 509 "parser.y"
{
			    yyval.trace_file  = trace_file_locate(yyvsp[-1].str .ptr,
						   (off_t) 0,
						   (u_int) 0,
						   yyvsp[0].flag );
			}
break;
case 21:
#line 516 "parser.y"
{
			    if (parse_limit_check("file-size",
						  (u_int) yyvsp[-2].num ,
						  TRACE_LIMIT_FILE_SIZE)) {
				PARSE_ERROR;
			    }

			    if (parse_limit_check("files",
						  (u_int) yyvsp[0].num ,
						  TRACE_LIMIT_FILE_COUNT)) {
				PARSE_ERROR;
			    }

			    yyval.trace_file  = trace_file_locate(yyvsp[-5].str .ptr,
						   yyvsp[-2].num ,
						   (u_int) yyvsp[0].num ,
						   yyvsp[-4].flag );
			}
break;
case 23:
#line 538 "parser.y"
{
			    yyval.num  = yyvsp[-1].num  * 1024;
			}
break;
case 24:
#line 542 "parser.y"
{
			    yyval.num  = yyvsp[-1].num  * 1024 * 1024;
			}
break;
case 25:
#line 548 "parser.y"
{
			    yyval.flag  = 0;
			}
break;
case 26:
#line 552 "parser.y"
{
			    yyval.flag  = TRF_REPLACE;
			}
break;
case 27:
#line 559 "parser.y"
{
			    yyval.flag  = 0;
			}
break;
case 28:
#line 563 "parser.y"
{
			    yyval.flag  = yyvsp[-1].flag  | yyvsp[0].flag ;
			}
break;
case 29:
#line 570 "parser.y"
{
			    yyval.flag  = TRC_NOSTAMP;
			}
break;
case 30:
#line 577 "parser.y"
{
			    yyval.flag  = (flag_t) 0;
			}
break;
case 31:
#line 581 "parser.y"
{
			    yyval.flag  = yyvsp[0].flag ;
			}
break;
case 32:
#line 585 "parser.y"
{
			    yyval.flag  = yyvsp[-2].flag  & ~ yyvsp[-1].num ;
			}
break;
case 33:
#line 592 "parser.y"
{
			    yyval.flag  = yyvsp[0].flag ;
			}
break;
case 34:
#line 596 "parser.y"
{
			    yyval.flag  = yyvsp[0].flag ;
			}
break;
case 35:
#line 600 "parser.y"
{
			    yyval.flag  = yyvsp[-1].flag  | yyvsp[0].flag ;
			}
break;
case 36:
#line 604 "parser.y"
{
			    yyval.flag  = yyvsp[-1].flag  | yyvsp[0].flag ;
			}
break;
case 37:
#line 612 "parser.y"
{
			    yyval.flag  = TR_PARSE;
			}
break;
case 38:
#line 616 "parser.y"
{
#ifdef	TR_ADV
			    yyval.flag  = TR_ADV;
#else	/* TR_ADV */
			    yyval.flag  = 0;
#endif	/* TR_ADV */
			}
break;
case 39:
#line 627 "parser.y"
{
			    yyval.flag  = TR_ROUTE;
			}
break;
case 40:
#line 631 "parser.y"
{
			    yyval.flag  = TR_NORMAL;
			}
break;
case 41:
#line 635 "parser.y"
{
			    yyval.flag  = TR_GENERAL;
			}
break;
case 42:
#line 639 "parser.y"
{
			    yyval.flag  = TR_STATE;
			}
break;
case 43:
#line 643 "parser.y"
{
			    yyval.flag  = TR_POLICY;
			}
break;
case 44:
#line 647 "parser.y"
{
			    yyval.flag  = TR_TASK;
			}
break;
case 45:
#line 651 "parser.y"
{
			    yyval.flag  = TR_TIMER;
			}
break;
case 46:
#line 655 "parser.y"
{
			    yyval.flag  = TR_ALL;
			}
break;
case 47:
#line 662 "parser.y"
{
			    yyval.flag  = (flag_t) 0;
			}
break;
case 48:
#line 666 "parser.y"
{
			    yyval.flag  = TR_PARSE_DETAIL;
			}
break;
case 49:
#line 673 "parser.y"
{
			    yyval.flag  = (flag_t) 0;
			}
break;
case 50:
#line 677 "parser.y"
{
			    yyval.flag  = TR_PARSE_SEND;
			}
break;
case 51:
#line 681 "parser.y"
{
			    yyval.flag  = TR_PARSE_RECV;
			}
break;
case 52:
#line 690 "parser.y"
{
				if (parse_new_state(PS_DEFINE)) {
					PARSE_ERROR;
				}
			}
break;
case 54:
#line 700 "parser.y"
{
			    parse_contig_masks = FALSE;
			}
break;
case 55:
#line 704 "parser.y"
{
			    parse_contig_masks = TRUE;

			    SI_MARTIANS(AF_INET) = yyvsp[-2].adv ;
			}
break;
case 56:
#line 710 "parser.y"
{
			    if (inet_autonomous_system) {
				(void) sprintf(parse_error, "autonomous-system already specified");
				PARSE_ERROR;
			    }
			    inet_autonomous_system = yyvsp[-2].as ;
#ifdef	PROTO_ASPATHS
			    aslocal_set(inet_autonomous_system, yyvsp[-1].u_num );
#endif	/* PROTO_ASPATHS */
			}
break;
case 57:
#line 721 "parser.y"
{
			    if (inet_parse_routerid(yyvsp[-1].sockaddr , parse_error)) {
				PARSE_ERROR;
			    }
			}
break;
case 58:
#line 730 "parser.y"
{
			    yyval.adv  = SI_MARTIANS(AF_INET);
			}
break;
case 59:
#line 734 "parser.y"
{
			    adv_entry *adv = adv_alloc(ADVFT_DM | yyvsp[-1].flag , (proto_t) 0);

			    adv->adv_dm = yyvsp[-2].dm ;
			    yyval.adv  = adv_destmask_insert(parse_error, yyvsp[-3].adv , adv);
			    if (!yyval.adv ) {
				PARSE_ERROR;
			    }
			}
break;
case 60:
#line 744 "parser.y"
{
			    yyerrok;
			}
break;
case 64:
#line 755 "parser.y"
{
			    yyval.flag  = ADVF_NO;
			}
break;
case 65:
#line 759 "parser.y"
{
			    yyval.flag  = (flag_t) 0;
			}
break;
case 66:
#line 767 "parser.y"
{
				if (parse_new_state(PS_OPTIONS)) {
					PARSE_ERROR;
				}
			}
break;
case 70:
#line 784 "parser.y"
{
			    task_newstate(TASKS_NOSEND, 0);
			}
break;
case 71:
#line 788 "parser.y"
{
			    task_newstate(TASKS_NORESOLV, 0);
			}
break;
case 72:
#line 792 "parser.y"
{
#ifdef	LOG_DAEMON
			    (void) setlogmask(yyvsp[0].num );
#else	/* LOG_DAEMON */
			    (void) sprintf(parse_error, "system does not support setlogmask()");
			    PARSE_ERROR;
#endif	/* LOG_DAEMON */
			}
break;
case 73:
#line 801 "parser.y"
{
			    task_mark_interval = yyvsp[0].time ;
			}
break;
case 74:
#line 805 "parser.y"
{
			    rt_default_needed = TRUE;
			    if (PARSE_METRIC_ISSET(&yyvsp[-1].metric )) {
				rt_default_rtparms.rtp_preference = yyvsp[-1].metric .metric;
			    }
			    rt_default_rtparms.rtp_router = yyvsp[0].sockaddr ;
			}
break;
case 75:
#line 817 "parser.y"
{
			    yyval.sockaddr  = (sockaddr_un *) 0;
			}
break;
case 76:
#line 821 "parser.y"
{
			    parse_gwlist = &parse_gwp;
			}
break;
case 77:
#line 825 "parser.y"
{
			    yyval.sockaddr  = sockdup(yyvsp[0].gwp ->gw_addr);
			    gw_freelist(*parse_gwlist);
			    *parse_gwlist = parse_gwp = (gw_entry *) 0;
			}
break;
case 78:
#line 833 "parser.y"
{
			    yyval.time  = TIME_MARK;
			}
break;
case 79:
#line 837 "parser.y"
{
			    yyval.time  = yyvsp[0].time ;
			}
break;
case 80:
#line 843 "parser.y"
{
#ifdef	LOG_DAEMON
			    yyval.num  = LOG_MASK(yyvsp[0].num );
#endif	/* LOG_DAEMON */
			}
break;
case 81:
#line 849 "parser.y"
{
#ifdef	LOG_DAEMON
			    yyval.num  = LOG_UPTO(yyvsp[0].num );
#endif	/* LOG_DAEMON */
			}
break;
case 82:
#line 855 "parser.y"
{
#ifdef	LOG_DAEMON
			    yyval.num  = yyvsp[-1].num  | LOG_MASK(yyvsp[0].num );
#endif	/* LOG_DAEMON */
			}
break;
case 83:
#line 863 "parser.y"
{
			    yyval.num  = LOG_EMERG;
			}
break;
case 84:
#line 867 "parser.y"
{
			    yyval.num  = LOG_ALERT;
			}
break;
case 85:
#line 871 "parser.y"
{
			    yyval.num  = LOG_CRIT;
			}
break;
case 86:
#line 875 "parser.y"
{
			    yyval.num  = LOG_ERR;
			}
break;
case 87:
#line 879 "parser.y"
{
			    yyval.num  = LOG_WARNING;
			}
break;
case 88:
#line 883 "parser.y"
{
			    yyval.num  = LOG_NOTICE;
			}
break;
case 89:
#line 887 "parser.y"
{
			    yyval.num  = LOG_INFO;
			}
break;
case 90:
#line 891 "parser.y"
{
			    yyval.num  = LOG_DEBUG;
			}
break;
case 91:
#line 899 "parser.y"
{
			    if (parse_new_state(PS_INTERFACE)) {
				PARSE_ERROR;
			    }
			}
break;
case 92:
#line 908 "parser.y"
{
			    PROTO_SEEN(RTPROTO_DIRECT);
			}
break;
case 93:
#line 912 "parser.y"
{
			    parse_proto = (proto_t) 0;
			}
break;
case 96:
#line 921 "parser.y"
{
			    yyerrok;
			}
break;
case 98:
#line 929 "parser.y"
{
			    config_list *list = config_list_alloc(yyvsp[0].config , 0);

			    parse_adv_propagate_config(yyvsp[-1].adv , list, (proto_t) 0);

			    if (parse_adv_append(&int_policy, yyvsp[-1].adv )) {
				PARSE_ERROR;
			    }

			    config_list_free(list);
			}
break;
case 99:
#line 941 "parser.y"
{
			    bzero((caddr_t) &parse_ifi, sizeof (parse_ifi));

			    parse_ifi.ifi_addr = yyvsp[0].sockaddr ;
			}
break;
case 100:
#line 947 "parser.y"
{
			    if (if_parse_add(&parse_ifi, parse_error)) {
				PARSE_ERROR;
			    }
			}
break;
case 103:
#line 961 "parser.y"
{
			    task_newstate(TASKS_STRICTIFS, 0);
			}
break;
case 104:
#line 965 "parser.y"
{
			    if (parse_limit_check("scan-interval",
						  (u_int) yyvsp[0].time ,
						  KRT_LIMIT_SCANTIMER)) {
				PARSE_ERROR;
			    }
			    
			    task_timer_set(krt_timer_ifcheck,
					   yyvsp[0].time ,
					   (time_t) 0);
			}
break;
case 105:
#line 980 "parser.y"
{
			    parse_ifi.ifi_addr_broadcast = yyvsp[0].sockaddr ;
			    BIT_SET(parse_ifi.ifi_state, IFS_BROADCAST);
			}
break;
case 106:
#line 985 "parser.y"
{
			    parse_ifi.ifi_addr_local = yyvsp[0].sockaddr ;
			    BIT_SET(parse_ifi.ifi_state, IFS_POINTOPOINT);
			}
break;
case 107:
#line 990 "parser.y"
{
			}
break;
case 110:
#line 1001 "parser.y"
{
			    parse_ifi.ifi_netmask = yyvsp[0].sockaddr ;
			}
break;
case 111:
#line 1005 "parser.y"
{
			    BIT_SET(parse_ifi.ifi_state, IFS_MULTICAST);
			}
break;
case 113:
#line 1013 "parser.y"
{
			    yyval.config  = config_append(yyvsp[-1].config , yyvsp[0].config );
			    if (!yyval.config ) {
				PARSE_ERROR;
			    }
			}
break;
case 114:
#line 1023 "parser.y"
{
			    yyval.config  = config_alloc(IF_CONFIG_PREFERENCE_UP, GS2A(yyvsp[0].pref ));
			}
break;
case 115:
#line 1027 "parser.y"
{
			    yyval.config  = config_alloc(IF_CONFIG_PREFERENCE_DOWN, GS2A(yyvsp[0].pref ));
			}
break;
case 116:
#line 1031 "parser.y"
{
			    yyval.config  = config_alloc(IF_CONFIG_PASSIVE, GS2A(TRUE));
			}
break;
case 117:
#line 1035 "parser.y"
{
			    yyval.config  = config_alloc(IF_CONFIG_SIMPLEX, GS2A(TRUE));
			}
break;
case 118:
#line 1039 "parser.y"
{
			    yyval.config  = config_alloc(IF_CONFIG_REJECT, GS2A(TRUE));
			}
break;
case 119:
#line 1043 "parser.y"
{
			    yyval.config  = config_alloc(IF_CONFIG_BLACKHOLE, GS2A(TRUE));
			}
break;
case 120:
#line 1047 "parser.y"
{
			    yyval.config  = config_alloc(IF_CONFIG_AS, GS2A(yyvsp[0].as ));
			}
break;
case 121:
#line 1054 "parser.y"
{
			    yyval.adv  = adv_alloc(ADVFT_ANY|ADVF_FIRST, (proto_t) 0);
			}
break;
case 122:
#line 1058 "parser.y"
{
			    yyval.adv  = yyvsp[0].adv ;
			}
break;
case 123:
#line 1065 "parser.y"
{
			    yyval.adv  = yyvsp[0].adv ;
			    parse_proto = (proto_t) 0;
			    parse_gwlist = (gw_entry **) 0;
			}
break;
case 124:
#line 1071 "parser.y"
{
			    yyval.adv  = yyvsp[0].adv ;
			    parse_proto = (proto_t) 0;
			    parse_gwlist = (gw_entry **) 0;
			}
break;
case 125:
#line 1080 "parser.y"
{
			    yyval.adv  = adv_alloc(ADVFT_ANY|ADVF_FIRST, (proto_t) 0);
			}
break;
case 127:
#line 1088 "parser.y"
{
			    BIT_SET(yyvsp[0].adv ->adv_flag, ADVF_FIRST);
			    yyval.adv  = yyvsp[0].adv ;
			}
break;
case 128:
#line 1093 "parser.y"
{
			    yyval.adv  = yyvsp[-1].adv ;
			    if (parse_adv_append(&yyval.adv , yyvsp[0].adv )) {
				PARSE_ERROR;
			    }
			}
break;
case 129:
#line 1102 "parser.y"
{
			    switch (yyvsp[0].adv ->adv_flag & ADVF_TYPE) {

			    case ADVFT_IFN:
				trace_tf(trace_global,
					 TR_PARSE,
					 0,
					 ("parse: %s INTERFACE: %A*",
					  parse_where(),
					  yyval.adv ->adv_ifn->ifae_addr));
				break;
				
			    case ADVFT_IFAE:
				trace_tf(trace_global,
					 TR_PARSE,
					 0,
					 ("parse: %s INTERFACE: %A",
					  parse_where(),
					  yyval.adv ->adv_ifae->ifae_addr));
			    }
			}
break;
case 130:
#line 1127 "parser.y"
{
			    int valid_ifname = 0;

			    switch (yyvsp[0].str .type) {
			    case L_KEYWORD:
			    case L_STRING:
				if (yyvsp[0].str .len < IFL_NAMELEN) {
				    valid_ifname++;
				}
				break;

			    default:
				break;
			    }

			    if (valid_ifname) {
				/* First try to look up as an interface name */

				yyval.adv  = if_parse_name(yyvsp[0].str .ptr, FALSE);
			    } else {
				/* Doesn't meet the criteria to be a name */

				yyval.adv  = (adv_entry *) 0;
			    }
			    
			    if (!yyval.adv  && !BIT_TEST(task_state, TASKS_NORESOLV)) {
				/* Try to resolve as a host name */

				sockaddr_un *addr = parse_addr_hostname(yyvsp[0].str .ptr, parse_error);

				if (addr) {
				    yyval.adv  = if_parse_address(addr);
				}

			    }

			    if (!yyval.adv ) {
				if (valid_ifname && !BIT_TEST(task_state, TASKS_STRICTIFS)) {
				    /* Assume it is a reference to an interface that does not exist */

				    yyval.adv  = if_parse_name(yyvsp[0].str .ptr, TRUE);
				} else {
				    /* Time to give up */
				    
				    (void) sprintf(parse_error, "Interface not found at '%s'",
						   yyvsp[0].str .ptr);
				    PARSE_ERROR;
				}
			    }
				
			    free_charstr(yyvsp[0].str );
			}
break;
case 131:
#line 1180 "parser.y"
{
			    yyval.adv  = if_parse_address(yyvsp[0].sockaddr );
			    if (!yyval.adv ) {
				(void) sprintf(parse_error, "Interface not found at '%A'",
					       yyvsp[0].sockaddr );
				PARSE_ERROR;
			    }

			    sockfree(yyvsp[0].sockaddr );
			}
break;
case 132:
#line 1193 "parser.y"
{
			    if_addr *ifap;
			    
			    ifap = if_withlcladdr(yyvsp[0].sockaddr , FALSE);
			    if (ifap) {
				yyval.ifae  = ifae_alloc(ifap->ifa_addrent_local);
			    } else {
				if (BIT_TEST(task_state, TASKS_STRICTIFS)) {
				    (void) sprintf(parse_error, "Invalid interface at '%A'",
						   yyvsp[0].sockaddr );
				    PARSE_ERROR;
				} else {
				    yyval.ifae  = ifae_locate(yyvsp[0].sockaddr , &if_local_list);
				}
			    }
			    sockfree(yyvsp[0].sockaddr );
			    trace_tf(trace_global,
				     TR_PARSE,
				     0,
				     ("parse: %s LOCAL ADDRESS: %A",
				      parse_where(),
				      yyval.ifae ->ifae_addr));
			}
break;
case 134:
#line 1220 "parser.y"
{
			    sockaddr_un *addr;

			    addr = parse_addr_hostname(yyvsp[0].str .ptr, parse_error);
			    if (!addr) {
				/* Name resolution failure */

				PARSE_ERROR;
			    }
			    yyval.sockaddr  = sockdup(addr);
			    free_charstr(yyvsp[0].str );
			}
break;
case 135:
#line 1240 "parser.y"
{
			    if (parse_new_state(PS_PROTO)) {
				PARSE_ERROR;
			    }
			}
break;
case 144:
#line 1261 "parser.y"
{
			    kernel_trace_options = trace_set_global(kernel_trace_types,
								    TR_KRT_SYMBOLS|TR_KRT_IFLIST);
			}
break;
case 149:
#line 1274 "parser.y"
{
			    trace_store(kernel_trace_options, yyvsp[0].flag , yyvsp[-1].flag , yyvsp[-2].trace_file , kernel_trace_types);
			}
break;
case 150:
#line 1277 "parser.y"
{}
break;
case 151:
#line 1279 "parser.y"
{
			    if (parse_limit_check("remnantholdtime",
						  (u_int) yyvsp[0].time ,
						  KRT_LIMIT_EXPIRE)) {
				PARSE_ERROR;
			    }
			    krt_t_expire = yyvsp[0].time ;
			}
break;
case 152:
#line 1287 "parser.y"
{}
break;
case 153:
#line 1288 "parser.y"
{}
break;
case 154:
#line 1290 "parser.y"
{
			    if (parse_limit_check("routes",
						  (u_int) yyvsp[0].num ,
						  KRT_LIMIT_ROUTES)) {
				PARSE_ERROR;
			    }
			    krt_limit_routes = yyvsp[0].num ;
			}
break;
case 157:
#line 1305 "parser.y"
{
			    BIT_SET(krt_options, KRT_OPT_NOCHANGE);
			}
break;
case 158:
#line 1309 "parser.y"
{
			    BIT_SET(krt_options, KRT_OPT_NOFLUSH);
			}
break;
case 161:
#line 1321 "parser.y"
{
			    krt_flash_routes = yyvsp[0].flag ;
			}
break;
case 162:
#line 1325 "parser.y"
{
			    if (parse_limit_check("flash limit",
						  (u_int) yyvsp[0].num ,
						  KRT_LIMIT_FLASH)) {
				PARSE_ERROR;
			    }
			    krt_flash_install_count = yyvsp[0].num ;
			}
break;
case 163:
#line 1337 "parser.y"
{
			    yyval.flag  = KRT_FLASH_INTERFACE;
			}
break;
case 164:
#line 1341 "parser.y"
{
			    yyval.flag  = KRT_FLASH_INTERNAL;
			}
break;
case 165:
#line 1345 "parser.y"
{
			    yyval.flag  = KRT_FLASH_ALL;
			}
break;
case 168:
#line 1357 "parser.y"
{
			    krt_install_priority = yyvsp[0].flag ;
			}
break;
case 169:
#line 1361 "parser.y"
{
			    if (parse_limit_check("install limit",
						  (u_int) yyvsp[0].num ,
						  KRT_LIMIT_INSTALL)) {
				PARSE_ERROR;
			    }
			    krt_install_count = yyvsp[0].num ;
			}
break;
case 170:
#line 1373 "parser.y"
{
			    yyval.flag  = KRT_INSTALL_PRIO_FLASH;
			}
break;
case 171:
#line 1377 "parser.y"
{
			    yyval.flag  = KRT_INSTALL_PRIO_HIGH;
			}
break;
case 172:
#line 1381 "parser.y"
{
			    yyval.flag  = KRT_INSTALL_PRIO_LOW;
			}
break;
case 173:
#line 1388 "parser.y"
{
			    yyval.flag  = (flag_t) 0;
			}
break;
case 174:
#line 1392 "parser.y"
{
			    yyval.flag  = yyvsp[0].flag ;
			}
break;
case 175:
#line 1396 "parser.y"
{
			    yyval.flag  = yyvsp[-2].flag  & ~yyvsp[0].flag ;
			}
break;
case 176:
#line 1403 "parser.y"
{
			    yyval.flag  = yyvsp[0].flag ;
			}
break;
case 177:
#line 1407 "parser.y"
{
			    yyval.flag  = yyvsp[-1].flag  | yyvsp[0].flag ;
			}
break;
case 179:
#line 1415 "parser.y"
{
			    yyval.flag  = trace_parse_packet(yyvsp[-2].flag , yyvsp[-1].flag , yyvsp[0].u_num );
			}
break;
case 180:
#line 1419 "parser.y"
{
			    yyval.flag  = TR_KRT_SYMBOLS;
			}
break;
case 181:
#line 1423 "parser.y"
{
			    yyval.flag  = TR_KRT_REMNANTS;
			}
break;
case 182:
#line 1427 "parser.y"
{
			    yyval.flag  = TR_KRT_IFLIST;
			}
break;
case 183:
#line 1431 "parser.y"
{
			    yyval.flag  = TR_KRT_REQUEST;
			}
break;
case 184:
#line 1435 "parser.y"
{
			    yyval.flag  = TR_KRT_INFO;
			}
break;
case 185:
#line 1442 "parser.y"
{
			    yyval.u_num  = TR_KRT_INDEX_PACKETS;
			}
break;
case 186:
#line 1446 "parser.y"
{
			    yyval.u_num  = TR_KRT_INDEX_ROUTES;
			}
break;
case 187:
#line 1450 "parser.y"
{
			    yyval.u_num  = TR_KRT_INDEX_REDIRECT;
			}
break;
case 188:
#line 1454 "parser.y"
{
			    yyval.u_num  = TR_KRT_INDEX_INTERFACE;
			}
break;
case 189:
#line 1458 "parser.y"
{
			    yyval.u_num  = TR_KRT_INDEX_OTHER;
			}
break;
case 190:
#line 1467 "parser.y"
{
			    PROTO_SEEN(RTPROTO_RIP);

			    parse_gwlist = &rip_gw_list;

			    rip_trace_options = trace_set_global(rip_trace_types,
								 (flag_t) 0);
			}
break;
case 191:
#line 1476 "parser.y"
{
			    switch (yyvsp[-1].num ) {
			    case T_OFF:
				BIT_RESET(rip_flags, RIPF_ON);
				break;

			    case T_ON:
				BIT_SET(rip_flags, RIPF_ON);
				break;
			    }

			    parse_proto = (proto_t) 0;
			    parse_gwlist = (gw_entry **) 0;
			}
break;
case 196:
#line 1499 "parser.y"
{
			    yyerrok;
			}
break;
case 197:
#line 1505 "parser.y"
{
			    rip_preference = yyvsp[0].pref ;
			}
break;
case 198:
#line 1509 "parser.y"
{
			    if (parse_metric_check(RTPROTO_RIP, &yyvsp[0].metric )) {
				PARSE_ERROR;
			    }
			    rip_default_metric = yyvsp[0].metric .metric;
			}
break;
case 199:
#line 1516 "parser.y"
{
			    BIT_RESET(rip_flags, RIPF_CHOOSE);
			    BIT_SET(rip_flags, RIPF_BROADCAST);
			}
break;
case 200:
#line 1521 "parser.y"
{
			    BIT_RESET(rip_flags, RIPF_CHOOSE|RIPF_BROADCAST);
			}
break;
case 201:
#line 1525 "parser.y"
{
			    trace_store(rip_trace_options, yyvsp[0].flag , yyvsp[-1].flag , yyvsp[-2].trace_file , rip_trace_types);
			}
break;
case 202:
#line 1529 "parser.y"
{
			    BIT_SET(rip_flags, RIPF_NOCHECK);
			}
break;
case 203:
#line 1533 "parser.y"
{
			    config_list *list = config_list_alloc(yyvsp[0].config , rip_config_free);

			    parse_adv_propagate_config(yyvsp[-1].adv , list, RTPROTO_RIP);

			    if (parse_adv_append(&rip_int_policy, yyvsp[-1].adv )) {
				PARSE_ERROR;
			    }

			    config_list_free(list);
			}
break;
case 204:
#line 1545 "parser.y"
{
			    rip_n_trusted += parse_gw_flag(yyvsp[0].adv , RTPROTO_RIP, GWF_TRUSTED);
			    if (!rip_n_trusted) {
				PARSE_ERROR;
			    }
			}
break;
case 205:
#line 1552 "parser.y"
{
			    rip_n_source += parse_gw_flag(yyvsp[0].adv , RTPROTO_RIP, GWF_SOURCE);
			    if (!rip_n_source) {
				PARSE_ERROR;
			    }
			    BIT_SET(rip_flags, RIPF_SOURCE);
			}
break;
case 206:
#line 1560 "parser.y"
{
			    rip_auth_query = (struct rip_authinfo *) yyvsp[0].vt ;
			}
break;
case 208:
#line 1568 "parser.y"
{
			    yyval.config  = config_append(yyvsp[-1].config , yyvsp[0].config );
			    if (!yyval.config ) {
				PARSE_ERROR;
			    }
			}
break;
case 209:
#line 1578 "parser.y"
{
			    yyval.config  = config_alloc(RIP_CONFIG_IN, GS2A(TRUE));
			}
break;
case 210:
#line 1582 "parser.y"
{
			    yyval.config  = config_alloc(RIP_CONFIG_IN, GS2A(FALSE));
			}
break;
case 211:
#line 1586 "parser.y"
{
			    yyval.config  = config_alloc(RIP_CONFIG_OUT, GS2A(TRUE));
			}
break;
case 212:
#line 1590 "parser.y"
{
			    yyval.config  = config_alloc(RIP_CONFIG_OUT, GS2A(FALSE));
			}
break;
case 213:
#line 1594 "parser.y"
{
			    if (parse_metric_check(RTPROTO_RIP, &yyvsp[0].metric )) {
				PARSE_ERROR;
			    }
			    yyval.config  = config_alloc(RIP_CONFIG_METRICIN, GS2A(yyvsp[0].metric .metric));
			}
break;
case 214:
#line 1601 "parser.y"
{
			    if (parse_metric_check(RTPROTO_RIP, &yyvsp[0].metric )) {
				PARSE_ERROR;
			    }
			    yyval.config  = config_alloc(RIP_CONFIG_METRICOUT, GS2A(yyvsp[0].metric .metric));
			}
break;
case 215:
#line 1608 "parser.y"
{
			    yyval.config  = config_alloc(RIP_CONFIG_AUTH, yyvsp[0].vt );
			}
break;
case 216:
#line 1612 "parser.y"
{
			    yyval.config  = config_alloc(RIP_CONFIG_AUTH2, yyvsp[0].vt );
			}
break;
case 217:
#line 1616 "parser.y"
{
			    flag_t flag;
			    
			    switch (yyvsp[-1].num ) {
			    case 1:
				if (yyvsp[0].flag ) {
				    (void) sprintf(parse_error, "options not valid with version 1");
				    PARSE_ERROR;
				}
				flag = RIP_IFPS_V1;
				break;

			    case 2:
				if (yyvsp[0].flag ) {
				    flag = yyvsp[0].flag ;
				} else {
				    /* This means whatever options the interface is capable of */
#ifdef	IP_MULTICAST
				    flag = RIP_IFPS_V2;
#else	/* IP_MULTICAST */
				    flag = RIP_IFPS_V2BC;
#endif	/* IP_MULTICAST */
				}
				break;

			    default:
				(void) sprintf(parse_error, "invalid version");
				PARSE_ERROR;
			    }
			    yyval.config  = config_alloc(RIP_CONFIG_FLAG, GS2A(flag));
			}
break;
case 218:
#line 1650 "parser.y"
{
			    yyval.flag  = 0;
			}
break;
case 219:
#line 1654 "parser.y"
{
			    yyval.flag  = RIP_IFPS_V2MC;
			}
break;
case 220:
#line 1658 "parser.y"
{
			    yyval.flag  = RIP_IFPS_V2BC;
			}
break;
case 221:
#line 1664 "parser.y"
{
			    yyval.vt  = (void_t) &rip_auth_none;
			}
break;
case 222:
#line 1668 "parser.y"
{
			    struct rip_authinfo *ap = (struct rip_authinfo *) task_block_alloc(rip_auth_block_index);

			    if (yyvsp[0].str .len > RIP_AUTH_SIZE) {
				sprintf(parse_error, "authentication-key \"%s\"%d longer than %d characters",
					yyvsp[0].str .ptr,
					yyvsp[0].str .len,
					RIP_AUTH_SIZE);
				PARSE_ERROR;
			    }

			    ap->auth_family = RIP_AF_AUTH;
			    ap->auth_type = htons((u_int16) yyvsp[-1].u_num );
			    /* Left justify the authentication key */
			    (void) strncpy((caddr_t) ap->auth_data, yyvsp[0].str .ptr, RIP_AUTH_SIZE);

			    free_charstr(yyvsp[0].str );

			    yyval.vt  = (void_t) ap;
			}
break;
case 223:
#line 1691 "parser.y"
{
			    yyval.u_num  = RIP_AUTH_SIMPLE;
			}
break;
case 224:
#line 1695 "parser.y"
{
			    yyval.u_num  = RIP_AUTH_SIMPLE;
			}
break;
case 225:
#line 1699 "parser.y"
{
			    yyval.u_num  = RIP_AUTH_MD5;
			}
break;
case 226:
#line 1706 "parser.y"
{
			    yyval.flag  = (flag_t) 0;
			}
break;
case 227:
#line 1710 "parser.y"
{
			    yyval.flag  = yyvsp[0].flag ;
			}
break;
case 228:
#line 1714 "parser.y"
{
			    yyval.flag  = yyvsp[-2].flag  & ~yyvsp[0].flag ;
			}
break;
case 229:
#line 1721 "parser.y"
{
			    yyval.flag  = yyvsp[0].flag ;
			}
break;
case 230:
#line 1725 "parser.y"
{
			    yyval.flag  = yyvsp[-1].flag  | yyvsp[0].flag ;
			}
break;
case 232:
#line 1733 "parser.y"
{
			    yyval.flag  = trace_parse_packet(yyvsp[-2].flag , yyvsp[-1].flag , yyvsp[0].u_num );
			}
break;
case 233:
#line 1740 "parser.y"
{
			    yyval.u_num  = TR_RIP_INDEX_PACKETS;
			}
break;
case 234:
#line 1744 "parser.y"
{
			    yyval.u_num  = TR_RIP_INDEX_REQUEST;
			}
break;
case 235:
#line 1748 "parser.y"
{
			    yyval.u_num  = TR_RIP_INDEX_RESPONSE;
			}
break;
case 236:
#line 1752 "parser.y"
{
			    yyval.u_num  = TR_RIP_INDEX_OTHER;
			}
break;
case 237:
#line 1758 "parser.y"
{
			    yyval.proto  = parse_proto = RTPROTO_RIP;
			    parse_gwlist = &rip_gw_list;
			    trace_tf(trace_global,
				     TR_PARSE,
				     0,
				     ("parse: %s PROTO: %s",
				      parse_where(),
				      gd_lower(trace_state(rt_proto_bits, yyval.proto ))));
			}
break;
case 238:
#line 1775 "parser.y"
{
			    PROTO_SEEN(RTPROTO_HELLO);

			    parse_gwlist = &hello_gw_list;

			    hello_trace_options = trace_set_global(hello_trace_types,
								   (flag_t) 0);
			}
break;
case 239:
#line 1784 "parser.y"
{
			    switch (yyvsp[-1].num ) {
			    case T_OFF:
				BIT_RESET(hello_flags, HELLOF_ON);
				break;

			    case T_ON:
				BIT_SET(hello_flags, HELLOF_ON);
				break;
			    }
			    
			    parse_proto = (proto_t) 0;
			    parse_gwlist = (gw_entry **) 0;
			}
break;
case 244:
#line 1808 "parser.y"
{
			    yyerrok;
			}
break;
case 245:
#line 1815 "parser.y"
{
			    hello_preference = yyvsp[0].pref ;
			}
break;
case 246:
#line 1819 "parser.y"
{
			    if (parse_metric_check(RTPROTO_HELLO, &yyvsp[0].metric )) {
				PARSE_ERROR;
			    }
			    hello_default_metric = yyvsp[0].metric .metric;
			}
break;
case 247:
#line 1826 "parser.y"
{
			    hello_flags = (hello_flags & ~HELLOF_CHOOSE) | HELLOF_BROADCAST;
			}
break;
case 248:
#line 1830 "parser.y"
{
			    hello_flags = hello_flags & ~(HELLOF_CHOOSE|HELLOF_BROADCAST);
			}
break;
case 249:
#line 1834 "parser.y"
{
			    trace_store(hello_trace_options, yyvsp[0].flag , yyvsp[-1].flag , yyvsp[-2].trace_file , hello_trace_types);
			}
break;
case 250:
#line 1838 "parser.y"
{
			    config_list *list = config_list_alloc(yyvsp[0].config , 0);

			    parse_adv_propagate_config(yyvsp[-1].adv , list, RTPROTO_HELLO);

			    if (parse_adv_append(&hello_int_policy, yyvsp[-1].adv )) {
				PARSE_ERROR;
			    }

			    config_list_free(list);
			}
break;
case 251:
#line 1850 "parser.y"
{
			    hello_n_trusted += parse_gw_flag(yyvsp[0].adv , RTPROTO_HELLO, GWF_TRUSTED);
			    if (!hello_n_trusted) {
				PARSE_ERROR;
			    }
			}
break;
case 252:
#line 1857 "parser.y"
{
			    hello_n_source += parse_gw_flag(yyvsp[0].adv , RTPROTO_HELLO, GWF_SOURCE);
			    if (!hello_n_source) {
				PARSE_ERROR;
			    }
			    BIT_SET(hello_flags, HELLOF_SOURCE);
			}
break;
case 254:
#line 1869 "parser.y"
{
			    yyval.config  = config_append(yyvsp[-1].config , yyvsp[0].config );
			    if (!yyval.config ) {
				PARSE_ERROR;
			    }
			}
break;
case 255:
#line 1879 "parser.y"
{
			    yyval.config  = config_alloc(HELLO_CONFIG_IN, GS2A(TRUE));
			}
break;
case 256:
#line 1883 "parser.y"
{
			    yyval.config  = config_alloc(HELLO_CONFIG_IN, GS2A(FALSE));
			}
break;
case 257:
#line 1887 "parser.y"
{
			    yyval.config  = config_alloc(HELLO_CONFIG_OUT, GS2A(TRUE));
			}
break;
case 258:
#line 1891 "parser.y"
{
			    yyval.config  = config_alloc(HELLO_CONFIG_OUT, GS2A(FALSE));
			}
break;
case 259:
#line 1895 "parser.y"
{
			    if (parse_metric_check(RTPROTO_HELLO, &yyvsp[0].metric )) {
				PARSE_ERROR;
			    }
			    yyval.config  = config_alloc(HELLO_CONFIG_METRICIN, GS2A(yyvsp[0].metric .metric));
			}
break;
case 260:
#line 1902 "parser.y"
{
			    if (parse_metric_check(RTPROTO_HELLO, &yyvsp[0].metric )) {
				PARSE_ERROR;
			    }
			    yyval.config  = config_alloc(HELLO_CONFIG_METRICOUT, GS2A(yyvsp[0].metric .metric));
			}
break;
case 261:
#line 1912 "parser.y"
{
			    yyval.flag  = (flag_t) 0;
			}
break;
case 262:
#line 1916 "parser.y"
{
			    yyval.flag  = yyvsp[0].flag ;
			}
break;
case 263:
#line 1920 "parser.y"
{
			    yyval.flag  = yyvsp[-2].flag  & ~yyvsp[0].flag ;
			}
break;
case 264:
#line 1927 "parser.y"
{
			    yyval.flag  = yyvsp[0].flag ;
			}
break;
case 265:
#line 1931 "parser.y"
{
			    yyval.flag  = yyvsp[-1].flag  | yyvsp[0].flag ;
			}
break;
case 267:
#line 1939 "parser.y"
{
			    yyval.flag  = trace_parse_packet(yyvsp[-2].flag , yyvsp[-1].flag , yyvsp[0].u_num );
			}
break;
case 268:
#line 1946 "parser.y"
{
			    yyval.u_num  = TR_HELLO_INDEX_PACKETS;
			}
break;
case 269:
#line 1953 "parser.y"
{
			    yyval.proto  = parse_proto = RTPROTO_HELLO;
			    parse_gwlist = &hello_gw_list;
			    trace_tf(trace_global,
				     TR_PARSE,
				     0,
				     ("parse: %s PROTO: %s",
				      parse_where(),
				      gd_lower(trace_state(rt_proto_bits, yyval.proto ))));
			}
break;
case 270:
#line 1970 "parser.y"
{
			    PROTO_SEEN(RTPROTO_OSPF);

			    parse_gwlist = &ospf.gw_list;

			    ospf.trace_options = trace_set_global(ospf_trace_types,
								  (flag_t) 0);

			    /* Somewhere around here is where we will figure out which OSPF global structure we are */
			    /* refering to and initialize it once we support multiple instantations of OSPF that is. */
			}
break;
case 271:
#line 1982 "parser.y"
{
			    if (yyvsp[-1].num  == T_OFF) {
                                ospf.ospf_admin_stat = OSPF_DISABLED;
                            } else {
                                if (ospf_parse_valid_check(parse_error)) {
                                    PARSE_ERROR;
                                }
                                ospf.ospf_admin_stat = OSPF_ENABLED;
                            }

                            parse_proto = (proto_t) 0;
                            parse_gwlist = (gw_entry **) 0;
                        }
break;
case 276:
#line 2004 "parser.y"
{
                                yyerrok;
                        }
break;
case 277:
#line 2010 "parser.y"
{
                        }
break;
case 278:
#line 2013 "parser.y"
{
                            trace_store(ospf.trace_options, yyvsp[0].flag , yyvsp[-1].flag , yyvsp[-2].trace_file , ospf_trace_types);
                        }
break;
case 279:
#line 2017 "parser.y"
{
                            ospf.export_interval = yyvsp[0].time ;
                        }
break;
case 280:
#line 2021 "parser.y"
{
                            ospf.export_limit = yyvsp[0].num ;
                        }
break;
case 281:
#line 2025 "parser.y"
{
			    ospf.mon_auth = yyvsp[0].ospf_auth ;
			}
break;
case 282:
#line 2029 "parser.y"
{
			    ospf_auth_length = yyvsp[0].hex .len;
			}
break;
case 283:
#line 2033 "parser.y"
{
			    ospf_auth *oap = (ospf_auth *)
			        task_block_alloc(ospf_auth_index);

			    oap->auth_type = yyvsp[-2].hex .type;
			    oap->auth_length = yyvsp[0].hex .len;
			    bcopy(yyvsp[0].hex .ptr, (caddr_t)oap->auth_key, yyvsp[0].hex .len);
			    free_bytestr(yyvsp[0].hex );

			    ospf.mon_auth = oap;
			}
break;
case 284:
#line 2045 "parser.y"
{
			    ospf_auth_length = OSPF_AUTH_SIMPLE_SIZE;
			}
break;
case 285:
#line 2049 "parser.y"
{
			    ospf_auth *oap = (ospf_auth *)
			        task_block_alloc(ospf_auth_index);

			    oap->auth_type = OSPF_AUTH_SIMPLE;
			    oap->auth_length = OSPF_AUTH_SIMPLE_SIZE;
			    bcopy(yyvsp[0].hex .ptr, (caddr_t)oap->auth_key, yyvsp[0].hex .len);
			    free_bytestr(yyvsp[0].hex );

			    ospf.mon_auth = oap;
			}
break;
case 286:
#line 2061 "parser.y"
{
			    ospf.log_first = yyvsp[-1].u_num ;
			    ospf.log_every = yyvsp[0].u_num ;
			}
break;
case 287:
#line 2066 "parser.y"
{
			    ospf_AREA = ospf_parse_area_alloc(yyvsp[0].u_num , parse_error);
			    if (!ospf_AREA) {
				PARSE_ERROR;
			    }
			    ospf_AREA->authtype = OSPF_AUTH_NONE;		/* Default to no authentication */
			}
break;
case 288:
#line 2074 "parser.y"
{
			    if (ospf_parse_area_check(ospf_AREA, parse_error)) {
				PARSE_ERROR;
			    }
			}
break;
case 289:
#line 2083 "parser.y"
{
			    yyval.u_num  = OSPF_LOG_FIRST;
			}
break;
case 290:
#line 2087 "parser.y"
{
			    if (parse_limit_check("syslog first",
						  (u_int) yyvsp[0].num ,
						  OSPF_LIMIT_LOG_FIRST)) {
				PARSE_ERROR;
			    }
			    yyval.u_num  = yyvsp[0].num ;
			}
break;
case 291:
#line 2099 "parser.y"
{
			    yyval.u_num  = OSPF_LOG_EVERY;
			}
break;
case 292:
#line 2103 "parser.y"
{
			    if (parse_limit_check("syslog every",
						  (u_int) yyvsp[0].num ,
						  OSPF_LIMIT_LOG_EVERY)) {
				PARSE_ERROR;
			    }
			    yyval.u_num  = yyvsp[0].num ;
			}
break;
case 295:
#line 2116 "parser.y"
{
			    yyerrok;
			}
break;
case 296:
#line 2122 "parser.y"
{
			    ospf.preference_ase = yyvsp[0].pref ;
			}
break;
case 297:
#line 2126 "parser.y"
{
			    ospf.export_metric = yyvsp[0].metric .metric;
			}
break;
case 298:
#line 2130 "parser.y"
{
			    switch (yyvsp[0].metric .state) {
			    case PARSE_METRICS_SET:
				ospf.export_tag = yyvsp[0].metric .metric;
				break;

			    case PARSE_METRICS_ALTERNATE:
				ospf.export_tag = PATH_OSPF_TAG_TRUSTED |
				    ((yyvsp[0].metric .metric << PATH_OSPF_TAG_USR_SHIFT) & PATH_OSPF_TAG_USR_MASK);
				break;
			    }
			}
break;
case 299:
#line 2143 "parser.y"
{
			    ospf.export_type = yyvsp[0].metric .metric;
			}
break;
case 302:
#line 2151 "parser.y"
{
			    yyerrok;
			}
break;
case 303:
#line 2157 "parser.y"
{
			    ospf_AREA->authtype = yyvsp[0].hex .type;
			    ospf_auth_length = yyvsp[0].hex .len;
			}
break;
case 304:
#line 2162 "parser.y"
{
			    if (ospf_AREA->area_id == OSPF_BACKBONE) {
				(void) sprintf(parse_error, "stub option not valid for backbone");
				PARSE_ERROR;
			    }

			    /* This is a stub */
			    BIT_SET(ospf_AREA->area_flags, OSPF_AREAF_STUB);

			    /* And we inject a default */
			    if (PARSE_METRIC_ISSET(&yyvsp[0].metric )) {
				BIT_SET(ospf_AREA->area_flags, OSPF_AREAF_STUB_DEFAULT);
				ospf_AREA->dflt_metric = yyvsp[0].metric .metric;
			    }
			}
break;
case 305:
#line 2177 "parser.y"
{}
break;
case 306:
#line 2179 "parser.y"
{
			    register dest_mask_internal *dmi;

			    if (ospf_AREA->nrcnt) {
				(void) sprintf(parse_error,
					"net range already specified");
				PARSE_ERROR;
			    }

			    if (yyvsp[-1].adv ) {

			    	/* Install each of the entries */
			    	DMI_WALK_ALL(yyvsp[-1].adv ->adv_dm.dm_internal, dmi, adv) {
					ospf_parse_add_net(ospf_AREA,
					   	adv->adv_dm.dm_dest,
					   	adv->adv_dm.dm_mask,
					   	adv->adv_result.res_flag);
			    	} DMI_WALK_ALL_END(yyvsp[-1].adv ->adv_dm.dm_internal,
					dmi, adv) ;

			    	adv_free_list(yyvsp[-1].adv );
			    }
			}
break;
case 307:
#line 2203 "parser.y"
{
			    if (parse_adv_append(&ospf_AREA->intf_policy, yyvsp[0].adv )) {
				PARSE_ERROR;
			    }
			}
break;
case 308:
#line 2209 "parser.y"
{
			    struct INTF *vif;

			    vif = ospf_parse_virt_parse(ospf_AREA, yyvsp[-3].sockaddr , yyvsp[-1].u_num , yyvsp[0].conflist , parse_error);
			    if (!vif) {
				PARSE_ERROR;
			    }

			    config_list_free(yyvsp[0].conflist );
			}
break;
case 309:
#line 2223 "parser.y"
{
			    config_list *list = yyvsp[0].conflist ;

			    if (PARSE_METRIC_ISSET(&yyvsp[-1].metric )) {

				list = config_list_add(list,
						       config_alloc(OSPF_CONFIG_COST,
								    GS2A(yyvsp[-1].metric .metric)),
						       ospf_config_free);
				if (!list) {
				    PARSE_ERROR;
				}
			    }

			    parse_adv_propagate_config(yyvsp[-2].adv , list, RTPROTO_OSPF);

			    config_list_free(list);

			    yyval.adv  = yyvsp[-2].adv ;
			}
break;
case 310:
#line 2244 "parser.y"
{
			    config_list *list = yyvsp[0].conflist ;

			    list = config_list_add(list,
						   config_alloc(OSPF_CONFIG_TYPE,
								GS2A(NONBROADCAST)),
						   ospf_config_free);
			    if (!list) {
				PARSE_ERROR;
			    }
			    
			    if (PARSE_METRIC_ISSET(&yyvsp[-1].metric )) {

				list = config_list_add(list,
						       config_alloc(OSPF_CONFIG_COST,
								    GS2A(yyvsp[-1].metric .metric)),
						       ospf_config_free);
				if (!list) {
				    PARSE_ERROR;
				}
			    }

			    parse_adv_propagate_config(yyvsp[-3].adv , list, RTPROTO_OSPF);

			    config_list_free(list);

			    yyval.adv  = yyvsp[-3].adv ;
			}
break;
case 311:
#line 2275 "parser.y"
{
			    yyval.conflist  = (config_list *) 0;
			}
break;
case 312:
#line 2279 "parser.y"
{
			    yyval.conflist  = config_list_alloc(yyvsp[-1].config , ospf_config_free);
			}
break;
case 314:
#line 2287 "parser.y"
{
			    yyval.config  = config_append(yyvsp[-2].config , yyvsp[-1].config );
			    if (!yyval.config ) {
				PARSE_ERROR;
			    }
			}
break;
case 315:
#line 2294 "parser.y"
{
				yyerrok;
			}
break;
case 316:
#line 2301 "parser.y"
{
			    yyval.conflist  = (config_list *) 0;
			}
break;
case 317:
#line 2305 "parser.y"
{
			    yyval.conflist  = config_list_alloc(yyvsp[-1].config , ospf_config_free);
			}
break;
case 319:
#line 2313 "parser.y"
{
			    yyval.config  = config_append(yyvsp[-2].config , yyvsp[-1].config );
			    if (!yyval.config ) {
				PARSE_ERROR;
			    }
			}
break;
case 320:
#line 2320 "parser.y"
{
				yyerrok;
			}
break;
case 321:
#line 2327 "parser.y"
{
			    yyval.config  = config_alloc(OSPF_CONFIG_ENABLE, GS2A(TRUE));
			}
break;
case 322:
#line 2331 "parser.y"
{
			    yyval.config  = config_alloc(OSPF_CONFIG_ENABLE, GS2A(FALSE));
			}
break;
case 323:
#line 2335 "parser.y"
{
			    if (parse_limit_check("retransmit-interval",
						  (u_int) yyvsp[0].time ,
						  OSPF_LIMIT_RETRANSMITINTERVAL)) {
				PARSE_ERROR;
			    }
			    yyval.config  = config_alloc(OSPF_CONFIG_RETRANSMIT, GS2A(yyvsp[0].time ));
			}
break;
case 324:
#line 2344 "parser.y"
{
			    if (parse_limit_check("transit-delay",
						  (u_int) yyvsp[0].time ,
						  OSPF_LIMIT_TRANSITDELAY)) {
				PARSE_ERROR;
			    }
			    yyval.config  = config_alloc(OSPF_CONFIG_TRANSIT, GS2A(yyvsp[0].time ));
			}
break;
case 325:
#line 2353 "parser.y"
{
			    yyval.config  = config_alloc(OSPF_CONFIG_PRIORITY, GS2A(yyvsp[0].num ));
			}
break;
case 326:
#line 2357 "parser.y"
{
			    if (parse_limit_check("hello-interval",
						  (u_int) yyvsp[0].time ,
						  OSPF_LIMIT_HELLOINTERVAL)) {
				PARSE_ERROR;
			    }
			    yyval.config  = config_alloc(OSPF_CONFIG_HELLO, GS2A(yyvsp[0].time ));
			}
break;
case 327:
#line 2366 "parser.y"
{
			    if (parse_limit_check("router-dead-interval",
						  (u_int) yyvsp[0].time ,
						  OSPF_LIMIT_ROUTERDEADINTERVAL)) {
				PARSE_ERROR;
			    }
			    yyval.config  = config_alloc(OSPF_CONFIG_ROUTERDEAD, GS2A(yyvsp[0].time ));
			}
break;
case 328:
#line 2375 "parser.y"
{
			    yyval.config  = config_alloc(OSPF_CONFIG_NOMULTI, GS2A(TRUE));
			}
break;
case 329:
#line 2379 "parser.y"
{
			    yyval.config  = config_alloc(OSPF_CONFIG_AUTH_MD5, GS2A(yyvsp[0].ospf_auth ));
			}
break;
case 330:
#line 2382 "parser.y"
{ ospf_auth_length = yyvsp[0].hex .len; }
break;
case 331:
#line 2384 "parser.y"
{
			    ospf_auth *oap = (ospf_auth *)
			        task_block_alloc(ospf_auth_index);

			    oap->auth_type = yyvsp[-2].hex .type;
			    oap->auth_length = yyvsp[0].hex .len;
			    bcopy(yyvsp[0].hex .ptr, (caddr_t)oap->auth_key, yyvsp[0].hex .len);
			    free_bytestr(yyvsp[0].hex );

			    yyval.config  = config_alloc(OSPF_CONFIG_AUTH, GS2A(oap));
			}
break;
case 332:
#line 2396 "parser.y"
{
			    ospf_auth *oap = (ospf_auth *) task_block_alloc(ospf_auth_index);

			    oap->auth_type = yyvsp[0].hex .type;
			    oap->auth_length = yyvsp[0].hex .len;
			    bcopy(yyvsp[0].hex .ptr, (caddr_t) oap->auth_key, yyvsp[0].hex .len);
			    free_bytestr(yyvsp[0].hex );

			    yyval.config  = config_alloc(OSPF_CONFIG_AUTH2, GS2A(oap));
			}
break;
case 333:
#line 2407 "parser.y"
{
			    /* Side-effect: The authentication type must be specified first */

			    ospf_auth *oap = (ospf_auth *) task_block_alloc(ospf_auth_index);
			    
			    switch (ospf_AREA->authtype) {
			    case OSPF_AUTH_NONE:
				sprintf(parse_error, "authentication-key specified when authentication-type is none");
				PARSE_ERROR;
				/* break; */

			    case OSPF_AUTH_SIMPLE:
				/* Simple password */
				break;

			    case OSPF_AUTH_MD5:
				/* MD5 crypto checksum */
				sprintf(parse_error,
				    "authkey not valid for md5 authentication");
				PARSE_ERROR;
				break;
				
			    default:
				sprintf(parse_error, "unknown authentication type: %d",
					ospf_AREA->authtype);
				PARSE_ERROR;
			    }

			    oap->auth_type = ospf_AREA->authtype;
			    oap->auth_length = yyvsp[0].hex .len;
			    bcopy(yyvsp[0].hex .ptr, (caddr_t) oap->auth_key, yyvsp[0].hex .len);

			    yyval.config  = config_alloc(OSPF_CONFIG_AUTH, GS2A(oap));

			    free_bytestr(yyvsp[0].hex );
			}
break;
case 335:
#line 2448 "parser.y"
{
			    if (parse_limit_check("poll-interval",
						  (u_int) yyvsp[0].time ,
						  OSPF_LIMIT_POLLINTERVAL)) {
				PARSE_ERROR;
			    }
			    yyval.config  = config_alloc(OSPF_CONFIG_POLL, GS2A(yyvsp[0].time ));
			}
break;
case 336:
#line 2457 "parser.y"
{
			    yyval.config  = config_alloc(OSPF_CONFIG_ROUTERS, GS2A(yyvsp[-1].ospf_router ));
			}
break;
case 337:
#line 2464 "parser.y"
{
			    PARSE_METRIC_CLEAR(&yyval.metric );
			}
break;
case 338:
#line 2468 "parser.y"
{
				yyval.metric  = yyvsp[0].metric ;
			}
break;
case 339:
#line 2474 "parser.y"
{
			    yyval.metric  = yyvsp[0].metric ;
			    if (parse_metric_check(RTPROTO_OSPF, &yyval.metric )) {
				PARSE_ERROR;
			    }
			}
break;
case 340:
#line 2484 "parser.y"
{
			    PARSE_METRIC_CLEAR(&yyval.metric );
			}
break;
case 341:
#line 2488 "parser.y"
{
			    yyval.metric  = yyvsp[0].metric ;
			}
break;
case 342:
#line 2494 "parser.y"
{
			    if (parse_limit_check("export-type",
						  (u_int) yyvsp[0].num ,
						  OSPF_LIMIT_EXPORTTYPE)) {
				PARSE_ERROR;
			    }
			    PARSE_METRIC_SET(&yyval.metric , yyvsp[0].num );
			}
break;
case 343:
#line 2506 "parser.y"
{
			    yyval.u_num  = yyvsp[0].u_num ;
			}
break;
case 344:
#line 2510 "parser.y"
{
			    yyval.u_num  = OSPF_BACKBONE;
			}
break;
case 345:
#line 2516 "parser.y"
{
			    if (parse_limit_check("area-number",
						  (u_int) yyvsp[0].num ,
						  OSPF_LIMIT_AREA)) {
				PARSE_ERROR;
			    }
			    yyval.u_num  = htonl((u_int32) yyvsp[0].num );
			}
break;
case 346:
#line 2527 "parser.y"
{
			    yyval.ospf_auth  = yyvsp[0].ospf_auth ;
			}
break;
case 347:
#line 2531 "parser.y"
{
			    yyval.ospf_auth  = yyvsp[-1].ospf_auth ;
			}
break;
case 348:
#line 2537 "parser.y"
{
			    /* initialize list this is leftmost */
			    ospf_md5_key_list = yyvsp[-1].ospf_auth ;
			}
break;
case 349:
#line 2542 "parser.y"
{
			    ospf_auth *kp;

			    /* make sure there are no duplicates */
			    for (kp = ospf_md5_key_list; kp;
				 kp = kp->auth_acc_next) {
				/* no two keys can have the same start time */
				if (kp->auth_generate.tr_start ==
				    yyvsp[-1].ospf_auth ->auth_generate.tr_start) {
				    sprintf(parse_error,
					"keys cannot have same start time");
				    PARSE_ERROR;
				}
				
			    }
			    /* add to list */
			    yyvsp[-2].ospf_auth ->auth_acc_next = yyvsp[-1].ospf_auth ;
			}
break;
case 350:
#line 2563 "parser.y"
{
			    /* set length for ospf_auth_key */
			    ospf_auth_length = OSPF_AUTH_MD5_SIZE;
			    ospf_md5_generate.tr_start = 0;
			    ospf_md5_generate.tr_stop = (time_t)-1;
			    ospf_md5_accept.tr_start = 0;
			    ospf_md5_accept.tr_stop = (time_t)-1;
			}
break;
case 351:
#line 2572 "parser.y"
{
			    ospf_auth *oap;

			    /* Check that key id is within range */
			    if (yyvsp[-1].num  < OSPF_MD5_MIN_KEY_ID
				|| yyvsp[-1].num  > OSPF_MD5_MAX_KEY_ID) {
				sprintf(parse_error,
					"key id must be between %d and %d",
					OSPF_MD5_MIN_KEY_ID,
					OSPF_MD5_MAX_KEY_ID);
				PARSE_ERROR;
			    }
			    if (ospf_md5_generate.tr_stop != (time_t)-1
				&& DIFFTIME(ospf_md5_generate.tr_stop,
					    ospf_md5_generate.tr_start) < 0) {
				sprintf(parse_error,
					"stop generate time must be later than start generate time");
				PARSE_ERROR;

			    }
			    if (ospf_md5_accept.tr_stop != (time_t)-1
				&& DIFFTIME(ospf_md5_accept.tr_stop,
					    ospf_md5_accept.tr_start) < 0) {
				sprintf(parse_error,
					"stop accept time must be later than start accept time");
				PARSE_ERROR;

			    }
			    
			    oap = (ospf_auth *)
			        task_block_alloc(ospf_auth_index);

			    bcopy(yyvsp[-3].hex .ptr, (caddr_t) oap->auth_key, yyvsp[-3].hex .len);
			    free_bytestr(yyvsp[-3].hex );

			    oap->auth_type = OSPF_AUTH_MD5;
			    oap->auth_id = yyvsp[-1].num ;
			    oap->auth_generate = ospf_md5_generate;
			    oap->auth_accept = ospf_md5_accept;

			    yyval.ospf_auth  = oap;
			}
break;
case 352:
#line 2617 "parser.y"
{
			}
break;
case 353:
#line 2620 "parser.y"
{
			}
break;
case 354:
#line 2626 "parser.y"
{
				}
break;
case 355:
#line 2630 "parser.y"
{
				}
break;
case 356:
#line 2636 "parser.y"
{
			    ospf_md5_accept.tr_start = yyvsp[0].time ;
			}
break;
case 357:
#line 2640 "parser.y"
{
			    ospf_md5_accept.tr_stop = yyvsp[0].time ;
			}
break;
case 358:
#line 2644 "parser.y"
{
			    ospf_md5_generate.tr_start = yyvsp[0].time ;
			}
break;
case 359:
#line 2648 "parser.y"
{
			    ospf_md5_generate.tr_stop = yyvsp[0].time ;
			}
break;
case 360:
#line 2654 "parser.y"
{
			    struct tm tm;
			    bzero(&tm, sizeof(tm));
			    if (yyvsp[-7].num  < 1970) {
				sprintf(parse_error,
					"year must be after 1970");
				PARSE_ERROR;
			    }
			    tm.tm_year = (yyvsp[-7].num  - 1900);
			    if (yyvsp[-5].num  < 1 || yyvsp[-5].num  > 12) {
				sprintf(parse_error,
					"month must be between 1 and 12");
				PARSE_ERROR;
			    }
			    tm.tm_mon = yyvsp[-5].num  - 1;
			    if (yyvsp[-3].num  < 1 || yyvsp[-3].num  > 31) {
				sprintf(parse_error,
					"day must be between 1 and 31");
				PARSE_ERROR;
			    }
			    tm.tm_mday = yyvsp[-3].num ;
			    if (yyvsp[-2].num  < 0 || yyvsp[-2].num  > 23) {
				sprintf(parse_error,
					"hour must be between 0 and 23");
				PARSE_ERROR;
			    }
			    tm.tm_hour = yyvsp[-2].num ;
			    if (yyvsp[0].num  < 0 || yyvsp[0].num  > 59) {
				sprintf(parse_error,
					"minute must be between 0 and 59");
				PARSE_ERROR;
			    }
			    tm.tm_min = yyvsp[0].num ;

			    tm.tm_isdst = -1;
			    yyval.time  = mktime(&tm);
			    if (yyval.time  == (time_t)-1) {
				sprintf(parse_error,
					"mktime: failed to parse date");
				PARSE_ERROR;

			    }
			}
break;
case 361:
#line 2700 "parser.y"
{
			    ospf_auth_length = yyvsp[0].hex .len;
			}
break;
case 362:
#line 2704 "parser.y"
{
			    /* Copy the key */
			    yyval.hex  = yyvsp[0].hex ;	/* struct copy */
			    /* And get type type */
			    yyval.hex .type = yyvsp[-2].hex .type;
			}
break;
case 363:
#line 2714 "parser.y"
{
			    yyval.hex .type = OSPF_AUTH_NONE;
			    yyval.hex .len = 0;
			    
			}
break;
case 364:
#line 2720 "parser.y"
{
			    yyval.hex .type = OSPF_AUTH_SIMPLE;
			    yyval.hex .len = OSPF_AUTH_SIMPLE_SIZE;
			}
break;
case 365:
#line 2727 "parser.y"
{
			    yyval.hex .type = OSPF_AUTH_NONE;
			    yyval.hex .len = 0;
			    
			}
break;
case 366:
#line 2733 "parser.y"
{
			    yyval.hex .type = OSPF_AUTH_SIMPLE;
			    yyval.hex .len = OSPF_AUTH_SIMPLE_SIZE;
			}
break;
case 367:
#line 2738 "parser.y"
{
			    yyval.hex .type = OSPF_AUTH_MD5;
			    yyval.hex .len = OSPF_AUTH_MD5_SIZE;
			}
break;
case 368:
#line 2745 "parser.y"
{
			    if (yyvsp[0].str .len > ospf_auth_length) {
				sprintf(parse_error, "authentication-key \"%s\"%d longer than %d characters",
					yyvsp[0].str .ptr,
					yyvsp[0].str .len,
					ospf_auth_length);
				PARSE_ERROR;
			    }

			    yyval.hex .ptr = (byte *) task_mem_calloc((task *) 0, sizeof (u_int32), OSPF_AUTH_SIZE);
			    yyval.hex .len = ospf_auth_length;
			    yyval.hex .type = 0;

			    /* Left justify */
			    strncpy((caddr_t) yyval.hex .ptr, yyvsp[0].str .ptr, yyval.hex .len);

			    free_charstr(yyvsp[0].str );
			}
break;
case 369:
#line 2764 "parser.y"
{
			    yyval.hex .ptr = (byte *) task_mem_calloc((task *) 0, sizeof (u_int32), OSPF_AUTH_SIZE);
			    yyval.hex .len = ospf_auth_length;
			    yyval.hex .type = 0;

			    /* Right justify the key */
			    bcopy((caddr_t) &yyvsp[0].num ,
				  yyval.hex .ptr + ospf_auth_length - sizeof (yyvsp[0].num ),
				  sizeof (yyvsp[0].num ));
			}
break;
case 370:
#line 2775 "parser.y"
{
			    if (yyvsp[0].hex .len > ospf_auth_length) {
				sprintf(parse_error, "authentication-key longer than %d characters",
					ospf_auth_length);
				PARSE_ERROR;
			    }

			    yyval.hex .ptr = (byte *) task_mem_calloc((task *) 0, sizeof (u_int32), OSPF_AUTH_SIZE);
			    yyval.hex .len = ospf_auth_length;
			    yyval.hex .type = 0;

			    /* Right justify the key */
			    bcopy((caddr_t) yyvsp[0].hex .ptr,
				  yyval.hex .ptr + ospf_auth_length - yyvsp[0].hex .len,
				  yyvsp[0].hex .len);

			    free_bytestr(yyvsp[0].hex );
			}
break;
case 373:
#line 2800 "parser.y"
{
			    yyerrok;
			}
break;
case 374:
#line 2806 "parser.y"
{
			    ospf_parse_add_host(ospf_AREA, sock2ip(yyvsp[-2].sockaddr ), yyvsp[0].metric .metric);

			    sockfree(yyvsp[-2].sockaddr );
			}
break;
case 376:
#line 2816 "parser.y"
{
			    (yyval.ospf_router  = yyvsp[-1].ospf_router )->ocr_next = yyvsp[-2].ospf_router ;
			}
break;
case 377:
#line 2820 "parser.y"
{
			    yyerrok;
			}
break;
case 378:
#line 2827 "parser.y"
{
			    yyval.ospf_router  = ospf_parse_router_alloc(sock2in(yyvsp[-1].gwp ->gw_addr), (u_int) yyvsp[0].num );
			}
break;
case 379:
#line 2834 "parser.y"
{
			    yyval.num  = 0;
			}
break;
case 380:
#line 2838 "parser.y"
{
			    yyval.num  = 1;
			}
break;
case 381:
#line 2844 "parser.y"
{
			    if (parse_limit_check("priority",
						  (u_int) yyvsp[0].num ,
						  OSPF_LIMIT_DRPRIORITY)) {
				PARSE_ERROR;
			    }
			    yyval.num  = yyvsp[0].num ;
			}
break;
case 382:
#line 2855 "parser.y"
{
			    yyval.proto  = parse_proto = RTPROTO_OSPF;
			    parse_gwlist = (gw_entry **) 0;
			    trace_tf(trace_global,
				     TR_PARSE,
				     0,
				     ("parse: %s PROTO: %s",
				      parse_where(),
				      gd_lower(trace_state(rt_proto_bits, yyval.proto ))));
			}
break;
case 383:
#line 2869 "parser.y"
{
			    /* Make sure this PS type is known */
			    adv_psfunc_add(RTPROTO_OSPF_ASE, &ospf_adv_psfunc);

			    yyval.proto  = parse_proto = RTPROTO_OSPF_ASE;
			    parse_gwlist = (gw_entry **) 0;
			    trace_tf(trace_global,
				     TR_PARSE,
				     0,
				     ("parse: %s PROTO: %s",
				      parse_where(),
				      gd_lower(trace_state(rt_proto_bits, yyval.proto ))));
			}
break;
case 384:
#line 2886 "parser.y"
{
			    yyval.flag  = (flag_t) 0;
			}
break;
case 385:
#line 2890 "parser.y"
{
			    yyval.flag  = yyvsp[0].flag ;
			}
break;
case 386:
#line 2894 "parser.y"
{
			    yyval.flag  = yyvsp[-2].flag  & ~yyvsp[0].flag ;
			}
break;
case 387:
#line 2901 "parser.y"
{
			    yyval.flag  = yyvsp[0].flag ;
			}
break;
case 388:
#line 2905 "parser.y"
{
			    yyval.flag  = yyvsp[-1].flag  | yyvsp[0].flag ;
			}
break;
case 390:
#line 2913 "parser.y"
{
			    yyval.flag  = trace_parse_packet(yyvsp[-2].flag , yyvsp[-1].flag , yyvsp[0].u_num );
			}
break;
case 391:
#line 2917 "parser.y"
{
			    yyval.flag  = TR_OSPF_LSA_BLD;
			}
break;
case 392:
#line 2921 "parser.y"
{
			    yyval.flag  = TR_OSPF_LSA_TX;
			}
break;
case 393:
#line 2925 "parser.y"
{
			    yyval.flag  = TR_OSPF_LSA_RX;
			}
break;
case 394:
#line 2929 "parser.y"
{
			    yyval.flag  = TR_OSPF_SPF;
			}
break;
case 395:
#line 2933 "parser.y"
{
			    yyval.flag  = TR_OSPF_TRAP;
			}
break;
case 396:
#line 2937 "parser.y"
{
#ifdef	DEBUG
			    yyval.flag  = TR_OSPF_DEBUG;
#else	/* DEBUG */
			    yyval.flag  = 0;
#endif	/* DEBUG */
			}
break;
case 397:
#line 2948 "parser.y"
{
			    yyval.u_num  = TR_OSPF_INDEX_PACKETS;
			}
break;
case 398:
#line 2952 "parser.y"
{
			    yyval.u_num  = TR_OSPF_INDEX_HELLO;
			}
break;
case 399:
#line 2956 "parser.y"
{
			    yyval.u_num  = TR_OSPF_INDEX_DD;
			}
break;
case 400:
#line 2960 "parser.y"
{
			    yyval.u_num  = TR_OSPF_INDEX_LSR;
			}
break;
case 401:
#line 2964 "parser.y"
{
			    yyval.u_num  = TR_OSPF_INDEX_LSU;
			}
break;
case 402:
#line 2968 "parser.y"
{
			    yyval.u_num  = TR_OSPF_INDEX_ACK;
			}
break;
case 403:
#line 2976 "parser.y"
{
			    yyval.adv  = (adv_entry *) 0;
			}
break;
case 404:
#line 2980 "parser.y"
{
			    if (sock2ip(yyvsp[-2].dm .dm_dest) == INADDR_ANY) {
				sprintf(parse_error, "invalid range %A/%A",
					yyvsp[-2].dm .dm_dest,
					yyvsp[-2].dm .dm_mask);
				PARSE_ERROR;
			    }
			    yyval.adv  = adv_alloc(ADVFT_DM, (proto_t) 0);
			    yyval.adv ->adv_dm = yyvsp[-2].dm ;
			    yyval.adv ->adv_result.res_flag = yyvsp[-1].flag ;
			    yyval.adv  = adv_destmask_insert(parse_error, yyvsp[-3].adv , yyval.adv );
			    if (!yyval.adv ) {
				PARSE_ERROR;
			    }
			}
break;
case 405:
#line 2996 "parser.y"
{
			    yyerrok;
			}
break;
case 408:
#line 3008 "parser.y"
{
			    yyval.flag  = Advertise;
			}
break;
case 409:
#line 3012 "parser.y"
{
			    yyval.flag  = DoNotAdvertise;
			}
break;
case 410:
#line 3037 "parser.y"
{
			    PROTO_SEEN(RTPROTO_BGP);

			    parse_gwlist = &parse_gwp;

			    bgp_default_trace_options = trace_set_global(bgp_trace_types,
									 (flag_t) 0);
			}
break;
case 411:
#line 3046 "parser.y"
{
			    switch (yyvsp[-1].num ) {
			    case T_ON:
				doing_bgp = TRUE;

				if (!bgp_conf_check(parse_error)) {
				    PARSE_ERROR;
				}
				break;

			    case T_OFF:
				doing_bgp = FALSE;
			    }

			    parse_proto = (proto_t) 0;
			    parse_gwlist = (gw_entry **) 0;
			}
break;
case 416:
#line 3072 "parser.y"
{
			    yyerrok;
			}
break;
case 417:
#line 3078 "parser.y"
{
			    bgp_default_preference = yyvsp[0].pref ;
			}
break;
case 418:
#line 3082 "parser.y"
{
			    bgp_default_preference2 = yyvsp[0].pref ;
			}
break;
case 419:
#line 3086 "parser.y"
{
			    if (parse_metric_check(RTPROTO_BGP, &yyvsp[0].metric )) {
				PARSE_ERROR;
			    }
			    bgp_default_metric = yyvsp[0].metric .metric;
			}
break;
case 420:
#line 3093 "parser.y"
{
			    trace_store(bgp_default_trace_options, yyvsp[0].flag , yyvsp[-1].flag , yyvsp[-2].trace_file , bgp_trace_types);
			}
break;
case 421:
#line 3097 "parser.y"
{
			    /* Allocate a Group structure */
			    bgp = bgp_conf_group_alloc();

			    /* Set the type and AS */
			    bgp->bgpg_type = yyvsp[-4].num ;
			    bgp->bgpg_peer_as = yyvsp[-2].as ;

			    /* Point at this configuration information */
			    bcp = &bgp->bgpg_conf;

			    /* Inherit global trace flags */
			    bgp->bgpg_trace_options = trace_alloc(bgp_default_trace_options);

			    /* Set the protocol for routing peers */
			    bgp->bgpg_proto = yyvsp[-1].proto ;

			    /* Set the interface policy */
			    bgp->bgpg_ifap_policy = yyvsp[0].adv ;
			}
break;
case 422:
#line 3118 "parser.y"
{
			    /* Add this group */
			    bgp = bgp_conf_group_add(bgp, parse_error);
			    if (!bgp) {
				PARSE_ERROR;
			    }
			}
break;
case 423:
#line 3126 "parser.y"
{
			    /* Check group */
			    if (!bgp_conf_group_check(bgp, parse_error)) {
				PARSE_ERROR;
			    }
			}
break;
case 428:
#line 3142 "parser.y"
{
				yyerrok;
			}
break;
case 429:
#line 3148 "parser.y"
{
			    /* Allocate this peer */
			    bnp = bgp_conf_peer_alloc(bgp);

			    /* Point at this configuration information */
			    bcp = &bnp->bgp_conf;
			}
break;
case 430:
#line 3156 "parser.y"
{
			    /* Set peer address */
			    bnp->bgp_addr = yyvsp[-1].sockaddr ;

			    bnp = bgp_conf_peer_add(bgp, bnp, parse_error);
			    if (!bnp) {
				PARSE_ERROR;
			    }
			}
break;
case 431:
#line 3166 "parser.y"
{
			    parse_contig_masks = FALSE;
			}
break;
case 432:
#line 3170 "parser.y"
{
			    parse_contig_masks = TRUE;

			    bgp->bgpg_allow = adv_destmask_finish(yyvsp[-1].adv );
			}
break;
case 433:
#line 3178 "parser.y"
{
			    if (parse_metric_check(RTPROTO_BGP, &yyvsp[0].metric )) {
				PARSE_ERROR;
			    }
			    BIT_SET(bcp->bgpc_options, BGPO_METRIC_OUT);
			    bcp->bgpc_metric_out = yyvsp[0].metric .metric;
			}
break;
case 434:
#line 3186 "parser.y"
{
			    BIT_SET(bcp->bgpc_options, BGPO_LOCAL_AS);
			    bcp->bgpc_local_as = yyvsp[0].as ;
			    aslocal_set(yyvsp[0].as , 0);
			}
break;
case 435:
#line 3192 "parser.y"
{
			    BIT_SET(bcp->bgpc_options, BGPO_NOGENDEFAULT);
			}
break;
case 436:
#line 3196 "parser.y"
{
			    if (bcp->bgpc_gateway) {
				/* Free the old one */
				sockfree(bcp->bgpc_gateway);
			    }
			    BIT_SET(bcp->bgpc_options, BGPO_GATEWAY);
			    bcp->bgpc_gateway = sockdup(yyvsp[0].gwp ->gw_addr);
			    gw_freelist(*parse_gwlist);
			    *parse_gwlist = parse_gwp = (gw_entry *) 0;
			}
break;
case 437:
#line 3207 "parser.y"
{
			    BIT_SET(bcp->bgpc_options, BGPO_PREFERENCE);
			    bcp->bgpc_preference = yyvsp[0].pref ;
			}
break;
case 438:
#line 3212 "parser.y"
{
			    BIT_SET(bcp->bgpc_options, BGPO_PREFERENCE2);
			    bcp->bgpc_preference2 = yyvsp[0].pref ;
			}
break;
case 439:
#line 3217 "parser.y"
{
			    if (bcp->bgpc_lcladdr) {
				/* Free the old one */
				ifae_free(bcp->bgpc_lcladdr);
			    }
			    BIT_SET(bcp->bgpc_options, BGPO_LCLADDR);
			    bcp->bgpc_lcladdr = yyvsp[0].ifae ;
			}
break;
case 440:
#line 3226 "parser.y"
{
			    if (parse_limit_check("hold time",
						  (u_int) yyvsp[0].time ,
						  BGP_LIMIT_HOLDTIME)) {
				PARSE_ERROR;
			    }
			    BIT_SET(bcp->bgpc_options, BGPO_HOLDTIME);
			    bcp->bgpc_holdtime_out = yyvsp[0].time ;
			}
break;
case 441:
#line 3236 "parser.y"
{
			    trace_store(bcp->bgpc_trace_options, yyvsp[0].flag , yyvsp[-1].flag , yyvsp[-2].trace_file , bgp_trace_types);
			}
break;
case 442:
#line 3240 "parser.y"
{
			    if (!BGP_KNOWN_VERSION(yyvsp[0].num )) {
				(void) sprintf(parse_error, "unsupported BGP version: %d",
					       yyvsp[0].num );
				PARSE_ERROR;
			    }
			    BIT_SET(bcp->bgpc_options, BGPO_VERSION);
			    bcp->bgpc_conf_version = yyvsp[0].num ;
			}
break;
case 443:
#line 3250 "parser.y"
{
			    BIT_SET(bcp->bgpc_options, BGPO_PASSIVE);
			}
break;
case 444:
#line 3254 "parser.y"
{
			    BIT_SET(bcp->bgpc_options, BGPO_KEEPALL);
			}
break;
case 445:
#line 3258 "parser.y"
{
			    BIT_SET(bcp->bgpc_options, BGPO_KEEPNONE);
			}
break;
case 446:
#line 3262 "parser.y"
{
			    BIT_SET(bcp->bgpc_options, BGPO_ANALRETENTIVE);
			}
break;
case 447:
#line 3266 "parser.y"
{
			    BIT_SET(bcp->bgpc_options, BGPO_KEEPALIVESALWAYS);
			}
break;
case 448:
#line 3270 "parser.y"
{
			    BIT_SET(bcp->bgpc_options, BGPO_NOAGGRID);
			}
break;
case 449:
#line 3274 "parser.y"
{
			    BIT_SET(bcp->bgpc_options, BGPO_NOAUTHCHECK);
			}
break;
case 450:
#line 3278 "parser.y"
{
			    BIT_SET(bcp->bgpc_options, BGPO_SETPREF);
			    bcp->bgpc_setpref = yyvsp[0].pref ;
			}
break;
case 451:
#line 3283 "parser.y"
{
				BIT_SET(bcp->bgpc_options, BGPO_IGNOREFIRSTASHOP);
			}
break;
case 452:
#line 3287 "parser.y"
{
			    bcp->bgpc_send_bufsize = yyvsp[0].num ;
			}
break;
case 453:
#line 3291 "parser.y"
{
			    bcp->bgpc_recv_bufsize = yyvsp[0].num ;
			}
break;
case 454:
#line 3295 "parser.y"
{
			    bcp->bgpc_rti_time = yyvsp[0].time ;
			}
break;
case 455:
#line 3299 "parser.y"
{
			    bcp->bgpc_rto_time = yyvsp[0].time ;
			}
break;
case 456:
#line 3303 "parser.y"
{
			    BIT_SET(bcp->bgpc_options, BGPO_TTL);
			    bcp->bgpc_ttl = yyvsp[0].num ;
			}
break;
case 457:
#line 3308 "parser.y"
{
			    BIT_SET(bcp->bgpc_options, BGPO_NOV4ASLOOP);
			}
break;
case 458:
#line 3312 "parser.y"
{
			    BIT_SET(bcp->bgpc_options, BGPO_V3ASLOOPOKAY);
			}
break;
case 459:
#line 3316 "parser.y"
{
			    BIT_SET(bcp->bgpc_options, BGPO_LOGUPDOWN);
			}
break;
case 460:
#line 3322 "parser.y"
{
			    if (parse_limit_check("kernel buffer size",
						  (u_int) yyvsp[0].num ,
						  BGP_LIMIT_KBUF)) {
				PARSE_ERROR;
			    }
			    yyval.num  = yyvsp[0].num ;
			}
break;
case 461:
#line 3333 "parser.y"
{
			    yyval.num  = BGPG_INTERNAL;
			}
break;
case 462:
#line 3337 "parser.y"
{
			    yyval.num  = BGPG_EXTERNAL;
			}
break;
case 463:
#line 3341 "parser.y"
{
			    yyval.num  = BGPG_INTERNAL_IGP;
			}
break;
case 464:
#line 3345 "parser.y"
{
			    yyval.num  = BGPG_INTERNAL_RT;
			}
break;
case 465:
#line 3349 "parser.y"
{
			    yyval.num  = BGPG_TEST;
			}
break;
case 466:
#line 3356 "parser.y"
{
			    yyval.flag  = (flag_t) 0;
			}
break;
case 467:
#line 3360 "parser.y"
{
			    yyval.flag  = yyvsp[0].flag ;
			}
break;
case 468:
#line 3364 "parser.y"
{
			    yyval.flag  = yyvsp[-2].flag  & ~yyvsp[0].flag ;
			}
break;
case 469:
#line 3371 "parser.y"
{
			    yyval.flag  = yyvsp[0].flag ;
			}
break;
case 470:
#line 3375 "parser.y"
{
			    yyval.flag  = yyvsp[-1].flag  | yyvsp[0].flag ;
			}
break;
case 471:
#line 3382 "parser.y"
{
			    yyval.proto  = (proto_t) 0;
			}
break;
case 472:
#line 3386 "parser.y"
{
			    yyval.proto  = yyvsp[0].proto ;
			}
break;
case 473:
#line 3393 "parser.y"
{
			    assert(FALSE);
			}
break;
case 474:
#line 3397 "parser.y"
{
			    yyval.proto  = RTPROTO_RIP;
			}
break;
case 475:
#line 3401 "parser.y"
{
			    yyval.proto  = RTPROTO_HELLO;
			}
break;
case 476:
#line 3405 "parser.y"
{
			    yyval.proto  = RTPROTO_OSPF;
			}
break;
case 477:
#line 3411 "parser.y"
{
			    yyval.adv  = (adv_entry *) 0;
			}
break;
case 478:
#line 3415 "parser.y"
{
			    parse_adv_propagate_config(yyvsp[0].adv , (config_list *) 0, RTPROTO_BGP);

			    yyval.adv  = yyvsp[0].adv ;
			}
break;
case 480:
#line 3425 "parser.y"
{
			    yyval.flag  = TR_BGP_ASPATH;
			}
break;
case 481:
#line 3429 "parser.y"
{
			    yyval.flag  = trace_parse_packet(yyvsp[-2].flag , yyvsp[-1].flag , yyvsp[0].u_num );
			}
break;
case 482:
#line 3436 "parser.y"
{
			    yyval.u_num  = TR_BGP_INDEX_PACKETS;
			}
break;
case 483:
#line 3440 "parser.y"
{
			    yyval.u_num  = TR_BGP_INDEX_OPEN;
			}
break;
case 484:
#line 3444 "parser.y"
{
			    yyval.u_num  = TR_BGP_INDEX_UPDATE;
			}
break;
case 485:
#line 3448 "parser.y"
{
			    yyval.u_num  = TR_BGP_INDEX_KEEPALIVE;
			}
break;
case 486:
#line 3455 "parser.y"
{
			    yyval.proto  = parse_proto = RTPROTO_BGP;
			    parse_gwlist = (gw_entry **) 0;
			    trace_tf(trace_global,
				     TR_PARSE,
				     0,
				     ("parse: %s PROTO: %s",
				      parse_where(),
				      gd_lower(trace_state(rt_proto_bits, yyval.proto ))));
			}
break;
case 487:
#line 3469 "parser.y"
{
			    yyval.adv  = bgp->bgpg_allow;
			}
break;
case 488:
#line 3473 "parser.y"
{
			    yyval.adv  = adv_alloc(ADVFT_DM, (proto_t) 0);
			    yyval.adv ->adv_dm = yyvsp[-1].dm ;
			    yyval.adv  = adv_destmask_insert(parse_error, yyvsp[-2].adv , yyval.adv );
			    if (!yyval.adv ) {
				PARSE_ERROR;
			    }
			}
break;
case 489:
#line 3482 "parser.y"
{
			    yyerrok;
			}
break;
case 493:
#line 3499 "parser.y"
{
			    icmp_trace_options = trace_set_global(icmp_trace_types,
								  (flag_t) 0);
			}
break;
case 497:
#line 3510 "parser.y"
{
			    yyerrok;
			}
break;
case 498:
#line 3518 "parser.y"
{
			    trace_store(icmp_trace_options, yyvsp[0].flag , yyvsp[-1].flag , yyvsp[-2].trace_file , icmp_trace_types);
			}
break;
case 499:
#line 3525 "parser.y"
{
			    yyval.flag  = (flag_t) 0;
			}
break;
case 500:
#line 3529 "parser.y"
{
			    yyval.flag  = yyvsp[0].flag ;
			}
break;
case 501:
#line 3533 "parser.y"
{
			    yyval.flag  = yyvsp[-2].flag  & ~yyvsp[0].flag ;
			}
break;
case 502:
#line 3540 "parser.y"
{
			    yyval.flag  = yyvsp[0].flag ;
			}
break;
case 503:
#line 3544 "parser.y"
{
			    yyval.flag  = yyvsp[-1].flag  | yyvsp[0].flag ;
			}
break;
case 505:
#line 3552 "parser.y"
{
			    yyval.flag  = trace_parse_packet(yyvsp[-2].flag , yyvsp[-1].flag , yyvsp[0].u_num );
			}
break;
case 506:
#line 3559 "parser.y"
{
			    yyval.u_num  = TR_ICMP_INDEX_PACKETS;
			}
break;
case 507:
#line 3563 "parser.y"
{
			    yyval.u_num  = TR_ICMP_INDEX_REDIRECT;
			}
break;
case 508:
#line 3567 "parser.y"
{
			    yyval.u_num  = TR_ICMP_INDEX_ROUTER;
			}
break;
case 509:
#line 3571 "parser.y"
{
			    yyval.u_num  = TR_ICMP_INDEX_INFO;
			}
break;
case 510:
#line 3575 "parser.y"
{
			    yyval.u_num  = TR_ICMP_INDEX_ERROR;
			}
break;
case 511:
#line 3586 "parser.y"
{
#if	defined(PROTO_ICMP) || defined(KRT_RT_SOCK)
			    PROTO_SEEN(RTPROTO_REDIRECT);

			    parse_gwlist = &redirect_gw_list;

			    redirect_trace_options = trace_set_global(redirect_trace_types,
								      (flag_t) 0);
#endif	/* defined(PROTO_ICMP) || defined(KRT_RT_SOCK) */
			}
break;
case 512:
#line 3597 "parser.y"
{
#if	defined(PROTO_ICMP) || defined(KRT_RT_SOCK)
			    switch (yyvsp[-1].num ) {
			    case T_ON:
				/* Enabled by default */
				break;

			    case T_OFF:
				redirect_disable(RTPROTO_REDIRECT);
				break;
			    }

			    parse_proto = (proto_t) 0;
			    parse_gwlist = (gw_entry **) 0;
#endif	/* defined(PROTO_ICMP) || defined(KRT_RT_SOCK) */
			}
break;
case 517:
#line 3623 "parser.y"
{
			    yyerrok;
			}
break;
case 518:
#line 3629 "parser.y"
{
#if	defined(PROTO_ICMP) || defined(KRT_RT_SOCK)
			    redirect_preference = yyvsp[0].pref ;
#endif	/* defined(PROTO_ICMP) || defined(KRT_RT_SOCK) */
			}
break;
case 519:
#line 3635 "parser.y"
{
#if	defined(PROTO_ICMP) || defined(KRT_RT_SOCK)
			    config_list *list = config_list_alloc(yyvsp[0].config , 0);

			    parse_adv_propagate_config(yyvsp[-1].adv , list, RTPROTO_REDIRECT);

			    if (parse_adv_append(&redirect_int_policy, yyvsp[-1].adv )) {
				PARSE_ERROR;
			    }

			    config_list_free(list);
#endif	/* defined(PROTO_ICMP) || defined(KRT_RT_SOCK) */
			}
break;
case 520:
#line 3649 "parser.y"
{
#if	defined(PROTO_ICMP) || defined(KRT_RT_SOCK)
			    redirect_n_trusted += parse_gw_flag(yyvsp[0].adv , RTPROTO_REDIRECT, GWF_TRUSTED);
			    if (!redirect_n_trusted) {
				PARSE_ERROR;
			    }
#endif	/* defined(PROTO_ICMP) || defined(KRT_RT_SOCK) */
			}
break;
case 521:
#line 3658 "parser.y"
{
			    trace_store(redirect_trace_options, yyvsp[0].flag , yyvsp[-1].flag , yyvsp[-2].trace_file , redirect_trace_types);
			}
break;
case 523:
#line 3666 "parser.y"
{
			    yyval.config  = config_append(yyvsp[-1].config , yyvsp[0].config );
			    if (!yyval.config ) {
				PARSE_ERROR;
			    }
			}
break;
case 524:
#line 3676 "parser.y"
{
#if	defined(PROTO_ICMP) || defined(KRT_RT_SOCK)
			    yyval.config  = config_alloc(REDIRECT_CONFIG_IN, GS2A(TRUE));
#endif	/* defined(PROTO_ICMP) || defined(KRT_RT_SOCK) */
			}
break;
case 525:
#line 3682 "parser.y"
{
#if	defined(PROTO_ICMP) || defined(KRT_RT_SOCK)
			    yyval.config  = config_alloc(REDIRECT_CONFIG_IN, GS2A(FALSE));
#endif	/* defined(PROTO_ICMP) || defined(KRT_RT_SOCK) */
			}
break;
case 526:
#line 3691 "parser.y"
{
			    yyval.flag  = (flag_t) 0;
			}
break;
case 527:
#line 3695 "parser.y"
{
			    yyval.flag  = yyvsp[0].flag ;
			}
break;
case 528:
#line 3699 "parser.y"
{
			    yyval.flag  = yyvsp[-2].flag  & ~yyvsp[0].flag ;
			}
break;
case 529:
#line 3706 "parser.y"
{
			    yyval.flag  = yyvsp[0].flag ;
			}
break;
case 530:
#line 3710 "parser.y"
{
			    yyval.flag  = yyvsp[-1].flag  | yyvsp[0].flag ;
			}
break;
case 532:
#line 3731 "parser.y"
{
			}
break;
case 533:
#line 3738 "parser.y"
{
			    if (parse_new_state(PS_ROUTE)) {
				PARSE_ERROR;
			    }
			}
break;
case 534:
#line 3746 "parser.y"
{
			    /* Need to set this for static routes, not used for interface routes */
			    parse_proto = RTPROTO_STATIC;
			    parse_gwlist = &rt_gw_list;
			    rt_open(rt_task);
			}
break;
case 535:
#line 3753 "parser.y"
{
			    rt_close(rt_task, (gw_entry *) 0, 0, NULL);
			    parse_proto = (proto_t) 0;
			    parse_gwlist = (gw_entry **) 0;
			}
break;
case 538:
#line 3763 "parser.y"
{
				yyerrok;
			}
break;
case 539:
#line 3769 "parser.y"
{
			    /* Route to an interface */

			    if (rt_parse_route(yyvsp[-4].dm .dm_dest,
					       yyvsp[-4].dm .dm_mask,
					       (adv_entry *) 0,
					       yyvsp[-2].adv ,
					       PARSE_METRIC_ISSET(&yyvsp[-1].metric ) ? (pref_t) yyvsp[-1].metric .metric : RTPREF_STATIC,
					       yyvsp[0].flag ,
					       parse_error)) {
				PARSE_ERROR;
			    }

			    sockfree(yyvsp[-4].dm .dm_dest);
			}
break;
case 540:
#line 3785 "parser.y"
{
			    /* Route to a gateway */

			    if (rt_parse_route_aspath(yyvsp[-6].dm .dm_dest,
						      yyvsp[-6].dm .dm_mask,
						      yyvsp[-4].adv ,
						      yyvsp[-3].adv ,
						      PARSE_METRIC_ISSET(&yyvsp[-1].metric ) ? (pref_t) yyvsp[-1].metric .metric : RTPREF_STATIC,
						      yyvsp[0].flag  | RTS_GATEWAY,
						      yyvsp[-2].aspath ,
						      parse_error)) {
				PARSE_ERROR;
			    }

			    sockfree(yyvsp[-6].dm .dm_dest);
			}
break;
case 541:
#line 3804 "parser.y"
{
			    yyval.aspath  = (as_path *) 0;
			}
break;
case 542:
#line 3808 "parser.y"
{
			    yyval.aspath  = aspath_create(yyvsp[0].as );
			}
break;
case 543:
#line 3814 "parser.y"
{
			    assert(FALSE);
			}
break;
case 546:
#line 3820 "parser.y"
{
			    yyval.dm .dm_dest = sockdup(inet_addr_default);
			    yyval.dm .dm_mask = inet_mask_default;
			    trace_tf(trace_global,
				     TR_PARSE,
				     0,
				     ("parse: %s DEST: %A MASK: %A",
				      parse_where(),
				      yyval.dm .dm_dest,
				      yyval.dm .dm_mask));
			}
break;
case 547:
#line 3834 "parser.y"
{
			    yyval.flag  = 0;
			}
break;
case 548:
#line 3838 "parser.y"
{
			    yyval.flag  = yyvsp[-1].flag  | yyvsp[0].flag ;
			}
break;
case 549:
#line 3844 "parser.y"
{
			    yyval.flag  = RTS_RETAIN;
			}
break;
case 550:
#line 3848 "parser.y"
{
			    yyval.flag  = RTS_REJECT;
			}
break;
case 551:
#line 3852 "parser.y"
{
			    yyval.flag  = RTS_BLACKHOLE;
			}
break;
case 552:
#line 3856 "parser.y"
{
			    yyval.flag  = RTS_NOTINSTALL;
			}
break;
case 553:
#line 3864 "parser.y"
{
			    if (parse_new_state(PS_CONTROL)) {
				PARSE_ERROR;
			    }
			}
break;
case 554:
#line 3873 "parser.y"
{
			    assert(FALSE);
			}
break;
case 555:
#line 3877 "parser.y"
{
#if	defined(PROTO_EGP) || defined(PROTO_BGP)
			    adv_entry *adv;
			    pmet_t *pp = &yyvsp[-1].metric_adv .metric;

			    /*
			     *	Tack the list of destinations onto the end of the list
			     *	for neighbors with the specified AS.
			     */
			    adv = adv_alloc(ADVFT_AS, yyvsp[-4].proto_lists .proto);
			    adv->adv_as = yyvsp[-2].as ;
			    adv->adv_list = yyvsp[-1].metric_adv .adv;
			    if (PARSE_METRIC_ISSET(pp)) {
				if (PARSE_METRIC_ISRESTRICT(pp)) {
				    BIT_SET(adv->adv_flag, ADVF_NO);
				} else {
				    BIT_SET(adv->adv_flag, ADVFOT_PREFERENCE);
				    adv->adv_result.res_preference = pp->metric;
				}
			    }

			    if (!parse_adv_as(yyvsp[-4].proto_lists .import_list, adv)) {
				PARSE_ERROR;
			    }
#endif	/* defined(PROTO_EGP) || defined(PROTO_BGP) */
			}
break;
case 556:
#line 3904 "parser.y"
{
			    adv_entry *adv;
			    pmet_t *pp = &yyvsp[-1].metric_adv .metric;

			    /*
			     *	Tack the list of destinations onto the end of the list
			     *	for neighbors with the specified AS.
			     */
			    adv = adv_alloc(ADVFT_ASPATH, RTPROTO_BGP);
			    adv->adv_aspath = yyvsp[-2].asp_match ;
			    adv->adv_list = yyvsp[-1].metric_adv .adv;
			    if (PARSE_METRIC_ISSET(pp)) {
				if (PARSE_METRIC_ISRESTRICT(pp)) {
				    BIT_SET(adv->adv_flag, ADVF_NO);
				} else {
				    BIT_SET(adv->adv_flag, ADVFOT_PREFERENCE);
				    adv->adv_result.res_preference = pp->metric;
				}
			    }

			    if (parse_adv_append(&bgp_import_aspath_list, adv)) {
				PARSE_ERROR;
			    }
			}
break;
case 557:
#line 3929 "parser.y"
{
			    /*
			     *	Append the dest_mask list to the end of the import list
			     *	for the specified protocol.
			     */
			    (void) parse_adv_propagate_preference(yyvsp[-2].adv , yyvsp[-3].proto_lists .proto, &yyvsp[-1].metric_adv .metric, yyvsp[-1].metric_adv .adv);

			    switch (yyvsp[-2].adv ->adv_flag & ADVF_TYPE) {
			    case ADVFT_GW:
				/* Gateway list */
			        {
				    adv_entry *adv = yyvsp[-2].adv ;
				    adv_entry *advn;

				    do {
					advn = adv->adv_next;
					adv->adv_next = NULL;
					/* Reset the type */
					adv->adv_flag = (adv->adv_flag & ~ADVF_TYPE) | ADVFT_ANY;
					if (parse_adv_append(&adv->adv_gwp->gw_import, adv)) {
					    PARSE_ERROR;
					}
				    } while ((adv = advn));
				}
				break;

			    case ADVFT_ANY:
				/* Global list */
				if (parse_adv_append(yyvsp[-3].proto_lists .import_list, yyvsp[-2].adv )) {
				    PARSE_ERROR;
				}
				break;

			    case ADVFT_IFN:
			    case ADVFT_IFAE:
				/* Interface list */
				if (parse_adv_append(&int_import[yyvsp[-3].proto_lists .proto], yyvsp[-2].adv )) {
				    PARSE_ERROR;
				}
				break;

			    default:
				assert(FALSE);
			    }
			}
break;
case 558:
#line 3975 "parser.y"
{
			    adv_entry *adv;
			    pmet_t *pp = &yyvsp[-1].metric_adv .metric;

			    /* Make sure this PS type is known */
			    adv_psfunc_add(RTPROTO_OSPF_ASE, &ospf_adv_psfunc);

			    /*
			     *	Append the dest_mask list to the end of the OSPF import list.
			     */
			    adv = adv_alloc(ADVFT_PS, RTPROTO_OSPF_ASE);
			    adv->adv_list = yyvsp[-1].metric_adv .adv;
			    adv->adv_ps = (caddr_t) adv;

			    /* Set Tag */
			    if (PARSE_METRIC_ISSET(&yyvsp[-2].metric )) {
				BIT_SET(adv->adv_flag, ADVFOT_METRIC);
				BIT_SET(adv->adv_result.res_flag, OSPF_EXPORT_TAG);
				adv->adv_result.res_metric = yyvsp[-2].metric .metric;
			    }

			    /* Set Preference */
			    if (PARSE_METRIC_ISSET(pp)) {
				if (PARSE_METRIC_ISRESTRICT(pp)) {
				    BIT_SET(adv->adv_flag, ADVF_NO);
				} else {
				    BIT_SET(adv->adv_flag, ADVFOT_PREFERENCE);
				    adv->adv_result.res_preference = pp->metric;
				}
			    }

			    if (parse_adv_append(&ospf.import_list, adv)) {
				PARSE_ERROR;
			    }
			}
break;
case 559:
#line 4011 "parser.y"
{
#if	defined(PROTO_EGP) || defined(PROTO_BGP)
			    adv_entry *adv;
			    adv_entry *adv_prop = yyvsp[-1].metric_adv .adv;
			    pmet_t *mp = &yyvsp[-1].metric_adv .metric;

			    /*
			     *	Tack the list of destinations onto the end of the list
			     *	for neighbors with the specified AS.
			     */
			    adv = adv_alloc(ADVFT_AS, yyvsp[-4].proto_lists .proto);
			    adv->adv_as = yyvsp[-2].as ;
			    adv->adv_list = adv_prop;
			    if (PARSE_METRIC_ISRESTRICT(mp)) {
				BIT_SET(adv->adv_flag, ADVF_NO);
			    } else if (PARSE_METRIC_ISSET(mp)) {
				BIT_SET(adv->adv_flag, ADVFOT_METRIC);
				adv->adv_result.res_metric = mp->metric;
			    }

			    if (!parse_adv_as(yyvsp[-4].proto_lists .export_list, adv)) {
				PARSE_ERROR;
			    }
#endif	/* defined(PROTO_EGP) || defined(PROTO_BGP) */
			}
break;
case 560:
#line 4037 "parser.y"
{
			    /*
			     *	Append the dest_mask list to the end of the export list
			     *	for the specified protocol.
			     */

			    (void) parse_adv_propagate_metric(yyvsp[-2].adv , yyvsp[-3].proto_lists .proto, &yyvsp[-1].metric_adv .metric, yyvsp[-1].metric_adv .adv);

			    switch (yyvsp[-2].adv ->adv_flag & ADVF_TYPE) {
			    case ADVFT_GW:
				/* Gateway list */
			        {
				    adv_entry *adv = yyvsp[-2].adv ;
				    adv_entry *advn;
				
				    do {
					advn = adv->adv_next;
					adv->adv_next = NULL;
					if (parse_adv_append(&adv->adv_gwp->gw_export, adv)) {
					    PARSE_ERROR;
					}
				    } while ((adv = advn));
				}
				break;

			    case ADVFT_ANY:
				/* Global list for the protocol */
				if (parse_adv_append(yyvsp[-3].proto_lists .export_list, yyvsp[-2].adv )) {
				    PARSE_ERROR;
				}
				break;

			    case ADVFT_IFN:
			    case ADVFT_IFAE:
				/* Interface list */
				if (parse_adv_append(&int_export[yyvsp[-3].proto_lists .proto], yyvsp[-2].adv )) {
				    PARSE_ERROR;
				}
				break;

			    default:
				assert(FALSE);
			    }
			}
break;
case 561:
#line 4082 "parser.y"
{
			    adv_entry *adv;
			    adv_entry *adv_prop = yyvsp[-1].metric_adv .adv;
			    pmet_t *mp = &yyvsp[-1].metric_adv .metric;

			    /* If we export then we are a border router */
			    ospf.asbr = TRUE;

			    /* Make sure this PS type is known */
			    adv_psfunc_add(RTPROTO_OSPF_ASE, &ospf_adv_psfunc);

			    /*
			     *	Append the dest_mask list to the end of the export list
			     *	for the specified protocol.
			     */
			    adv = adv_alloc(ADVFT_PS, RTPROTO_OSPF_ASE);
			    adv->adv_ps = (caddr_t) adv;
			    adv->adv_list = adv_prop;

			    /* Metric */
			    if (PARSE_METRIC_ISRESTRICT(mp)) {
				BIT_SET(adv->adv_flag, ADVF_NO);
			    } else if (PARSE_METRIC_ISSET(mp)) {
				BIT_SET(adv->adv_flag, ADVFOT_METRIC);
				adv->adv_result.res_metric = mp->metric;
			    }

			    /* Type */
			    if (PARSE_METRIC_ISSET(&yyvsp[-3].metric )) {
				BIT_SET(adv->adv_flag, ADVFOT_FLAG);
				adv->adv_result.res_flag = yyvsp[-3].metric .metric;
			    }

			    /* Tag */
			    switch (yyvsp[-2].metric .state) {
			    case PARSE_METRICS_SET:
				BIT_SET(adv->adv_flag, ADVFOT_METRIC2|ADVFOT_FLAG);
				BIT_SET(adv->adv_result.res_flag, OSPF_EXPORT_TAG|OSPF_EXPORT_TAG_METRIC2);
				adv->adv_result.res_metric2 = yyvsp[-2].metric .metric;
				break;

			    case PARSE_METRICS_ALTERNATE:
				BIT_SET(adv->adv_flag, ADVFOT_METRIC2|ADVFOT_FLAG);
				BIT_SET(adv->adv_result.res_flag, OSPF_EXPORT_TAG|OSPF_EXPORT_TAG_METRIC2);
				adv->adv_result.res_metric2 = PATH_OSPF_TAG_TRUSTED |
				    ((yyvsp[-2].metric .metric << PATH_OSPF_TAG_USR_SHIFT) & PATH_OSPF_TAG_USR_MASK);
				break;
			    }

			    if (parse_adv_append(&ospf.export_list, adv)) {
				PARSE_ERROR;
			    }
			}
break;
case 562:
#line 4136 "parser.y"
{
			    adv_entry *adv;
			    pmet_t *pp = &yyvsp[-5].metric ;

			    /*
			     *	Tack the list of destinations onto the end of the list
			     *	for neighbors with the specified AS.
			     */
			    adv = adv_alloc(ADVFT_DM | yyvsp[-4].flag , 0);
			    adv->adv_dm = yyvsp[-6].dm ;	/* struct copy */
			    BIT_RESET(adv->adv_dm.dm_flags, DMF_EXACT);
			    BIT_SET(adv->adv_dm.dm_flags, DMF_REFINE);
			    adv->adv_list = yyvsp[-2].adv ;
			    if (PARSE_METRIC_ISSET(pp)) {
				if (PARSE_METRIC_ISRESTRICT(pp)) {
				    BIT_SET(adv->adv_flag, ADVF_NO);
				} else {
				    BIT_SET(adv->adv_flag, ADVFOT_PREFERENCE);
				    adv->adv_result.res_preference = pp->metric;
				}
			    }

			    aggregate_list_inet = adv_destmask_insert(parse_error, aggregate_list_inet, adv);
			    if (!aggregate_list_inet) {
				PARSE_ERROR;
			    }
			}
break;
case 563:
#line 4164 "parser.y"
{
			    adv_entry *adv;
			    pmet_t *pp = &yyvsp[-4].metric ;

			    /*
			     *	Tack the list of destinations onto the end of the list
			     *	for neighbors with the specified AS.
			     */
			    adv = adv_alloc(ADVFT_DM | ADVF_AGGR_GENERATE, 0);
			    adv->adv_dm = yyvsp[-5].dm ;	/* struct copy */
			    BIT_RESET(adv->adv_dm.dm_flags, DMF_EXACT);
			    BIT_SET(adv->adv_dm.dm_flags, DMF_REFINE);
			    adv->adv_list = yyvsp[-2].adv ;
			    if (PARSE_METRIC_ISSET(pp)) {
				if (PARSE_METRIC_ISRESTRICT(pp)) {
				    BIT_SET(adv->adv_flag, ADVF_NO);
				} else {
				    BIT_SET(adv->adv_flag, ADVFOT_PREFERENCE);
				    adv->adv_result.res_preference = pp->metric;
				}
			    }

			    aggregate_list_inet = adv_destmask_insert(parse_error, aggregate_list_inet, adv);
			    if (!aggregate_list_inet) {
				PARSE_ERROR;
			    }
			}
break;
case 564:
#line 4199 "parser.y"
{
			    yyval.flag  = 0;
			}
break;
case 565:
#line 4203 "parser.y"
{
			    yyval.flag  = ADVF_AGGR_BRIEF;
			}
break;
case 568:
#line 4215 "parser.y"
{
			    yyval.adv  = yyvsp[0].adv ;
			    BIT_SET(yyval.adv ->adv_flag, ADVF_FIRST);
			}
break;
case 569:
#line 4220 "parser.y"
{
			    yyval.adv  = yyvsp[-1].adv ;
			    if (parse_adv_append(&yyval.adv , yyvsp[0].adv )) {
				PARSE_ERROR;
			    }
			}
break;
case 570:
#line 4227 "parser.y"
{
			    yyerrok;
			}
break;
case 571:
#line 4234 "parser.y"
{
			    pmet_t *pp = &yyvsp[-1].metric_adv .metric;
			    
			    yyval.adv  = adv_alloc(ADVFT_ANY, yyvsp[-2].proto );
			    yyval.adv ->adv_list = yyvsp[-1].metric_adv .adv;
			    if (PARSE_METRIC_ISSET(pp)) {
				if (PARSE_METRIC_ISRESTRICT(pp)) {
				    BIT_SET(yyval.adv ->adv_flag, ADVF_NO);
				} else {
				    BIT_SET(yyval.adv ->adv_flag, ADVFOT_PREFERENCE);
				    yyval.adv ->adv_result.res_preference = pp->metric;
				}
			    }
			}
break;
case 572:
#line 4249 "parser.y"
{
			    pmet_t *pp = &yyvsp[-1].metric_adv .metric;
			    
			    yyval.adv  = adv_alloc(ADVFT_AS, yyvsp[-4].proto );
			    yyval.adv ->adv_as = yyvsp[-2].as ;
			    yyval.adv ->adv_list = yyvsp[-1].metric_adv .adv;
			    if (PARSE_METRIC_ISSET(pp)) {
				if (PARSE_METRIC_ISRESTRICT(pp)) {
				    BIT_SET(yyval.adv ->adv_flag, ADVF_NO);
				} else {
				    BIT_SET(yyval.adv ->adv_flag, ADVFOT_PREFERENCE);
				    yyval.adv ->adv_result.res_preference = pp->metric;
				}
			    }
			}
break;
case 573:
#line 4265 "parser.y"
{
			    pmet_t *pp = &yyvsp[-1].metric_adv .metric;
			    
			    yyval.adv  = adv_alloc(ADVFT_ASPATH, yyvsp[-3].proto );
			    yyval.adv ->adv_aspath = yyvsp[-2].asp_match ;
			    yyval.adv ->adv_list = yyvsp[-1].metric_adv .adv;
			    if (PARSE_METRIC_ISSET(pp)) {
				if (PARSE_METRIC_ISRESTRICT(pp)) {
				    BIT_SET(yyval.adv ->adv_flag, ADVF_NO);
				} else {
				    BIT_SET(yyval.adv ->adv_flag, ADVFOT_PREFERENCE);
				    yyval.adv ->adv_result.res_preference = pp->metric;
				}
			    }
			}
break;
case 574:
#line 4281 "parser.y"
{
			    pmet_t *pp = &yyvsp[-1].metric_adv .metric;
			    
			    yyval.adv  = adv_alloc(ADVFT_TAG, (proto_t) yyvsp[-4].proto );
			    yyval.adv ->adv_tag = yyvsp[-2].metric .metric;
			    yyval.adv ->adv_list = yyvsp[-1].metric_adv .adv;
			    if (PARSE_METRIC_ISSET(pp)) {
				if (PARSE_METRIC_ISRESTRICT(pp)) {
				    BIT_SET(yyval.adv ->adv_flag, ADVF_NO);
				} else {
				    BIT_SET(yyval.adv ->adv_flag, ADVFOT_PREFERENCE);
				    yyval.adv ->adv_result.res_preference = pp->metric;
				}
			    }
			}
break;
case 575:
#line 4300 "parser.y"
{
			    yyval.proto  = RTPROTO_ANY;
			}
break;
case 576:
#line 4304 "parser.y"
{
			    yyval.proto  = RTPROTO_DIRECT;
			}
break;
case 577:
#line 4308 "parser.y"
{
			    yyval.proto  = RTPROTO_STATIC;
			}
break;
case 578:
#line 4312 "parser.y"
{
			    yyval.proto  = RTPROTO_KERNEL;
			}
break;
case 579:
#line 4316 "parser.y"
{
			    yyval.proto  = RTPROTO_AGGREGATE;
			}
break;
case 580:
#line 4320 "parser.y"
{
			    yyval.proto  = RTPROTO_HELLO;
			}
break;
case 581:
#line 4324 "parser.y"
{
			    yyval.proto  = RTPROTO_RIP;
			}
break;
case 582:
#line 4328 "parser.y"
{
			    yyval.proto  = RTPROTO_OSPF;
			}
break;
case 583:
#line 4332 "parser.y"
{
			    yyval.proto  = RTPROTO_OSPF_ASE;
			}
break;
case 584:
#line 4336 "parser.y"
{
			    yyval.proto  = RTPROTO_BGP;
			}
break;
case 585:
#line 4347 "parser.y"
{
			    parse_export_proto = yyval.proto_lists .proto = yyvsp[0].proto ;

			    switch (yyvsp[0].proto ) {
#if	defined(PROTO_EGP)
			    case RTPROTO_EGP:
				yyval.proto_lists .import_list = &egp_import_list;
				yyval.proto_lists .export_list = &egp_export_list;
				break;
#endif

#if	defined(PROTO_BGP)
			    case RTPROTO_BGP:
				yyval.proto_lists .import_list = &bgp_import_list;
				yyval.proto_lists .export_list = &bgp_export_list;
				break;
#endif

			    default:
				(void) sprintf(parse_error,
					       "unknown import protocol: %s",
					       gd_lower(trace_state(rt_proto_bits, yyvsp[0].proto )));
				PARSE_ERROR;
			    }
			}
break;
case 586:
#line 4377 "parser.y"
{
			}
break;
case 587:
#line 4380 "parser.y"
{
			    trace_log_tf(trace_global,
					 0,
					 LOG_WARNING,
					 ("parse: %s use of the ``listen'' keyword is deprecated; removal is recommended",
					  parse_where()));
			}
break;
case 588:
#line 4390 "parser.y"
{
			    yyval.proto_lists  = yyvsp[0].proto_lists ;
			    
			    parse_proto = yyvsp[0].proto_lists .proto;
			    parse_gwlist = yyvsp[0].proto_lists .gw_list;

			    if (!yyvsp[0].proto_lists .import_list) {
				(void) sprintf(parse_error,
					       "unknown import protocol: %s",
					       gd_lower(trace_state(rt_proto_bits, yyvsp[0].proto_lists .proto)));
				PARSE_ERROR;
			    }
			    
			    trace_tf(trace_global,
				     TR_PARSE,
				     0,
				     ("parse: %s PROTO: %s",
				      parse_where(),
				      gd_lower(trace_state(rt_proto_bits, yyvsp[0].proto_lists .proto))));
			}
break;
case 589:
#line 4417 "parser.y"
{
			    PARSE_METRIC_RESTRICT(&yyval.metric_adv .metric);
			    yyval.metric_adv .adv = (adv_entry *) 0;
			}
break;
case 590:
#line 4422 "parser.y"
{
			    yyval.metric_adv .metric = yyvsp[-3].metric ;
			    yyval.metric_adv .adv = adv_destmask_finish(yyvsp[-1].adv );
			}
break;
case 591:
#line 4430 "parser.y"
{
			    yyval.adv  = (adv_entry *) 0;
			}
break;
case 592:
#line 4434 "parser.y"
{
			    
			    yyval.adv  = adv_destmask_insert(parse_error, yyvsp[-2].adv , yyvsp[-1].adv );
			    if (!yyval.adv ) {
				PARSE_ERROR;
			    }
			}
break;
case 593:
#line 4442 "parser.y"
{
			    yyerrok;
			}
break;
case 594:
#line 4449 "parser.y"
{
			    parse_contig_masks = FALSE;
			}
break;
case 595:
#line 4453 "parser.y"
{
			    parse_contig_masks = TRUE;

			    if (yyvsp[-1].dm .dm_dest
				&& (socktype(yyvsp[-1].dm .dm_dest) != AF_INET)) {
				sprintf(parse_error, "address not INET");
				PARSE_ERROR
			    }

			    yyval.adv  = adv_alloc(ADVFT_DM, (proto_t) 0);
			    yyval.adv ->adv_dm = yyvsp[-1].dm ;
			    if (PARSE_METRIC_ISSET(&yyvsp[0].metric )) {
				if (PARSE_METRIC_ISRESTRICT(&yyvsp[0].metric )) {
				    BIT_SET(yyval.adv ->adv_flag, ADVF_NO);
				} else {
				    BIT_SET(yyval.adv ->adv_flag, ADVFOT_PREFERENCE);
				    yyval.adv ->adv_result.res_preference = yyvsp[0].metric .metric;
				}
			    }
			}
break;
case 596:
#line 4481 "parser.y"
{
			    yyval.proto_lists  = yyvsp[0].proto_lists ;
			    
			    parse_export_proto = parse_proto = yyvsp[0].proto_lists .proto;
			    parse_gwlist = yyvsp[0].proto_lists .gw_list;

			    if (!yyvsp[0].proto_lists .export_list) {
				(void) sprintf(parse_error,
					       "unknown export protocol: %s",
					       gd_lower(trace_state(rt_proto_bits, yyvsp[0].proto_lists .proto)));
				PARSE_ERROR;
			    }
			    
			    trace_tf(trace_global,
				     TR_PARSE,
				     0,
				     ("parse: %s PROTO: %s",
				      parse_where(),
				      gd_lower(trace_state(rt_proto_bits, yyvsp[0].proto_lists .proto))));
			}
break;
case 597:
#line 4505 "parser.y"
{
			    yyval.proto  = parse_export_proto = RTPROTO_OSPF_ASE;
			    trace_tf(trace_global,
				     TR_PARSE,
				     0,
				     ("parse: %s PROTO: %s",
				      parse_where(),
				      gd_lower(trace_state(rt_proto_bits, yyval.proto ))));
			}
break;
case 598:
#line 4519 "parser.y"
{
			}
break;
case 599:
#line 4522 "parser.y"
{
			    trace_log_tf(trace_global,
					 0,
					 LOG_WARNING,
					 ("parse: %s use of the ``announce'' keyword is deprecated; removal is recommended",
					  parse_where()));
			}
break;
case 600:
#line 4535 "parser.y"
{
			    PARSE_METRIC_RESTRICT(&yyval.metric_adv .metric);
			    yyval.metric_adv .adv = (adv_entry *) 0;
			}
break;
case 601:
#line 4540 "parser.y"
{
			    yyval.metric_adv .metric = yyvsp[-3].metric ;
			    yyval.metric_adv .adv = yyvsp[-1].adv ;
			}
break;
case 602:
#line 4548 "parser.y"
{
			    yyval.adv  = (adv_entry *) 0;
			}
break;
case 603:
#line 4552 "parser.y"
{
			    yyval.adv  = yyvsp[-2].adv ;
			    if (parse_adv_append(&yyval.adv , yyvsp[-1].adv )) {
				PARSE_ERROR;
			    }
			}
break;
case 604:
#line 4559 "parser.y"
{
			    yyerrok;
			}
break;
case 605:
#line 4567 "parser.y"
{
			    yyval.adv  = parse_adv_propagate_metric(yyvsp[-1].adv , yyvsp[-2].proto , &yyvsp[0].metric_adv .metric, yyvsp[0].metric_adv .adv);
			    parse_proto = (proto_t) 0;
			}
break;
case 606:
#line 4573 "parser.y"
{
			    yyval.adv  = parse_adv_propagate_metric(yyvsp[-1].adv , yyvsp[-2].proto , &yyvsp[0].metric_adv .metric, yyvsp[0].metric_adv .adv);
			    parse_proto = (proto_t) 0;
			    parse_gwlist = (gw_entry **) 0;
			}
break;
case 607:
#line 4580 "parser.y"
{
			    yyval.adv  = parse_adv_propagate_metric(adv_alloc(ADVFT_ANY | ADVF_FIRST,
								      (proto_t) 0),
							    yyvsp[-1].proto ,
							    &yyvsp[0].metric_adv .metric,
							    yyvsp[0].metric_adv .adv);
			    parse_proto = (proto_t) 0;
			}
break;
case 608:
#line 4590 "parser.y"
{
			    yyval.adv  = parse_adv_propagate_metric(yyvsp[-1].adv , yyvsp[-2].proto , &yyvsp[0].metric_adv .metric, yyvsp[0].metric_adv .adv);
			    parse_proto = (proto_t) 0;
			    parse_gwlist = (gw_entry **) 0;
			}
break;
case 609:
#line 4597 "parser.y"
{
			    yyval.adv  = parse_adv_propagate_metric(adv_alloc(ADVFT_ANY | ADVF_FIRST,
								      (proto_t) 0),
							    yyvsp[-1].proto ,
							    &yyvsp[0].metric_adv .metric,
							    yyvsp[0].metric_adv .adv);
			    parse_proto = (proto_t) 0;
			}
break;
case 610:
#line 4606 "parser.y"
{
			    yyval.adv  = parse_adv_propagate_metric(adv_alloc(ADVFT_AS | ADVF_FIRST,
								      (proto_t) 0),
							    yyvsp[-3].proto ,
							    &yyvsp[0].metric_adv .metric,
							    yyvsp[0].metric_adv .adv);
			    yyval.adv ->adv_as = yyvsp[-1].as ;
			    parse_proto = (proto_t) 0;
			}
break;
case 611:
#line 4617 "parser.y"
{
			    yyval.adv  = parse_adv_propagate_metric(adv_alloc(ADVFT_ASPATH | ADVF_FIRST,
								      (proto_t) 0),
							    yyvsp[-2].proto ,
							    &yyvsp[0].metric_adv .metric,
							    yyvsp[0].metric_adv .adv);
			    yyval.adv ->adv_aspath = yyvsp[-1].asp_match ;
			    parse_proto = (proto_t) 0;
			    parse_gwlist = (gw_entry **) 0;
			}
break;
case 612:
#line 4629 "parser.y"
{
			    yyval.adv  = parse_adv_propagate_metric(adv_alloc(ADVFT_TAG|ADVF_FIRST,
								      (proto_t) 0),
							    yyvsp[-3].proto ,
							    &yyvsp[0].metric_adv .metric,
							    yyvsp[0].metric_adv .adv);
			    yyval.adv ->adv_tag = yyvsp[-1].metric .metric;
			    parse_proto = (proto_t) 0;
			    parse_gwlist = (gw_entry **) 0;
			}
break;
case 613:
#line 4641 "parser.y"
{
			    yyval.adv  = parse_adv_propagate_metric(yyvsp[-1].adv , yyvsp[-2].proto , &yyvsp[0].metric_adv .metric, yyvsp[0].metric_adv .adv);
			    parse_proto = (proto_t) 0;
			    parse_gwlist = (gw_entry **) 0;
			}
break;
case 614:
#line 4648 "parser.y"
{
			    yyval.adv  = parse_adv_propagate_metric(yyvsp[-1].adv , yyvsp[-2].proto , &yyvsp[0].metric_adv .metric, yyvsp[0].metric_adv .adv);
			    parse_proto = (proto_t) 0;
			    parse_gwlist = (gw_entry **) 0;
			}
break;
case 615:
#line 4655 "parser.y"
{
			    yyval.adv  = parse_adv_propagate_metric(adv_alloc(ADVFT_ANY | ADVF_FIRST,
								      (proto_t) 0),
							    yyvsp[-1].proto ,
							    &yyvsp[0].metric_adv .metric,
							    yyvsp[0].metric_adv .adv);
			    parse_proto = (proto_t) 0;
			}
break;
case 616:
#line 4664 "parser.y"
{
			    yyval.adv  = parse_adv_propagate_metric(adv_alloc(ADVFT_ANY | ADVF_FIRST,
								      (proto_t) 0),
							    yyvsp[-1].proto ,
							    &yyvsp[0].metric_adv .metric,
							    yyvsp[0].metric_adv .adv);
			    parse_proto = (proto_t) 0;
			}
break;
case 617:
#line 4675 "parser.y"
{
			    yyval.proto  = parse_proto = RTPROTO_DIRECT;
			    parse_gwlist = (gw_entry **) 0;
			    trace_tf(trace_global,
				     TR_PARSE,
				     0,
				     ("parse: %s PROTO: %s",
				      parse_where(),
				      gd_lower(trace_state(rt_proto_bits, yyval.proto ))));
			}
break;
case 618:
#line 4688 "parser.y"
{
			    yyval.proto  = parse_proto = RTPROTO_STATIC;
			    parse_gwlist = &rt_gw_list;
			    trace_tf(trace_global,
				     TR_PARSE,
				     0,
				     ("parse: %s PROTO: %s",
				      parse_where(),
				      gd_lower(trace_state(rt_proto_bits, yyval.proto ))));
			}
break;
case 619:
#line 4701 "parser.y"
{
			    yyval.proto  = parse_proto = RTPROTO_AGGREGATE;
			    parse_gwlist = (gw_entry **) 0;
			    trace_tf(trace_global,
				     TR_PARSE,
				     0,
				     ("parse: %s PROTO: %s",
				      parse_where(),
				      gd_lower(trace_state(rt_proto_bits, yyval.proto ))));
			}
break;
case 620:
#line 4714 "parser.y"
{
			    yyval.proto  = parse_proto = RTPROTO_KERNEL;
			    parse_gwlist = &krt_gw_list;
			    trace_tf(trace_global,
				     TR_PARSE,
				     0,
				     ("parse: %s PROTO: %s",
				      parse_where(),
				      gd_lower(trace_state(rt_proto_bits, yyval.proto ))));
			}
break;
case 621:
#line 4727 "parser.y"
{
			    yyval.proto  = parse_proto = RTPROTO_DEFAULT;
			    parse_gwlist = (gw_entry **) 0;
			    trace_tf(trace_global,
				     TR_PARSE,
				     0,
				     ("parse: %s PROTO: %s",
				      parse_where(),
				      gd_lower(trace_state(rt_proto_bits, yyval.proto ))));
			}
break;
case 622:
#line 4741 "parser.y"
{
			    PARSE_METRIC_RESTRICT(&yyval.metric_adv .metric);
			    yyval.metric_adv .adv = (adv_entry *) 0;
			}
break;
case 623:
#line 4746 "parser.y"
{
			    yyval.metric_adv .metric = yyvsp[-1].metric ;
			    yyval.metric_adv .adv = yyvsp[0].adv ;
			}
break;
case 624:
#line 4754 "parser.y"
{
			    yyval.adv  = (adv_entry *) 0;
			}
break;
case 625:
#line 4758 "parser.y"
{
			    yyval.adv  = adv_destmask_finish(yyvsp[-1].adv );
			}
break;
case 626:
#line 4766 "parser.y"
{
			    yyval.adv  = (adv_entry *) 0;
			}
break;
case 627:
#line 4770 "parser.y"
{
			    yyval.adv  = adv_destmask_insert(parse_error, yyvsp[-2].adv , yyvsp[-1].adv );
			    if (!yyval.adv ) {
				PARSE_ERROR;
			    }
			}
break;
case 628:
#line 4777 "parser.y"
{
			    yyerrok;
			}
break;
case 629:
#line 4784 "parser.y"
{
			    parse_contig_masks = FALSE;
			}
break;
case 630:
#line 4788 "parser.y"
{
			    parse_contig_masks = TRUE;

			    if (yyvsp[-1].dm .dm_dest
				&& (socktype(yyvsp[-1].dm .dm_dest) != AF_INET)) {
				sprintf(parse_error, "address not INET");
				PARSE_ERROR
			    }

			    yyval.adv  = adv_alloc(ADVFT_DM, (proto_t) 0);
			    yyval.adv ->adv_dm = yyvsp[-1].dm ;
			    if (PARSE_METRIC_ISRESTRICT(&yyvsp[0].metric )) {
				BIT_SET(yyval.adv ->adv_flag, ADVF_NO);
			    } else if (PARSE_METRIC_ISSET(&yyvsp[0].metric )) {
				yyval.adv ->adv_result.res_metric = yyvsp[0].metric .metric;
				BIT_SET(yyval.adv ->adv_flag, ADVFOT_METRIC);
			    }
			}
break;
case 635:
#line 4822 "parser.y"
{
			    yyval.dm .dm_dest = sockdup(inet_addr_default);
			    yyval.dm .dm_mask = inet_mask_default;
			    yyval.dm .dm_flags = (flag_t) 0;
			    trace_tf(trace_global,
				     TR_PARSE,
				     0,
				     ("parse: %s DEST: all",
				      parse_where()));
			}
break;
case 636:
#line 4836 "parser.y"
{
			    yyval.dm .dm_dest = sockdup(inet_addr_default);
			    yyval.dm .dm_mask = inet_mask_default;
			    yyval.dm .dm_flags = (flag_t) DMF_EXACT;
			    trace_tf(trace_global,
				     TR_PARSE,
				     0,
				     ("parse: %s DEST: %A",
				      parse_where(),
				      yyval.dm .dm_dest));
			}
break;
case 637:
#line 4852 "parser.y"
{
			    yyval.adv  = adv_alloc(ADVFT_GW | ADVF_FIRST, (proto_t) 0);
			    yyval.adv ->adv_gwp = yyvsp[0].gwp ;
			}
break;
case 638:
#line 4857 "parser.y"
{
			    adv_entry *adv = adv_alloc(ADVFT_GW, (proto_t) 0);

			    adv->adv_gwp = yyvsp[0].gwp ;
			    yyval.adv  = yyvsp[-1].adv ;
			    if (parse_adv_append(&yyval.adv , adv)) {
				PARSE_ERROR;
			    }
			}
break;
case 639:
#line 4870 "parser.y"
{
			    /*
			     *	Make sure host is on a locally attached network then
			     *	find or create a gw structure for it.  Requires that
			     *	parse_proto and parse_gwlist are previously set
			     */
			    if (!if_parse_withdst(yyvsp[0].sockaddr )) {
				(void) sprintf(parse_error, "gateway not a host address on an attached network: '%A'",
					       yyvsp[0].sockaddr );
				if (BIT_TEST(task_state, TASKS_STRICTIFS)) {
				    PARSE_ERROR;
				} else {
				    yyerror(parse_error);
				}
			    }
			    if (!parse_gwlist) {
				(void) sprintf(parse_error, "gateway specification not valid for %s",
					       trace_state(rt_proto_bits, parse_proto));
				PARSE_ERROR;
			    }
			    yyval.gwp  = gw_locate(parse_gwlist,
					   parse_proto,
					   (task *) 0,
					   (as_t) 0,
					   (as_t) 0,
					   yyvsp[0].sockaddr ,
					   (flag_t) 0);
			    trace_tf(trace_global,
				     TR_PARSE,
				     0,
				     ("parse: %s GATEWAY: %A  PROTO: %s",
				      parse_where(),
				      yyval.gwp ->gw_addr,
				      gd_lower(trace_state(rt_proto_bits, yyval.gwp ->gw_proto))));
			    sockfree(yyvsp[0].sockaddr );
			}
break;
case 640:
#line 4909 "parser.y"
{ 
			    sockaddr_un *addr;

			    addr = parse_addr_hostname(yyvsp[0].str .ptr, parse_error);
			    if (!addr) {
				/* Name resolution failure */
				
				PARSE_ERROR;
			    }
			    yyval.sockaddr  = sockdup(addr);
			    trace_tf(trace_global,
				     TR_PARSE,
				     0,
				     ("parse: %s HOST: %A",
				      parse_where(),
				      yyval.sockaddr ));
			    free_charstr(yyvsp[0].str );
			}
break;
case 641:
#line 4928 "parser.y"
{
			    yyval.sockaddr  = yyvsp[0].sockaddr ;
			    trace_tf(trace_global,
				     TR_PARSE,
				     0,
				     ("parse: %s HOST: %A",
				      parse_where(),
				      yyval.sockaddr ));
			}
break;
case 642:
#line 4940 "parser.y"
{
			    yyval.dm .dm_dest = yyvsp[0].sockaddr ;
			    yyval.dm .dm_mask = inet_mask_host;
			    yyval.dm .dm_flags = DMF_EXACT;

			    trace_tf(trace_global,
				     TR_PARSE,
				     0,
				     ("parse: %s DEST: %A",
				      parse_where(),
				      yyval.dm .dm_dest));
			}
break;
case 643:
#line 4955 "parser.y"
{
			    sockaddr_un *addr;

			    addr = parse_addr_netname(yyvsp[0].str .ptr, parse_error);
			    if (!addr) {
				/* Name resolution failure */

				PARSE_ERROR;
			    }
			    yyval.sockaddr  = sockdup(addr);
			    trace_tf(trace_global,
				     TR_PARSE,
				     0,
				     ("parse: %s NETWORK: %A",
				      parse_where(),
				      yyval.sockaddr ));
			    free_charstr(yyvsp[0].str );
			}
break;
case 644:
#line 4974 "parser.y"
{
			    yyval.sockaddr  = yyvsp[0].sockaddr ;
			    trace_tf(trace_global,
				     TR_PARSE,
				     0,
				     ("parse: %s NETWORK: %A",
				      parse_where(),
				      yyval.sockaddr ));
			}
break;
case 645:
#line 4987 "parser.y"
{
			    yyval.dm .dm_dest = yyvsp[0].sockaddr ;
			    yyval.dm .dm_mask = inet_mask_natural(yyvsp[0].sockaddr );
			    yyval.dm .dm_flags = (flag_t) 0;
			    if (sock2ip(yyval.dm .dm_dest) & ~sock2ip(yyval.dm .dm_mask)) {
				sprintf(parse_error, "host or mask option needed");
				PARSE_ERROR;
			    }
			    trace_tf(trace_global,
				     TR_PARSE,
				     0,
				     ("parse: %s DEST: %A MASK: %A",
				      parse_where(),
				      yyval.dm .dm_dest,
				      yyval.dm .dm_mask));
			}
break;
case 646:
#line 5004 "parser.y"
{
			    yyval.dm .dm_dest = yyvsp[-2].sockaddr ;
			    if (parse_limit_check("inet mask bits",
						  (u_int) yyvsp[0].num ,
						  INET_LIMIT_MASKLEN)) {
				PARSE_ERROR;
			    }
			    yyval.dm .dm_mask = inet_masks[yyvsp[0].num ];
			    yyval.dm .dm_flags = (flag_t) 0;
			    trace_tf(trace_global,
				     TR_PARSE,
				     0,
				     ("parse: %s DEST: %A MASK: %A",
				      parse_where(),
				      yyval.dm .dm_dest,
				      yyval.dm .dm_mask));
			}
break;
case 647:
#line 5022 "parser.y"
{
			    register byte *ap = yyvsp[-2].sockaddr ->a.ga_data;
			    register byte *mp = yyvsp[0].sockaddr ->a.ga_data;
			    register byte *alp = (byte *) yyvsp[-2].sockaddr  + socksize(yyvsp[-2].sockaddr );
			    register byte *mlp = (byte *) yyvsp[0].sockaddr  + socksize(yyvsp[0].sockaddr );

			    /* Skip zero fields */
			    while (!*ap && ap < alp) {
				ap++, mp++;
			    }
			    
			    while (ap < alp) {
				if (mp < mlp) {
				    if (*ap++ & ~*mp++) {
					break;
				    }
				} else {
				    if (*ap++) {
					break;
				    }
				}
			    }
			    if (ap < alp) {
				sprintf(parse_error, "Non-masked bits not zero for %A mask %A",
					yyvsp[-2].sockaddr ,
					yyvsp[0].sockaddr );
				PARSE_ERROR;
			    }

			    yyval.dm .dm_dest = yyvsp[-2].sockaddr ;
			    yyval.dm .dm_mask = yyvsp[0].sockaddr ;
			    yyval.dm .dm_flags = (flag_t) 0;
			    trace_tf(trace_global,
				     TR_PARSE,
				     0,
				     ("parse: %s DEST: %A MASK: %A",
				      parse_where(),
				      yyval.dm .dm_dest,
				      yyval.dm .dm_mask));
			}
break;
case 648:
#line 5066 "parser.y"
{
			    yyval.dm  = yyvsp[-1].dm ;
			    yyval.dm .dm_flags = yyvsp[0].num ;
			}
break;
case 649:
#line 5073 "parser.y"
{
			    yyval.sockaddr  = inet_mask_locate(yyvsp[0].in_addr .s_addr);
			    if (!yyval.sockaddr ) {
				if (parse_contig_masks) {
				    (void) sprintf(parse_error, "mask not contiguous");
				    PARSE_ERROR;
				} else {
				    yyval.sockaddr  = mask_locate(sockbuild_in(0, yyvsp[0].in_addr .s_addr));
				}
			    }
			    trace_tf(trace_global,
				     TR_PARSE,
				     0,
				     ("parse: %s MASK: %A",
				      parse_where(),
				      yyval.sockaddr ));
			}
break;
case 650:
#line 5094 "parser.y"
{
			    yyval.sockaddr  = sockdup(sockbuild_in(0, yyvsp[0].in_addr .s_addr));

			    trace_tf(trace_global,
				     TR_PARSE,
				     0,
				     ("parse: %s IP_ADDR: %A",
				      parse_where(),
				      yyval.sockaddr ));
			}
break;
case 651:
#line 5108 "parser.y"
{
			    u_int32 addr = yyvsp[0].num ;
			    
			    if (addr) {
				/* Normalize the address */
				while (!(addr & 0xff000000)) {
				    addr <<= 8;
				}
			    }
			    yyval.in_addr .s_addr = htonl(addr);
			}
break;
case 652:
#line 5120 "parser.y"
{
			    int i;
			    u_int32 addr = 0;

			    if (yyvsp[0].hex .len > sizeof(addr)) {
				sprintf(parse_error, "octet or hex string too long to be an IP address");
				PARSE_ERROR;
			    }

			    for (i = 0; i < sizeof(addr); i++) {
				addr <<= 8;
				if (i < yyvsp[0].hex .len) {
				    BIT_SET(addr, yyvsp[0].hex .ptr[i]);
				}
			    }
			    yyval.in_addr .s_addr = htonl(addr);
			    free_bytestr(yyvsp[0].hex );
			}
break;
case 653:
#line 5142 "parser.y"
{
			    if (parse_limit_check("ttl",
						  (u_int) yyvsp[0].num ,
						  INET_LIMIT_TTL)) {
				PARSE_ERROR;
			    }
			    yyval.num  = yyvsp[0].num ;
			}
break;
case 654:
#line 5159 "parser.y"
{
			    assert(FALSE);
			}
break;
case 656:
#line 5166 "parser.y"
{
			    assert(FALSE);
			}
break;
case 658:
#line 5173 "parser.y"
{
			    assert(FALSE);
			}
break;
case 660:
#line 5180 "parser.y"
{
			    assert(FALSE);
			}
break;
case 662:
#line 5187 "parser.y"
{
			    yyval.num  = (flag_t) 0;
			}
break;
case 663:
#line 5191 "parser.y"
{
			    yyval.num  = DMF_EXACT;
			}
break;
case 664:
#line 5195 "parser.y"
{
			    yyval.num  = DMF_REFINE;
			}
break;
case 665:
#line 5204 "parser.y"
{
			    yyval.str  = yyvsp[0].str ;
			    yyval.str .ptr = task_mem_strdup((task *) 0, yyvsp[0].str .ptr);
			}
break;
case 666:
#line 5209 "parser.y"
{
			    yyval.str  = yyvsp[0].str ;
			    yyval.str .ptr = task_mem_strdup((task *) 0, yyvsp[0].str .ptr);
			}
break;
case 668:
#line 5218 "parser.y"
{
#if	defined(PROTO_ICMP) || defined(KRT_RT_SOCK)
			    yyval.proto_lists .proto = RTPROTO_REDIRECT;
			    yyval.proto_lists .import_list = &redirect_import_list;
			    yyval.proto_lists .export_list = (adv_entry **) 0;
			    yyval.proto_lists .gw_list = &redirect_gw_list;
#endif	/* defined(PROTO_ICMP) || defined(KRT_RT_SOCK) */
			}
break;
case 669:
#line 5227 "parser.y"
{
			    yyval.proto_lists .proto = RTPROTO_HELLO;
			    yyval.proto_lists .import_list = &hello_import_list;
			    yyval.proto_lists .export_list = &hello_export_list;
			    yyval.proto_lists .gw_list = &hello_gw_list;
			}
break;
case 670:
#line 5234 "parser.y"
{
			    yyval.proto_lists .proto = RTPROTO_RIP;
			    yyval.proto_lists .import_list = &rip_import_list;
			    yyval.proto_lists .export_list = &rip_export_list;
			    yyval.proto_lists .gw_list = &rip_gw_list;
			}
break;
case 671:
#line 5244 "parser.y"
{
			    assert(FALSE);
			}
break;
case 673:
#line 5250 "parser.y"
{ yyval.num  = T_ON; }
break;
case 674:
#line 5251 "parser.y"
{ yyval.num  = T_OFF; }
break;
case 675:
#line 5257 "parser.y"
{
			    PARSE_METRIC_SET(&yyval.metric , yyvsp[0].num );
			    trace_tf(trace_global,
				     TR_PARSE,
				     0,
				     ("parse: %s METRIC: %d",
				      parse_where(),
				      yyval.metric .metric));
			}
break;
case 676:
#line 5267 "parser.y"
{
			    PARSE_METRIC_INFINITY(&yyval.metric );
			    trace_tf(trace_global,
				     TR_PARSE,
				     0,
				     ("parse: %s METRIC: infinity",
				      parse_where()));
			}
break;
case 677:
#line 5278 "parser.y"
{
			    PARSE_METRIC_CLEAR(&yyval.metric );
			}
break;
case 678:
#line 5282 "parser.y"
{
			    if (parse_metric_check(parse_export_proto, &yyvsp[0].metric )) {
				PARSE_ERROR;
			    }
			    yyval.metric  = yyvsp[0].metric ;
			}
break;
case 679:
#line 5292 "parser.y"
{
			    yyval.metric  = yyvsp[0].metric ;
			}
break;
case 680:
#line 5296 "parser.y"
{
			    PARSE_METRIC_RESTRICT(&yyval.metric );
			}
break;
case 681:
#line 5304 "parser.y"
{
			    PARSE_METRIC_CLEAR(&yyval.metric );
			}
break;
case 682:
#line 5308 "parser.y"
{
			    PARSE_METRIC_SET(&yyval.metric , yyvsp[0].pref );
			}
break;
case 684:
#line 5316 "parser.y"
{
			    PARSE_METRIC_RESTRICT(&yyval.metric );
			}
break;
case 685:
#line 5322 "parser.y"
{
			    if (parse_limit_check("preference",
						  (u_int) yyvsp[0].num ,
						  LIMIT_PREFERENCE)) {
				PARSE_ERROR;
			    }
			    yyval.pref  = yyvsp[0].num ;
			}
break;
case 686:
#line 5335 "parser.y"
{
			    /* Remove quotes from the string if present */
			    char *cp;

			    yyval.str  = yyvsp[0].str ;
				
			    switch (*yyval.str .ptr) {
			    case '<':
				*yyval.str .ptr = '>';
				/* Fall through */

			    case '"':
				cp = &yyval.str .ptr[yyval.str .len-1];
				if (*cp == *yyval.str .ptr) {
				    *cp = (char) 0;
				    yyval.str .len -= 2;
				    cp = yyval.str .ptr + 1;
				    break;
				}
				/* Fall Through */

			    default:
				cp = yyval.str .ptr;
			    }

			    yyval.str .ptr = task_mem_strdup((task *) 0, cp);
			    trace_tf(trace_global,
				     TR_PARSE,
				     0,
				     ("parse: %s STRING: \"%s\" length: %d",
				      parse_where(),
				      yyval.str .ptr,
				      yyval.str .len));
			}
break;
case 688:
#line 5374 "parser.y"
{
			    yyval.num  = -yyvsp[0].num ;
			}
break;
case 690:
#line 5381 "parser.y"
{
			    int i = 0;
			    int len = sizeof(yyval.num );

			    yyval.num  = 0;
			    if (yyvsp[0].hex .len > len) {
				sprintf(parse_error, "octet or hex string too long to be a number");
				PARSE_ERROR;
			    } else if (yyvsp[0].hex .len < len) {
				len = yyvsp[0].hex .len;
			    }
			    while (i < len) {
				yyval.num  = (yyval.num  << 8) | yyvsp[0].hex .ptr[i++];
			    }
			    free_bytestr(yyvsp[0].hex );
			}
break;
case 693:
#line 5406 "parser.y"
{
			    if (parse_limit_check("seconds",
						  (u_int) yyvsp[0].num ,
						  LIMIT_NATURAL)) {
				PARSE_ERROR;
			    };
			    yyval.time  = yyvsp[0].num ;
			}
break;
case 694:
#line 5415 "parser.y"
{
			    if (parse_limit_check("minutes",
						  (u_int) yyvsp[-2].num ,
						  LIMIT_NATURAL)) {
				PARSE_ERROR;
			    }
			    if (parse_limit_check("seconds",
						  (u_int) yyvsp[0].num ,
						  LIMIT_SECONDS)) {
				PARSE_ERROR;
			    }
			    yyval.time  = (yyvsp[-2].num  * 60) + yyvsp[0].num ;
			}
break;
case 695:
#line 5429 "parser.y"
{
			    if (parse_limit_check("hours",
						  (u_int) yyvsp[-4].num ,
						  LIMIT_NATURAL)) {
				PARSE_ERROR;
			    }
			    if (parse_limit_check("minutes",
						  (u_int) yyvsp[-2].num ,
						  LIMIT_MINUTES)) {
				PARSE_ERROR;
			    }
			    if (parse_limit_check("seconds",
						  (u_int) yyvsp[0].num ,
						  LIMIT_SECONDS)) {
				PARSE_ERROR;
			    }
			    yyval.time  = ((yyvsp[-4].num  * 60) + yyvsp[-2].num ) * 60 + yyvsp[0].num ;
			}
break;
case 696:
#line 5455 "parser.y"
{
			    yyval.u_num  = 1;
			}
break;
case 697:
#line 5459 "parser.y"
{
			    if (parse_limit_check("as loop limit",
						  (u_int) yyvsp[0].num ,
						  LIMIT_AS_LOOPS)) {
				PARSE_ERROR;
			    }
			    yyval.u_num  = yyvsp[0].num ;
			}
break;
case 698:
#line 5471 "parser.y"
{
			    if (parse_limit_check("autonomous system",
						  (u_int) yyvsp[0].num ,
						  LIMIT_AS)) {
				PARSE_ERROR;
			    }
			    yyval.as  = yyvsp[0].num ;
			}
break;
case 699:
#line 5488 "parser.y"
{
			    PARSE_METRIC_CLEAR(&yyval.metric );
			}
break;
case 700:
#line 5492 "parser.y"
{
			    yyval.metric  = yyvsp[0].metric ;
			}
break;
case 702:
#line 5499 "parser.y"
{
			    PARSE_METRIC_ALTERNATE(&yyval.metric , 0);
			}
break;
case 703:
#line 5503 "parser.y"
{
			    if (parse_limit_check("external-route-tag",
						  (u_int) yyvsp[0].num ,
						  PATH_OSPF_TAG_USR_LIMIT)) {
				PARSE_ERROR;
			    }
			    PARSE_METRIC_ALTERNATE(&yyval.metric , yyvsp[0].num );
			}
break;
case 704:
#line 5514 "parser.y"
{
			    PARSE_METRIC_CLEAR(&yyval.metric );
			}
break;
case 705:
#line 5518 "parser.y"
{
			    yyval.metric  = yyvsp[0].metric ;
			}
break;
case 706:
#line 5524 "parser.y"
{
			    if (parse_limit_check("external-route-tag",
						  (u_int) yyvsp[0].num ,
						  PATH_OSPF_TAG_LIMIT)) {
				PARSE_ERROR;
			    }
			    PARSE_METRIC_SET(&yyval.metric , yyvsp[0].num );
			}
break;
case 712:
#line 5548 "parser.y"
{
			    yyval.proto  = parse_proto = RTPROTO_ANY;
			    parse_gwlist = (gw_entry **) 0;
			    trace_tf(trace_global,
				     TR_PARSE,
				     0,
				     ("parse: %s PROTO: %s",
				      parse_where(),
				      gd_lower(trace_state(rt_proto_bits, yyval.proto ))));
			}
break;
case 718:
#line 5566 "parser.y"
{
			    aspath_init_regex();
			}
break;
case 719:
#line 5570 "parser.y"
{
			    yyval.asp_match  = aspath_consume_current(yyvsp[-2].asp_stack );
			    yyval.asp_match ->origin_mask = yyvsp[0].flag ;
			    trace_tf(trace_global,
				     TR_PARSE,
				     0,
				     ("%s: path %X origins %s",
				      parse_where(),
				      yyval.asp_match ,
				      aspath_adv_origins(yyval.asp_match ->origin_mask)));
			}
break;
case 720:
#line 5584 "parser.y"
{
			    /* Set all bits on */
			    yyval.flag  = (flag_t) -1;
			}
break;
case 721:
#line 5589 "parser.y"
{
			    /* Set all bits on */
			    yyval.flag  = (flag_t) -1;
			}
break;
case 722:
#line 5594 "parser.y"
{
			    yyval.flag  = yyvsp[0].flag ;
			}
break;
case 724:
#line 5602 "parser.y"
{
			    yyval.flag  = yyvsp[-2].flag  | yyvsp[0].flag ;
			}
break;
case 725:
#line 5608 "parser.y"
{
			    yyval.flag  = (1 << PATH_ORG_IGP);
			}
break;
case 726:
#line 5612 "parser.y"
{
			    yyval.flag  = (1 << PATH_ORG_EGP);
			}
break;
case 727:
#line 5616 "parser.y"
{
			    yyval.flag  = (1 << PATH_ORG_XX);
			}
break;
case 728:
#line 5622 "parser.y"
{
			    aspath_simple_regex(&yyval.asp_stack , &yyvsp[0].asp_stack );
			}
break;
case 729:
#line 5626 "parser.y"
{
			    aspath_merge_regex(&yyval.asp_stack , &yyvsp[-2].asp_stack , &yyvsp[0].asp_stack );
			}
break;
case 730:
#line 5632 "parser.y"
{
			    aspath_copy_regex(&yyval.asp_stack , &yyvsp[0].asp_stack );
			}
break;
case 731:
#line 5636 "parser.y"
{
			    if(!aspath_prepend_regex(&yyval.asp_stack , &yyvsp[-1].asp_stack , &yyvsp[0].asp_stack )) {
				YYERROR;
			    }
			}
break;
case 732:
#line 5644 "parser.y"
{
			    aspath_copy_regex(&yyval.asp_stack , &yyvsp[0].asp_stack );
			}
break;
case 733:
#line 5648 "parser.y"
{
			    aspath_zero_or_more_term(&yyval.asp_stack , &yyvsp[-1].asp_stack );
			}
break;
case 734:
#line 5652 "parser.y"
{
			    aspath_one_or_more_term(&yyval.asp_stack , &yyvsp[-1].asp_stack );
			}
break;
case 735:
#line 5656 "parser.y"
{
			    aspath_zero_or_one_term(&yyval.asp_stack , &yyvsp[-1].asp_stack );
			}
break;
case 736:
#line 5660 "parser.y"
{
			    if (!aspath_range_term(&yyval.asp_stack , &yyvsp[-3].asp_stack , &yyvsp[-1].asp_range )) {
				sprintf(parse_error,
					"too many AS path regexp states");
				PARSE_ERROR;
			    }
			}
break;
case 737:
#line 5670 "parser.y"
{
			    if (!aspath_as_transition(&yyval.asp_stack , yyvsp[0].as )) {
				sprintf(parse_error,
					"too many AS path regexp states");
				PARSE_ERROR;
			    }
			}
break;
case 738:
#line 5678 "parser.y"
{
			    if (!aspath_any_transition(&yyval.asp_stack )) {
				sprintf(parse_error,
					"too many AS path regexp states");
				PARSE_ERROR;
			    }
			}
break;
case 739:
#line 5686 "parser.y"
{
			    aspath_copy_regex(&yyval.asp_stack , &yyvsp[-1].asp_stack );
			}
break;
case 740:
#line 5691 "parser.y"
{
			    yyval.asp_range .begin = yyvsp[0].num ;
			    yyval.asp_range .end = yyvsp[0].num ;
			    if (!yyvsp[0].num ) {
				sprintf(parse_error, "invalid range start: %d",
					yyvsp[0].num );
				PARSE_ERROR;
			    }
			    trace_tf(trace_global,
				     TR_PARSE,
				     0,
				     ("%s: RANGE: {%d}",
				      parse_where(),
				      yyval.asp_range .begin));
			}
break;
case 741:
#line 5707 "parser.y"
{
			    yyval.asp_range .begin = yyvsp[-1].num ;
			    yyval.asp_range .end = 0;
			    if (!yyvsp[-1].num ) {
				sprintf(parse_error, "invalid range start: %d",
					yyvsp[-1].num );
				PARSE_ERROR;
			    }
			    trace_tf(trace_global,
				     TR_PARSE,
				     0,
				     ("%s: RANGE: {%d,}",
				      parse_where(),
				      yyval.asp_range .begin));
			}
break;
case 742:
#line 5723 "parser.y"
{
			    yyval.asp_range .begin = yyvsp[-2].num ;
			    yyval.asp_range .end = yyvsp[0].num ;
			    if (!yyvsp[-2].num ) {
				sprintf(parse_error, "invalid range start: %d",
					yyvsp[-2].num );
				PARSE_ERROR;
			    }
			    if (yyvsp[0].num  < yyvsp[-2].num  || !yyvsp[0].num ) {
				sprintf(parse_error, "invalid range end: %d",
					yyvsp[0].num );
				PARSE_ERROR;
			    }
			    trace_tf(trace_global,
				     TR_PARSE,
				     0,
				     ("%s: RANGE: {%d,%d}",
				      parse_where(),
				      yyval.asp_range .begin,
				      yyval.asp_range .end));
			}
break;
#line 8168 "y.tab.c"
    }
    yyssp -= yym;
    yystate = *yyssp;
    yyvsp -= yym;
    yym = yylhs[yyn];
    if (yystate == 0 && yym == 0)
    {
#ifdef YYDEBUG
        if (yydebug)
            printf("yydebug: after reduction, shifting from state 0 to\
 state %d\n", YYFINAL);
#endif
        yystate = YYFINAL;
        *++yyssp = YYFINAL;
        *++yyvsp = yyval;
        if (yychar < 0)
        {
            if ((yychar = yylex()) < 0) yychar = 0;
#if YYDEBUG
            if (yydebug)
            {
                yys = 0;
                if (yychar <= YYMAXTOKEN) yys = yyname[yychar];
                if (!yys) yys = "illegal-symbol";
                printf("yydebug: state %d, reading %d (%s)\n",
                        YYFINAL, yychar, yys);
            }
#endif
        }
        if (yychar == 0) goto yyaccept;
        goto yyloop;
    }
    if ((yyn = yygindex[yym]) && (yyn += yystate) >= 0 &&
            yyn <= YYTABLESIZE && yycheck[yyn] == yystate)
        yystate = yytable[yyn];
    else
        yystate = yydgoto[yym];
#ifdef YYDEBUG
    if (yydebug)
        printf("yydebug: after reduction, shifting from state %d \
to state %d\n", *yyssp, yystate);
#endif
    if (yyssp >= yyss + yystacksize - 1)
    {
        goto yyoverflow;
    }
    *++yyssp = yystate;
    *++yyvsp = yyval;
    goto yyloop;
yyoverflow:
    yyerror("yacc stack overflow");
yyabort:
    return (1);
yyaccept:
    return (0);
}
